// ---------------------------------------------------------------------------
// CollectIdentifiers.java
// A class which is dedicated to the collection of all identifier names in declarations.
//
// Author: Franz Reichl <franz.x.reichl@gmail.com>
// $Id: CollectIdentifiers.java,v 1.3 2020/05/25 11:42:22 schreine Exp $
// 
// Author: Franz Reichl <franz.x.reichl@gmail.com>
// Copyright (C) 2019-, Research Institute for Symbolic Computation (RISC)
// Johannes Kepler University, Linz, Austria, http://www.risc.jku.at
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
// ---------------------------------------------------------------------------
package riscal.smt;

import java.util.HashSet;
import java.util.List;
import java.util.Set;

import riscal.smt.Utilities.Triple;
import riscal.syntax.AST;


public class CollectIdentifiers
{
	private Set<String> funcIdentStrings = new HashSet<String>();
	private Set<String> valIdentStrings = new HashSet<String>();
	private Set<String> typeIdentStrings = new HashSet<String>();
	
	/***************************************************************************
	 * Returns the the strings of the identifiers which occur in the given declarations
	 * @param declst A list of declarations which shall be processed.
	 * @param decl An additional declaration which shall be processed.
	 * @return The first component of the pair contains the strings of the function 
	 * 		like entities, the second the strings of the value like entities
	 * 		and the third the names of introduced types.
	 **************************************************************************/
	public static Triple<Set<String>,Set<String>,Set<String>> compute(List<AST.Declaration> declst, AST.Declaration decl)
  {
  	CollectIdentifiers visitor = new CollectIdentifiers();
  	for(AST.Declaration dec: declst)
  	{
  		visitor.process(dec);
  	}
  	visitor.process(decl);
    return new Triple<Set<String>,Set<String>,Set<String>>(visitor.funcIdentStrings,visitor.valIdentStrings,visitor.typeIdentStrings);
  }
	
	public static Triple<Set<String>,Set<String>,Set<String>> compute(List<AST.Declaration> declst1, List<AST.Declaration.TheoremParamDefinition> declst2)
  {
  	CollectIdentifiers visitor = new CollectIdentifiers();
  	for(AST.Declaration dec: declst1)
  	{
  		visitor.process(dec);
  	}
  	for(AST.Declaration dec: declst2)
  	{
  		visitor.process(dec);
  	}
    return new Triple<Set<String>,Set<String>,Set<String>>(visitor.funcIdentStrings,visitor.valIdentStrings,visitor.typeIdentStrings);
  }
	
	/***************************************************************************
	 * Returns the the strings of the identifiers which occur in the given declarations
	 * @param declst A list of declarations which shall be processed.
	 * @return The first component of the pair contains the strings of the function 
	 * 		like entities, the second the strings of the value like entities
	 * 		and the third the names of introduced types.
	 **************************************************************************/
  public static Triple<Set<String>,Set<String>,Set<String>> compute(List<AST.Declaration> declst)
  {
  	CollectIdentifiers visitor = new CollectIdentifiers();
  	for(AST.Declaration dec: declst)
  	{
  		visitor.process(dec);
  	}
  	return new Triple<Set<String>,Set<String>,Set<String>>(visitor.funcIdentStrings,visitor.valIdentStrings,visitor.typeIdentStrings);
  }
  
  /***************************************************************************
   * Returns the the strings of the identifiers which occur in the given declarations
   * @param dec The declaration which shall be processed.
   * @return The first component of the pair contains the strings of the function 
   * 		like entities, the second the strings of the value like entities
   * 		and the third the names of introduced types.
   **************************************************************************/
  public static Triple<Set<String>,Set<String>,Set<String>> compute(AST.Declaration dec)
  {
  	CollectIdentifiers visitor = new CollectIdentifiers();
  	visitor.process(dec);
  	return new Triple<Set<String>,Set<String>,Set<String>>(visitor.funcIdentStrings,visitor.valIdentStrings,visitor.typeIdentStrings);
  }
  
  /***************************************************************************
   * Determines the identifier names introduced by the given declaration.
   * @param decl The declaration whose identifier name we want to retrieve.
   * @return The identifier name.
   **************************************************************************/
  private Void process(AST.Declaration decl)
  {
  	if(decl instanceof AST.Declaration.FunctionDeclaration)
  		funcIdentStrings.add(((AST.Declaration.FunctionDeclaration)decl).ident.string);
  	else if(decl instanceof AST.Declaration.PredicateDeclaration)
  		funcIdentStrings.add(((AST.Declaration.PredicateDeclaration)decl).ident.string);
  	else if(decl instanceof AST.Declaration.ProcedureDeclaration)
  		funcIdentStrings.add(((AST.Declaration.ProcedureDeclaration)decl).ident.string);
  	else if(decl instanceof AST.Declaration.FunctionDefinition)
  		funcIdentStrings.add(((AST.Declaration.FunctionDefinition)decl).ident.string);
  	else if(decl instanceof AST.Declaration.PredicateDefinition)
  		funcIdentStrings.add(((AST.Declaration.PredicateDefinition)decl).ident.string);
  	else if(decl instanceof AST.Declaration.TheoremParamDefinition)
  		funcIdentStrings.add(((AST.Declaration.TheoremParamDefinition)decl).ident.string);
  	else if(decl instanceof AST.Declaration.ProcedureDefinition)
  		funcIdentStrings.add(((AST.Declaration.ProcedureDefinition)decl).ident.string);
  	else if(decl instanceof AST.Declaration.ValueDeclaration)
  		valIdentStrings.add(((AST.Declaration.ValueDeclaration)decl).ident.string);
  	else if(decl instanceof AST.Declaration.ValueDefinition)
  		valIdentStrings.add(((AST.Declaration.ValueDefinition)decl).ident.string);
  	else if(decl instanceof AST.Declaration.PredicateValueDefinition)
  		valIdentStrings.add(((AST.Declaration.PredicateValueDefinition)decl).ident.string);
  	else if(decl instanceof AST.Declaration.TheoremDefinition)
  		valIdentStrings.add(((AST.Declaration.TheoremDefinition)decl).ident.string);
  	else if(decl instanceof AST.Declaration.TypeDefinition)
  		typeIdentStrings.add(((AST.Declaration.TypeDefinition)decl).ident.string);
  	return null;
  }

  
}
//---------------------------------------------------------------------------
// end of file
//---------------------------------------------------------------------------
