// ---------------------------------------------------------------------------
// DeclarationTranslator.java
// A class which provides the functionality to determine the symbols from a declarations
//		a given theorem relys on.
//
// Author: Franz Reichl <franz.x.reichl@gmail.com>
// $Id: DependingSymbols.java,v 1.2 2020/02/18 08:04:58 schreine Exp $
// 
// Author: Franz Reichl <franz.x.reichl@gmail.com>
// Copyright (C) 2019-, Research Institute for Symbolic Computation (RISC)
// Johannes Kepler University, Linz, Austria, http://www.risc.jku.at
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
// ---------------------------------------------------------------------------
package riscal.smt;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;

import riscal.syntax.AST;
import riscal.syntax.AST.Declaration;
import riscal.syntax.AST.Declaration.FunctionDeclaration;
import riscal.syntax.AST.Declaration.FunctionDefinition;
import riscal.syntax.AST.Declaration.PredicateDeclaration;
import riscal.syntax.AST.Declaration.PredicateDefinition;
import riscal.syntax.AST.Declaration.PredicateValueDefinition;
import riscal.syntax.AST.Declaration.ProcedureDeclaration;
import riscal.syntax.AST.Declaration.ProcedureDefinition;
import riscal.syntax.AST.Declaration.TheoremDefinition;
import riscal.syntax.AST.Declaration.TheoremParamDefinition;
import riscal.syntax.AST.Declaration.TypeDefinition;
import riscal.syntax.AST.Declaration.ValueDefinition;
import riscal.syntax.AST.Expression;
import riscal.syntax.AST.Expression.ApplicationExp;
import riscal.syntax.AST.Expression.IdentifierExp;
import riscal.syntax.AST.FunctionSpec;
import riscal.syntax.AST.Parameter;
import riscal.syntax.AST.Type;
import riscal.syntax.AST.Type.ArrayType;
import riscal.syntax.AST.Type.IntType;
import riscal.syntax.AST.Type.MapType;
import riscal.syntax.AST.Type.NatType;
import riscal.syntax.AST.Type.RecordType;
import riscal.syntax.AST.Type.SetType;
import riscal.syntax.AST.Type.SubType;
import riscal.syntax.AST.Type.TupleType;
import riscal.syntax.AST.Type.UnitType;
import riscal.syntax.ASTVisitor;
import riscal.types.Environment.Symbol;
import riscal.types.Environment.Symbol.FunctionSymbol;

public class DependingSymbols extends ASTVisitor.Base<Void>
{	
	/***************************************************************************
	 * The symbol introduced by a visited declaration
	 **************************************************************************/
	private Symbol sym;

	/***************************************************************************
	 * The symbols of functions, which occur in the visited declaration.
	 **************************************************************************/
	private Set<Symbol> occuringFunctionSymbols;

	/***************************************************************************
	 * The symbols of values, which occur in the visited declarations.
	 **************************************************************************/
	private Set<Symbol> occuringValueSymbols;
	
	/***************************************************************************
	 * True if the visited declaration is either a function or
	 * a predicate definition, whose definition shall be preserved
	 **************************************************************************/
	private boolean isFunctionDefinition;
	
	
	
	private DependingSymbols()	
	{
		sym=null;
		occuringFunctionSymbols=new HashSet<Symbol>();
		occuringValueSymbols=new HashSet<Symbol>();
		isFunctionDefinition=false;		
	}
	
	/***************************************************************************
	 * Determines all the symbols the denoted theorem depends on. 
	 * This also includes symbols on which the symbols on which the theorem
	 * depends, depend (transitive closure).
	 * This set is affected by SMT.functionDefinitionTreatmentMethod
	 * If a function shall be treated modular also axioms (theorems)
	 * which contain the respective symbols are considered for this computation.
	 * @param decs The declarations on which the theorem may depend
	 * @param th The theorem which is the base of this dependency-tree
	 * @return A set containing the symbols the theorem (transitively)
	 * 		depends on.
	 **************************************************************************/
	public static Set<Symbol> giveDependingSymbols(List<Declaration> decs,TheoremParamDefinition th)
	{
		if(SMT.skipDeclarations==SMT.SkipMode.RetainAll)
			return null;
		int len=decs.size()+1;
		DependingSymbols dep=new DependingSymbols();
		Map<FunctionSymbol,Boolean> isDefined=new HashMap<FunctionSymbol, Boolean>(len);
		Map<Symbol,Set<Symbol>> occuringFunSymbols=new LinkedHashMap<Symbol, Set<Symbol>>();
		Map<Symbol,Set<Symbol>> occuringValSymbols=new LinkedHashMap<Symbol, Set<Symbol>>();
		Map<Symbol,Set<Symbol>> occuringFunSymbolsInTheorems=new LinkedHashMap<Symbol, Set<Symbol>>();
		Map<Symbol,Set<Symbol>> occuringValSymbolsInTheorems=new LinkedHashMap<Symbol, Set<Symbol>>();
		List<Symbol> values=new ArrayList<Symbol>();
		Map<Symbol,Set<Symbol>>typeFunSyms=new LinkedHashMap<Symbol, Set<Symbol>>();
		Map<Symbol,Set<Symbol>>typeValSyms=new LinkedHashMap<Symbol, Set<Symbol>>();
		for(Declaration dec:decs)
		{
			if(SMT.skipDeclarations==SMT.SkipMode.SkipUnused)
			{
				if(dec instanceof TheoremDefinition||dec instanceof TheoremParamDefinition)
					continue;
			}						
			if(SMT.skipDeclarations==SMT.SkipMode.RetainAxioms)
			{
				if(dec instanceof TheoremDefinition)
				{
					if(!((TheoremDefinition)dec).axiom)
						continue;
				}
				else if(dec instanceof TheoremParamDefinition)
				{
					if(!((TheoremParamDefinition)dec).axiom)
						continue;
				}
			}
			dep=compute(dec);
			if(dec instanceof TypeDefinition)
			{
				typeFunSyms.put(dep.sym,dep.occuringFunctionSymbols);
				typeValSyms.put(dep.sym, dep.occuringValueSymbols);
				continue;
			}
			else if(dec instanceof ValueDefinition||dec instanceof PredicateValueDefinition)
			{
				values.add(dep.sym);
				occuringFunSymbols.put(dep.sym, dep.occuringFunctionSymbols);
				occuringValSymbols.put(dep.sym, dep.occuringValueSymbols);
			}
			else if(dec instanceof FunctionDeclaration||dec instanceof PredicateDeclaration
					||dec instanceof ProcedureDeclaration||dec instanceof ProcedureDefinition)
			{
				isDefined.put((FunctionSymbol)dep.sym, Boolean.valueOf(false));
				occuringFunSymbols.put(dep.sym, dep.occuringFunctionSymbols);
				occuringValSymbols.put(dep.sym, dep.occuringValueSymbols);
			}
			else if(dec instanceof FunctionDefinition||dec instanceof PredicateDefinition)
			{
				isDefined.put((FunctionSymbol)dep.sym, Boolean.valueOf(dep.isFunctionDefinition));
				occuringFunSymbols.put(dep.sym, dep.occuringFunctionSymbols);
				occuringValSymbols.put(dep.sym, dep.occuringValueSymbols);
			}
			else if(dec instanceof TheoremParamDefinition||dec instanceof TheoremDefinition)
			{
				occuringFunSymbolsInTheorems.put(dep.sym, dep.occuringFunctionSymbols);
				occuringValSymbolsInTheorems.put(dep.sym, dep.occuringValueSymbols);
			}			
		}
		Iterator<Entry<Symbol,Set<Symbol>>> it1=occuringValSymbols.entrySet().iterator();
		while(it1.hasNext())
		{
			Entry<Symbol,Set<Symbol>>e=it1.next();
			Symbol sym=e.getKey();
			Set<Symbol> syms=e.getValue();
			syms.retainAll(values);
			occuringFunSymbols.get(sym).addAll(syms);
		}
		Iterator<Entry<Symbol,Set<Symbol>>> it2=occuringValSymbolsInTheorems.entrySet().iterator();
		while(it2.hasNext())
		{
			Entry<Symbol,Set<Symbol>>e=it2.next();
			Symbol sym=e.getKey();
			Set<Symbol> syms=e.getValue();
			syms.retainAll(values);
			occuringFunSymbolsInTheorems.get(sym).addAll(syms);
		}		
		Iterator<Entry<Symbol,Set<Symbol>>> it3=typeValSyms.entrySet().iterator();
		while(it3.hasNext())
		{
			Entry<Symbol,Set<Symbol>>e=it3.next();
			Symbol sym=e.getKey();
			Set<Symbol> syms=e.getValue();
			syms.retainAll(values);
			typeFunSyms.get(sym).addAll(syms);
		}
		
		dep=compute(th);

//		Set<Symbol> dependingSyms=new HashSet<Symbol>();
////	dependingSyms.add(dep.sym);
//	Set<Symbol> thvals=dep.occuringValueSymbols;
//	thvals.retainAll(values);
//	thvals.addAll(dep.occuringFunctionSymbols);
//	
//	Queue<Symbol> remainingDependingSyms=new LinkedList<Symbol>(thvals);
//	Symbol s=remainingDependingSyms.poll();
//	while(s!=null)
//	{
//		s=remainingDependingSyms.poll();
//		dependingSyms.add(s);
//		for(Symbol sy:occuringFunSymbols.get(s))
//		{
//			if(!dependingSyms.contains(sy))
//				remainingDependingSyms.add(sy);
//		}
//	}
		
		Set<Symbol> dependingSyms=new HashSet<Symbol>();
		dependingSyms.add(dep.sym);
		Set<Symbol> thvals=dep.occuringValueSymbols;
		thvals.retainAll(values);
		thvals.addAll(dep.occuringFunctionSymbols);
		for(Symbol s:thvals)
		{
			addSymbols(s,occuringFunSymbols,dependingSyms);
		}
		Iterator<Entry<Symbol,Set<Symbol>>> it4=typeFunSyms.entrySet().iterator();
		while(it4.hasNext())
		{
			for(Symbol s:it4.next().getValue())
			{
				addSymbols(s, occuringFunSymbols, dependingSyms);
			}			
		}	
		if(SMT.skipDeclarations!=SMT.SkipMode.SkipUnused)
		{
			boolean stop=true;
			do
			{
				stop=true;
				Iterator<Entry<Symbol,Set<Symbol>>> it=occuringFunSymbolsInTheorems.entrySet().iterator();
				while(it.hasNext())
				{
					Entry<Symbol,Set<Symbol>>e=it.next();
					Symbol sym=e.getKey();
					Set<Symbol> syms=e.getValue();
					if(syms==null)
						continue;
					boolean add=false;
					for(Symbol s:syms)
					{
						if(s instanceof FunctionSymbol)
						{
							FunctionSymbol fs=(FunctionSymbol)s;
							if(!isDefined.get(fs))
							{
								if(dependingSyms.contains(fs))
								{
									add=true;
									break;
								}							
							}
						}
					}
					if(add)
					{
						dependingSyms.add(sym);
						syms.removeAll(dependingSyms);
						if(!syms.isEmpty())
						{
							stop=false;
							dependingSyms.addAll(syms);
							for(Symbol s:syms)
							{
								addSymbols(s, occuringFunSymbols, dependingSyms);
							}
						}
						occuringFunSymbolsInTheorems.put(sym, null);
					}
				}
			}while(!stop);
		}
		return dependingSyms;		
	}
	
	
	/***************************************************************************
	 * Auxiliary function for the dependency computation.
	 * Adds the symbol sym and the symbols it depends on,
	 * which are not already in set, to set.
	 * @param sym The symbol of which we want to determine the symbols it depends on.
	 * @param occuringSymbols Assigns to each symbol the symbols it depends on.
	 * @param set The set to which we want to add the symbols on which sym depends on.
	  **************************************************************************/
	private static void addSymbols(Symbol sym,Map<Symbol,Set<Symbol>> occuringSymbols,Set<Symbol> set)
	{
		set.add(sym);
		Set<Symbol> aux=occuringSymbols.get(sym);
		if(aux==null)
		{
			SMT.illegalUsage(null, "A problem occured during cutting the specification. You probably used a theorem as a predicate/value");
		}
		aux.removeAll(set);
		set.addAll(aux);
		for(Symbol s:aux)
		{
			addSymbols(s, occuringSymbols, set);
		}
		return;
	}
	
	private static DependingSymbols compute(AST.Declaration dec)
	{
		DependingSymbols dep=new DependingSymbols();
		dec.accept(dep);
		return dep;
	}
	
	//The visit methods
	
  public Void visit(FunctionDeclaration decl)
  {
  	sym=(FunctionSymbol)decl.ident.getSymbol();
  	decl.type.accept(this);
  	for(Parameter p:decl.param)
  	{
  		p.accept(this);
  	}
    return null;
  }
  public Void visit(FunctionDefinition decl)
  {
  	sym=(FunctionSymbol)decl.ident.getSymbol();
  	decl.type.accept(this);
  	for(Parameter p:decl.param)
  	{
  		p.accept(this);
  	}
  	if(DeclarationPreprocessor.applyModularReasoning(decl.spec))
  	{
  		for (FunctionSpec s : decl.spec) s.accept(this);  
  	}
  	else
  	{
  		isFunctionDefinition=true;
  		decl.exp.accept(this);
  	}
    return null;
  }
  public Void visit(PredicateDeclaration decl)
  {
  	sym=(FunctionSymbol)decl.ident.getSymbol();
  	for(Parameter p:decl.param)
  	{
  		p.accept(this);
  	}
  	return null;
  }
  public Void visit(PredicateDefinition decl)
  {
  	sym=(FunctionSymbol)decl.ident.getSymbol();
  	for(Parameter p:decl.param)
  	{
  		p.accept(this);
  	}
  	if(DeclarationPreprocessor.applyModularReasoning(decl.spec))
  	{
  		for (FunctionSpec s : decl.spec) s.accept(this);  
  	}
  	else
  	{
  		decl.exp.accept(this);
  	}
    return null;
  }
  public Void visit(TheoremParamDefinition decl)
  {
  	sym=decl.ident.getSymbol();
  	for(Parameter p:decl.param)
  	{
  		p.accept(this);
  	}
    for (FunctionSpec s : decl.spec) s.accept(this);    
    decl.exp.accept(this);
    return null;
  }
  public Void visit(ProcedureDeclaration decl)
  {
  	sym=(FunctionSymbol)decl.ident.getSymbol();
  	decl.type.accept(this);
  	for(Parameter p:decl.param)
  	{
  		p.accept(this);
  	}
  	return null;
  }
  public Void visit(ProcedureDefinition decl)
  {
  	sym=(FunctionSymbol)decl.ident.getSymbol();
  	decl.type.accept(this);
  	for(Parameter p:decl.param)
  	{
  		p.accept(this);
  	}
  	for (FunctionSpec s : decl.spec) s.accept(this);  
    return null;
  }

  public Void visit(TheoremDefinition decl)
  {
  	sym=decl.ident.getSymbol();
    decl.exp.accept(this);
    return null;
  }
  public Void visit(PredicateValueDefinition decl)
  { 	
  	sym=decl.ident.getSymbol();
    decl.exp.accept(this);
    return null;
  }

  public Void visit(TypeDefinition decl)
  {
  	sym=decl.ident.getSymbol();
  	decl.type.accept(this);
    if (decl.exp != null) decl.exp.accept(this);
    return null;
  }

  public Void visit(ValueDefinition decl)
  {
  	sym=decl.ident.getSymbol();
  	if(decl.type!=null)
  		decl.type.accept(this);
    decl.exp.accept(this);
    return null;
  }
		
	public Void visit(ApplicationExp exp)
	{
		occuringFunctionSymbols.add(exp.ident.getSymbol());
		for(Expression e:exp.exps)
		{
			e.accept(this);
		}
		return null;
	}
	
	public Void visit(IdentifierExp exp)
	{
		occuringValueSymbols.add(exp.ident.getSymbol());
		return null;
	}
	
	public Void visit(ArrayType type)
  {
    if (type.isCanonical()) return null;
    type.exp.accept(this);
    type.type.accept(this);
    return null;
  }

  public Void visit(IntType type)
  {
    if (type.isCanonical()) return null;
    type.exp1.accept(this);
    type.exp2.accept(this);
    return null;
  }
  public Void visit(MapType type)
  {
    if (type.isCanonical()) return null;
    type.type1.accept(this);
    type.type2.accept(this);
    return null;
  }     
  public Void visit(NatType type)
  {
    if (type.isCanonical()) return null;
    type.exp.accept(this);
    return null;
  }
  public Void visit(RecordType type)
  {
    if (type.isCanonical()) return null;
    for (Parameter p : type.param) p.accept(this);
    return null;
  }

  public Void visit(SetType type)
  {
    if (type.isCanonical()) return null;
    type.type.accept(this);
    return null;
  }    
  public Void visit(SubType type)
  {
    if (type.isCanonical()) return null;
//    type.ident.accept(this);
    type.type.accept(this);
    // do not visit predicate symbol
    return null;
  } 
  public Void visit(TupleType type)
  {
    if (type.isCanonical()) return null;
    for (Type t : type.types) t.accept(this);
    return null;
  }   
  public Void visit(UnitType type)
  {
    return null;
  }
  public Void visit(Parameter p)
  {
  	p.type.accept(this);
  	return null;
  }
}
//---------------------------------------------------------------------------
// end of file
//---------------------------------------------------------------------------
