// ---------------------------------------------------------------------------
// ExpressionTranslator.java
// A class which is dedicated to translate expressions to SMT-LIB strings.
//
// Author: Franz Reichl <franz.x.reichl@gmail.com>
// $Id: ExpressionTranslator.java,v 1.9 2021/03/31 06:38:45 schreine Exp $
// 
// Author: Franz Reichl <franz.x.reichl@gmail.com>
// Copyright (C) 2019-, Research Institute for Symbolic Computation (RISC)
// Johannes Kepler University, Linz, Austria, http://www.risc.jku.at
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
// ---------------------------------------------------------------------------
package riscal.smt;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.ListIterator;
import java.util.Map;
import java.util.Objects;
import java.util.Set;
import java.util.StringJoiner;
import java.util.stream.Collectors;

import riscal.semantics.Types.Value;
import riscal.semantics.Values;
import riscal.smt.SMT.TooLargeTypeSize;
import riscal.syntax.AST.Binder;
import riscal.syntax.AST.Expression;
import riscal.syntax.AST.Expression.AndExp;
import riscal.syntax.AST.Expression.ApplicationExp;
import riscal.syntax.AST.Expression.AssertExp;
import riscal.syntax.AST.Expression.BigIntersectExp;
import riscal.syntax.AST.Expression.BigUnionExp;
import riscal.syntax.AST.Expression.CartesianExp;
import riscal.syntax.AST.Expression.CheckExp;
import riscal.syntax.AST.Expression.ChooseExp;
import riscal.syntax.AST.Expression.ChooseInElseExp;
import riscal.syntax.AST.Expression.ChooseInExp;
import riscal.syntax.AST.Expression.DividesExp;
import riscal.syntax.AST.Expression.EmptySetExp;
import riscal.syntax.AST.Expression.EnumeratedSetExp;
import riscal.syntax.AST.Expression.EqualsExp;
import riscal.syntax.AST.Expression.EquivExp;
import riscal.syntax.AST.Expression.ExistsExp;
import riscal.syntax.AST.Expression.FactorialExp;
import riscal.syntax.AST.Expression.FalseExp;
import riscal.syntax.AST.Expression.ForallExp;
import riscal.syntax.AST.Expression.GreaterEqualExp;
import riscal.syntax.AST.Expression.GreaterExp;
import riscal.syntax.AST.Expression.HeightExp;
import riscal.syntax.AST.Expression.IdentifierExp;
import riscal.syntax.AST.Expression.IfThenElseExp;
import riscal.syntax.AST.Expression.ImpliesExp;
import riscal.syntax.AST.Expression.InSetExp;
import riscal.syntax.AST.Expression.IntersectExp;
import riscal.syntax.AST.Expression.IntervalExp;
import riscal.syntax.AST.Expression.LessEqualExp;
import riscal.syntax.AST.Expression.LessExp;
import riscal.syntax.AST.Expression.LetExp;
import riscal.syntax.AST.Expression.LetParExp;
import riscal.syntax.AST.Expression.MapBuilderExp;
import riscal.syntax.AST.Expression.MapSelectionExp;
import riscal.syntax.AST.Expression.MapUpdateExp;
import riscal.syntax.AST.Expression.MatchExp;
import riscal.syntax.AST.Expression.MaxExp;
import riscal.syntax.AST.Expression.MinExp;
import riscal.syntax.AST.Expression.MinusExp;
import riscal.syntax.AST.Expression.NegationExp;
import riscal.syntax.AST.Expression.NotEqualsExp;
import riscal.syntax.AST.Expression.NotExp;
import riscal.syntax.AST.Expression.NumberExp;
import riscal.syntax.AST.Expression.NumberLiteralExp;
import riscal.syntax.AST.Expression.OrExp;
import riscal.syntax.AST.Expression.PlusExp;
import riscal.syntax.AST.Expression.PlusExpMult;
import riscal.syntax.AST.Expression.PowerExp;
import riscal.syntax.AST.Expression.Powerset1Exp;
import riscal.syntax.AST.Expression.Powerset2Exp;
import riscal.syntax.AST.Expression.PowersetExp;
import riscal.syntax.AST.Expression.PrintExp;
import riscal.syntax.AST.Expression.PrintInExp;
import riscal.syntax.AST.Expression.PrintTraceInExp;
import riscal.syntax.AST.Expression.ProductExp;
import riscal.syntax.AST.Expression.RecApplicationExp;
import riscal.syntax.AST.Expression.RecIdentifierExp;
import riscal.syntax.AST.Expression.RecordExp;
import riscal.syntax.AST.Expression.RecordSelectionExp;
import riscal.syntax.AST.Expression.RecordUpdateExp;
import riscal.syntax.AST.Expression.RemainderExp;
import riscal.syntax.AST.Expression.SetBuilderExp;
import riscal.syntax.AST.Expression.SetSizeExp;
import riscal.syntax.AST.Expression.SubsetExp;
import riscal.syntax.AST.Expression.SumExp;
import riscal.syntax.AST.Expression.TimesExp;
import riscal.syntax.AST.Expression.TimesExpMult;
import riscal.syntax.AST.Expression.TrueExp;
import riscal.syntax.AST.Expression.TupleExp;
import riscal.syntax.AST.Expression.TupleSelectionExp;
import riscal.syntax.AST.Expression.TupleUpdateExp;
import riscal.syntax.AST.Expression.UnionExp;
import riscal.syntax.AST.Expression.UnitExp;
import riscal.syntax.AST.Expression.WithoutExp;
import riscal.syntax.AST.Identifier;
import riscal.syntax.AST.QuantifiedVariable;
import riscal.syntax.AST.QuantifiedVariableCore.IdentifierSetQuantifiedVar;
import riscal.syntax.AST.QuantifiedVariableCore.IdentifierTypeQuantifiedVar;
import riscal.syntax.AST.Type;
import riscal.syntax.AST.Type.BoolType;
import riscal.syntax.AST.Type.IdentifierType;
import riscal.syntax.AST.Type.IntType;
import riscal.syntax.AST.Type.MapType;
import riscal.syntax.AST.Type.SetType;
import riscal.syntax.AST.Type.SubType;
import riscal.syntax.AST.Type.TupleType;
import riscal.syntax.AST.Type.UnitType;
import riscal.syntax.ASTVisitor;
import riscal.types.Environment.Symbol.FunctionSymbol;
import riscal.types.Environment.Symbol.TypeSymbol;
import riscal.types.Environment.Symbol.ValueSymbol;
import riscal.util.Seq;

public class ExpressionTranslator extends ASTVisitor.Base<String>
{
	/**
	 * The underlying DeclarationTranslator
	 */
	private DeclarationTranslator decTrans;
	
	private boolean unquantify;
	boolean getUnquantify()
	{
		return unquantify;
	}
	
	/*
	 * The translators for expressions involving the respective types
	 */
	private BoolTranslator boolTrans;
	private IntTranslator intTrans;
	private TupleTranslator tupleTrans;
	private MapTranslator mapTrans;
	private SetTranslator setTrans;
	
	/**
	 * The name of the function that maps boolean values to bit vectors.
	 * null if yet not needed
	 */
	private String boolean2BitVector=null;
	/**
	 * The name of the function that maps bit vectors to boolean values.
	 * null if yet not needed
	 */
	private String bitVector2boolean=null;
	
	/**
	 * Assigns to the respective type all the information about
	 * the enumeration of the type.
	 */
	private Map<Type,TypeEnumerationInformation> typeEnumerations=new HashMap<Type, TypeEnumerationInformation>();
	
	/**
	 * Used to determine whether auxiliary functions in quantifier bodies pay off.
	 * This number determines the maximal length of a quantifier body for which
	 * no auxiliary functions are generated (if AuxilaryQuantfierFunctionUsage is
	 * set to EnabledRespectSize)
	 */
	public static final long maximalQuantifiedExpressionLengthInQVExpansion=2500;
	/**
	 * The basis for the generation of the name of auxiliary functions in 
	 * quantifier bodies.
	 */
	private static final String defaultAuxillaryExpansionFunctionName="qvExpand";
	/**
	 * The basis for the generation of the name of the enumeration of the UnitType.
	 */
	private static final String defaultUnitTypeEnumerationName="unitEnum";
	
	/**
	 * Contains the information needed for enumerating the bit vector representations
	 * of the values of a RISCAL type.
	 */
	public static class TypeEnumerationInformation
	{
		public String functionName;
		public int enumerationFunctionResultVectorLength;
		public long numberOfElements;
		public TypeEnumerationInformation(String name,int len,long num)
		{
			this.functionName=name;
			this.enumerationFunctionResultVectorLength=len;
			this.numberOfElements=num;
		}		
	}
	
	/**
	 * Adds an additionally generated declaration (SMT-LIB)
	 * to the underlying DeclarationTranslator.
	 * @param str The SMT-LIB declaration.
	 */
	void addDeclaration(String str)
	{
		decTrans.addDeclaration(str);
	}
	
	void addDeclarationRemovable(String str)
	{
		decTrans.addDeclarationRemovable(str);
	} 
	
	/***************************************************************************
	 * Retrieves the symbols of all the values from the currently processed
	 * RISCAL declarations.
	 * @return The values from the underlying DeclarationTranslator.
	 **************************************************************************/
	Set<ValueSymbol> getVals()
	{
		return decTrans.getVals();
	}
	
	/***************************************************************************
	 * Retrieves the name of the name of the function that maps boolean values
	 * to bit vectors. If this function was not yet used, it is generated
	 * and the respective SMT-LIB function definition is added to the
	 * underlying DeclarationTranslator.
	 * @return The name of the function that maps boolean values to bit vectors.
	 **************************************************************************/
	String getBoolean2BitVector()
	{
		if(boolean2BitVector==null)
		{
			boolean2BitVector=getFunctionName("boolean2BitVector");
			String dec="(define-fun "+boolean2BitVector+"((x Bool)) (_ BitVec 1) (ite x #b1 #b0))";
			addDeclaration(dec);
		}
			
		return boolean2BitVector;		
	}
	
	/***************************************************************************
	 * Retrieves the name of the name of the function that maps bit vectors
	 * to boolean values. If this function was not yet used, it is generated
	 * and the respective SMT-LIB function definition is added to the
	 * underlying DeclarationTranslator.
	 * @return The name of the function that maps bit vectors to boolean values.
	 **************************************************************************/
	String getBitVector2boolean()
	{
		if(bitVector2boolean==null)
		{
			bitVector2boolean=getFunctionName("bitVector2boolean");
			String dec="(define-fun "+bitVector2boolean+"((x (_ BitVec 1))) Bool (ite (= x #b1) true false))";
			addDeclaration(dec);
		}
		return bitVector2boolean;
	}	
	
	/***************************************************************************
	 * Create an ExpressionTranslator with the given underlying DeclarationTranslator.
	 * @param d The underlying DeclarationTranslator.
	 * Condition: d!=null
	 **************************************************************************/
	public ExpressionTranslator(DeclarationTranslator d)
	{
		super();
		decTrans=Objects.requireNonNull(d, "An expression translator can only be generated by a declaration translator");
		this.unquantify=d.getUnquantify();
		this.boolTrans=new BoolTranslator(this);
		this.intTrans=new IntTranslator(this);
		this.tupleTrans=new TupleTranslator(this);
		this.mapTrans=new MapTranslator(this);
		this.setTrans=new SetTranslator(this);
	}
	
	/***************************************************************************
	 * Translate the given expression to SMT-LIB.
	 * @param exp The expression that shall be translated.
	 * @return The result of the translation.
	 **************************************************************************/
	public String compute(Expression exp)
	{
		return exp.accept(this);
	}
	
	/***************************************************************************
	 * Translate the given expression to SMT-LIB.
	 * The sort of the result is the SMT-LIB sort that corresponds
	 * to the given type.
	 * @param exp The expression that shall be translated.
	 * @param type Determines the sort the result shall have.
	 * @return The result of the translation.
	 **************************************************************************/
	public String compute(Expression exp, Type type)
	{
		type=Utilities.getBaseType(type);
		if(type instanceof IntType)
		{
			IntType iT=(IntType)type;
			return intTrans.compute(exp, iT.ivalue1, iT.ivalue2);
		}
		else if(type instanceof TupleType)
		{
			TupleType tT=(TupleType)type;
			return tupleTrans.compute(exp,tT);
		}
		else if(type instanceof MapType)
		{
			MapType mT=(MapType)type;
			return mapTrans.compute(exp, mT);
		}
		else if(type instanceof SetType)
		{
			SetType sT=(SetType)type;
			return setTrans.compute(exp, sT);
		}
		return exp.accept(this);
	}
	
	/***************************************************************************
	 * Translate the given expression to SMT-LIB.
	 * If the expression is a boolean expression
	 * than skolemize determines whether skolemisation
	 * shall be applied. 
	 * Note:The other compute methods do not makes use of
	 * skolemisation.
	 * @param exp The expression that shall be translated.
	 * @param skolemize If true skolemisation shall be applied.
	 * @return The result of the translation.
	 **************************************************************************/
	public String compute(Expression exp, boolean skolemize)
	{
		if(exp.getType() instanceof BoolType)
		{
			return boolTrans.compute(exp, skolemize);
		}
		return exp.accept(this);
	}
	
	/***************************************************************************
	 * Takes two expressions that have a common super type.
	 * This method determines such a common super type and
	 * translates both expressions to SMT-LIB expressions
	 * of the sort that corresponds to this type.
	 * Condition:
	 * exp1 and exp2 have a common super type.
	 * @param exp1 A expression that shall be translated
	 * @param exp2 A expression that shall be translated
	 * @return
	 **************************************************************************/
	public Utilities.Pair<String,String> computeUseSameType(Expression exp1, Expression exp2)
	{
		Type t1=exp1.getType();
		Type t2=exp2.getType();
		Type t=determineCommonSuperType(t1, t2);
		String s1=compute(exp1, t);
		String s2=compute(exp2, t);
		return new Utilities.Pair<String, String>(s1, s2);
	}
	
	/***************************************************************************
	 * Takes a SMT-LIB expression that has the sort that corresponds
	 * to oldType. The method transforms this SMT-LIB expression to
	 * a SMT-LIB expression of the sort that corresponds to newType.
	 * Condition:
	 * oldType must be contained in newType
	 * @param s The SMT-LIB expression that shall be transformed.
	 * @param oldType The type that corresponds to the current sort.
	 * @param newType The type that corresponds to the sort the result
	 * shall have.
	 * @return A SMT-LIB expression of the sort that corresponds to
	 * newType
	 **************************************************************************/
	String adaptTranslation(String s,Type oldType,Type newType)
	{
		Type t1=Utilities.getBaseType(oldType);
		Type t2=Utilities.getBaseType(newType);
		assert(t1.getClass()==t2.getClass());
		if(t1 instanceof UnitType||t1 instanceof BoolType)
			return s;
		else if(t1 instanceof IntType)
			return IntTranslator.adaptTranslation(s, (IntType)t1, (IntType)t2);
		else if(t1 instanceof TupleType)
		{
			return tupleTrans.adaptTranslation(s, (TupleType)t1, (TupleType)t2);
		}
		else if(t1 instanceof MapType)
		{
			return mapTrans.adaptTranslation(s, (MapType)t1, (MapType)t2);
		}
		else if(t1 instanceof SetType)
		{
			return setTrans.adaptTranslation(s, (SetType)t1, (SetType)t2);
		}
		assert(false);
		return null;
	}
	
	/***************************************************************************
	 * Generates a SMT-LIB function application.
	 * Condition:
	 * The elements of params must correspond to the arity of funName.
	 * @param funName The name of the function.
	 * @param params The arguments of the function.
	 * @return The application of funName to params.
	 **************************************************************************/
	public String generateSMTLIBFunctionApplication(String funName,List<String> params)
	{
		if(params.size()==0)
		{
			return funName;
		}
		String str=String.join(" ",params);
		return "("+funName+" "+str+")";			
	}
	
	/***************************************************************************
	 * Universally quantifies the given SMT-LIB string with respect to variables with
	 * names given by paramNames and sorts corresponding to the types given by
	 * paramTypes. The method either returns the expansion of the quantifier or
	 * a SMT-LIB universal quantifier, depending on SMT.getUnquantify().
	 * Note:
	 * The SMT-LIB quantifier should not be used.
	 * Condition:
	 * paramNames.size()==paramTypes.size()
	 * @param quantifiedExpression The SMT-LIB expression that shall be quantified.
	 * @param paramNames The names of the variables.
	 * @param paramTypes The types that determine the sorts of the variables
	 * @return Either an expansion of the respective universal quantifier
	 * or a universally quantified SMT-LIB expression.
	 **************************************************************************/
	public String generateQuantifiedSMTLibExpression(String quantifiedExpression,List<String>paramNames,List<Type>paramTypes)
	{
		assert(paramNames.size()==paramTypes.size());
		String str=quantifiedExpression;
		ListIterator<String> nameIt=paramNames.listIterator(paramNames.size());
		ListIterator<Type> typeIt=paramTypes.listIterator(paramTypes.size());
		int ind=paramNames.size()-1;
		if(getUnquantify())
		{
			while(nameIt.hasPrevious()&&typeIt.hasPrevious())
			{
				Type ty=typeIt.previous();
				String name=nameIt.previous();
				if(ExpressionTranslator.useAuxillaryFunctionInExpansions(getTypeSize(ty), str.length()))
				{
					if(SMT.plugValuesDirectlyIntoAuxiliaryFunctions)
					{
						List<String> lst=generateAuxillaryExpansionFunction(str, paramNames.subList(0, ind+1), 
								paramTypes.subList(0, ind+1), "Bool", ind);
						str=BoolTranslator.conjoin(lst);
						ind--;
						continue;
					}
					else
					{
						str=generateAuxillaryExpansionFunction(str, paramNames.subList(0, ind+1),
								paramTypes.subList(0, ind+1),"Bool");
					}					
				}
				ind--;
				str=boolTrans.constructForallExpansion(str, name, ty);
			}
		}
		else
		{
			while(nameIt.hasPrevious()&&typeIt.hasPrevious())
			{
				str=boolTrans.constructSMTLibForall(str, nameIt.previous(), typeIt.previous());
			}
		}
		return str;
	}
	
	/***************************************************************************
	 * Determines the SMT-LIB sort that corresponds to the given type.
	 * @param ty The type of interest.
	 * @return The corresponding SMT-LIB sort.
	 **************************************************************************/
	public String translateType(Type ty)
	{
		return ty.accept(this);
	}
	
	/**
	 * Factory for creating ValueContainer
	 */
	static class ValueContainerGenerator
	{
		static ValueContainer getValueContainer(ExpressionTranslator eT,Type type)
		{
			if(type instanceof SubType)
				type=Utilities.getBaseType(type);
			if(type instanceof IntType)
			{
				return new IntTranslator.IntValueContainer((IntType)type);
			}
			else if(type instanceof BoolType)
			{
				return new BoolTranslator.BoolValueContainer((BoolType)type);
			}
			else if(type instanceof UnitType)
			{
				return new UnitValueContainer();
			}
			else if(type instanceof TupleType)
			{
				return new TupleTranslator.TupleValueContainer(eT,eT.tupleTrans,(TupleType)type);
			}
			else if(type instanceof MapType)
			{
				return new MapTranslator.MapValueContainer(eT, eT.mapTrans, (MapType)type);
			}
			else if(type instanceof SetType)
			{
				return new SetTranslator.SetValueContainer(eT, (SetType)type);
			}
			
			assert(false);
			return null;
		}		
	}
	
	/***************************************************************************
	 * We want to provide a ValueContainer for all types (non recursive)
	 * except the UnitType.
	 * These ValueContainers shall provide the functionalities to translate a
	 * given value corresponding to the respective type into SMT-LIB.
	 * The idea is that for the translation of values of certain types
	 * auxiliary computations have to be done. (e.g. the the necessary bit vector length
	 * for integers) A ValueContainer shall provide the result of these
	 * auxiliary computations such that these computations only have to be conducted once.
	 **************************************************************************/
	static abstract class ValueContainer
	{
		/***************************************************************************
		 * Generate a bit vector constant representing the given value.
		 * @param val The value that shall be represented by a bit vector.
		 * @return The SMT-LIB representation of val
		 **************************************************************************/
		abstract String process(Value val);
		/***************************************************************************
		 * Generate a bit vector constant representing the given value without
		 * the #b prefix.
		 * @param val The value that shall be represented by a bit vector.
		 * @return The SMT-LIB representation of val
		 **************************************************************************/
		abstract String processToBareVector(Value val);
	}
	
	/**
	 * ValueContainer for the UnitType
	 */
	private static class UnitValueContainer extends ValueContainer
	{
		/***************************************************************************
		 * Generate a bit vector constant representing the given value.
		 * @param val The value that shall be represented by a bit vector.
		 * @return The SMT-LIB representation of val
		 **************************************************************************/
		String process(Value val)
		{
			return DeclarationTranslator.unit;
		}
		/***************************************************************************
		 * Generate a bit vector constant representing the given value without
		 * the #b prefix.
		 * @param val The value that shall be represented by a bit vector.
		 * @return The SMT-LIB representation of val
		 **************************************************************************/
		String processToBareVector(Value val)
		{
			return DeclarationTranslator.unitBareVector;
		}		
	}
	
	/***************************************************************************
	 * Compute the SMT-LIB representations of the values of the given type.
	 * Do not add the #b prefix to these bit vector representations.
	 * @param type The type whose values shall be computed
	 * @return The bit vector representations of the values of type,
	 * given in the order of Values.getValueSeq(type)
	 **************************************************************************/
	public List<String> computeValuesBare(Type type)
	{
		return computeValues(type,true);
	}
	
	/***************************************************************************
	 * Compute the SMT-LIB representations of the values of the given type.
	 * @param type The type whose values shall be computed
	 * @return If type is neither the BoolType nor a SubType of BoolType
	 * the bit vector representations of the values of type,
	 * given in the order of Values.getValueSeq(type). 
	 * Otherwise the respective boolean values from SMT-LIB
	 * given in the order of Values.getValueSeq(type).
	 **************************************************************************/
	public List<String> computeValues(Type type)
	{
		return computeValues(type,false);
	}
	
	/***************************************************************************
	 * Compute the SMT-LIB representations of the values of the given type.
	 * Add the #b prefix to these representations depending on bare.
	 * @param type The type whose values shall be computed
	 * @param bare If true do not add the #b prefix to the values
	 * @return If type is neither the BoolType nor a SubType of BoolType
	 * or bare is true the bit vector representations of the values of type,
	 * given in the order of Values.getValueSeq(type). 
	 * Otherwise the respective boolean values from SMT-LIB
	 * given in the order of Values.getValueSeq(type).
	 **************************************************************************/
	public List<String> computeValues(Type type,boolean bare)
	{
		if(type instanceof IdentifierType)
		{
			type=((TypeSymbol)((IdentifierType)type).ident.getSymbol()).getType();
		}	
		Type type0=Utilities.getBaseType(type);
		if(SMT.expandBaseTypesOfSubTypes)
		{
			type=type0;
		}
		if(type0 instanceof UnitType)//cannot have a proper subtype
		{
			List<String> vals=new ArrayList<String>(1);
			if(bare)
				vals.add(DeclarationTranslator.unitBareVector);
			else
				vals.add(DeclarationTranslator.unit);
			return vals;
		}
		Seq<Value> seq=null;
		boolean isSubType=false;
		if(type instanceof SubType)
		{
			isSubType=true;
			seq=Values.getValueSeq(type);
		}		
		if(type0 instanceof BoolType)
		{
			if(isSubType)
				return BoolTranslator.processValueSequence(type0,seq,bare);
			else
				return BoolTranslator.computeValues(bare);
		}
		else if(type0 instanceof IntType)
		{
			if(isSubType)
				return IntTranslator.processValueSequence(type0, seq,bare);
			else
				return IntTranslator.computeValues((IntType)type,bare);
		}
		else if(type0 instanceof TupleType)
		{
			if(isSubType)
				return tupleTrans.processValueSequence(type0, seq,bare);
			else
				return tupleTrans.computeValues((TupleType)type,bare);
		}
		else if(type0 instanceof MapType)
		{
			if(isSubType)
				return mapTrans.processValueSequence(type0, seq,bare);
			else
				return mapTrans.computeValues((MapType)type,bare);
		}
		else if(type0 instanceof SetType)
		{
			if(isSubType)
				return setTrans.processValueSequence(type0, seq,bare);
			else
				return SetTranslator.computeValues((SetType)type,bare);
		}
		else
		{
			SMT.notImplemented(type, "Values of the specified type cannot be generated!");
			return null;
		}		
	}
	
	/***************************************************************************
	 * Determines whether auxiliary functions shall applied.
	 * @param numberOfValues The number of values of the type of the respective quantification.
	 * @param stringLength The length of the SMT-LIB representation of the quantified
	 * expression.
	 * @return true if auxiliary functions shall applied.
	 **************************************************************************/
	static boolean useAuxillaryFunctionInExpansions(int numberOfValues,int stringLength)
	{
		switch(SMT.quantifierAuxFunctionMode)
		{
		case Disabled:
			return false;
		case EnabledRespectSize:
			if(((long)numberOfValues)*stringLength>ExpressionTranslator.maximalQuantifiedExpressionLengthInQVExpansion)
				return true;
			else
				return false;
		case EnabledAlwaysUse:
			return true;
		}
		assert(false);
		return false;
	}
	
	/***************************************************************************
	 * Generate a SMT-LIB function definition.
	 * @param funName The name of the function that shall be defined.
	 * @param paramNames The names of the parameters.
	 * @param paramTypes The types of the corresponding RISCAL function. 
	 * Shall be canonical types.
	 * @param typeStr The result type of the corresponding RISCAL function.
	 * @param definingStr The defining SMT-LIB expression.
	 * @return A SMT-LIB function definition command.
	 **************************************************************************/
	private String generateSMT_LIBFunctionDefinition(String funName,List<String>paramNames,
			List<Type>paramTypes,String typeStr,String definingStr)
	{
		assert(paramNames.size()==paramTypes.size());
		Iterator<String> it1=paramNames.iterator();
		Iterator<Type> it2=paramTypes.iterator();
		StringBuilder strB=new StringBuilder();
		while(it1.hasNext()&&it2.hasNext())
		{
			strB.append("("+it1.next()+" "+translateType(Utilities.getBaseType(it2.next()))+")");
		}
		String params=strB.toString();
		String funStr="(define-fun "+funName+" ("+params+") "+typeStr+" "+definingStr+")";
		return funStr;
	}
	
	/***************************************************************************
	 * Define an auxiliary function for a quantifier body and return its application.
	 * @param definingStr The SMT-LIB representation of the quantified expression.
	 * @param freeSyms The symbols that correspond to the free variables in the quantified
	 * expression.
	 * @param typeStr The SMT-LIB sort of the result of the auxiliary function.
	 * @return The application of the auxiliary function to variables whose names are
	 * given by the identifier strings of the symbols.
	 **************************************************************************/
	String generateAuxillaryExpansionFunction(String definingStr,Set<ValueSymbol> freeSyms,String typeStr)
	{
		return generateAuxillaryExpansionFunction(definingStr,Utilities.sortSymbols(freeSyms),typeStr);
	}
	
	/***************************************************************************
	 * Define an auxiliary function for a quantifier body and return its application.
	 * @param definingStr The SMT-LIB representation of the quantified expression.
	 * @param free The symbols that correspond to the free variables in the quantified
	 * expression.
	 * The symbols in this list should be sorted with respect to Utilities.sortSymbolSet.
	 * @param typeStr The SMT-LIB sort of the result of the auxiliary function.
	 * @return The application of the auxiliary function to variables whose names are
	 * given by the identifier strings of the symbols.
	 **************************************************************************/
	private String generateAuxillaryExpansionFunction(String definingStr,List<ValueSymbol> free,String typeStr)
	{
		List<String> names=new ArrayList<String>();
		List<Type>types=new ArrayList<Type>();
		for(ValueSymbol val:free)
		{
			names.add(val.ident.string);
			types.add(val.type);
		}
		return generateAuxillaryExpansionFunction(definingStr, names, types,typeStr);
	}
	
	/***************************************************************************
	 * Define an auxiliary function for a quantifier body and return its application.
	 * Condition
	 * paramNames.size()==paramTypes.size()
	 * @param definingStr The SMT-LIB representation of the quantified expression.
	 * @param paramNames The names of the free variables part of the quantified expression.
	 * @param paramTypes The types of the free variables. Shall be canonical types.
	 * @param typeStr The SMT-LIB sort of the result of the auxiliary function.
	 * @return The application of the auxiliary function to variables whose names are
	 * given by paramNames.
	 **************************************************************************/
	String generateAuxillaryExpansionFunction(String definingStr,List<String>paramNames,List<Type>paramTypes,String typeStr)
	{
		assert(paramNames.size()==paramTypes.size());
		String funName=getFunctionName(defaultAuxillaryExpansionFunctionName);
		String funStr=generateSMT_LIBFunctionDefinition(funName,paramNames,paramTypes,typeStr,definingStr);
		addDeclarationRemovable(funStr);
		if(paramNames.size()==0||paramTypes.size()==0)
		{
			return funName;
		}
		else
		{
			String applicationParams=paramNames.stream().collect(Collectors.joining(" "));
			return "("+funName+" "+applicationParams+")";
		}
	}
	
	/***************************************************************************
	 * Define an auxiliary function for a quantifier body.
	 * Generate for each value of the type of quantifiedVariable a function application
	 * of the auxiliary function where the bit vector representation of the respective type
	 * is plugged in.
	 * @param definingStr The SMT-LIB representation of the quantified expression.
	 * @param freeSyms The symbols that correspond to the free variables in the quantified
	 * expression.
	 * @param quantifiedVariable The symbol of variable we are currently dealing with.
	 * @param typeStr The SMT-LIB sort of the result of the auxiliary function.
	 * @return A list containing applications of the generated auxiliary function
	 * where in each element the representation of a different value of the type
	 * of quantifiedVariable is applied to the auxiliary function.
	 **************************************************************************/
	List<String> generateAuxillaryExpansionFunction(String definingStr,Set<ValueSymbol> freeSyms,
			ValueSymbol quantifiedVariable,String typeStr)
	{
		return generateAuxillaryExpansionFunction(definingStr,Utilities.sortSymbols(freeSyms),quantifiedVariable,typeStr);
	}
	
	/***************************************************************************
	 * Define an auxiliary function for a quantifier body.
	 * Generate for each value of the type of quantifiedVariable a function application
	 * of the auxiliary function where the bit vector representation of the respective type
	 * is plugged in.
	 * @param definingStr The SMT-LIB representation of the quantified expression.
	 * @param free The symbols that correspond to the free variables in the quantified
	 * expression.
	 * @param quantifiedVariable The symbol of variable we are currently dealing with.
	 * @param typeStr The SMT-LIB sort of the result of the auxiliary function.
	 * @return A list containing applications of the generated auxiliary function
	 * where in each element the representation of a different value of the type
	 * of quantifiedVariable is applied to the auxiliary function.
	 **************************************************************************/
	private List<String> generateAuxillaryExpansionFunction(String definingStr,List<ValueSymbol> free,
			ValueSymbol quantifiedVariable,String typeStr)
	{
		List<String> names=new ArrayList<String>();
		List<Type>types=new ArrayList<Type>();
		int index=-1;
		int aux=0;
		for(ValueSymbol val:free)
		{
			if(val.equals(quantifiedVariable))
				index=aux;
			names.add(val.ident.string);
			types.add(val.type);
			aux++;
		}
		return generateAuxillaryExpansionFunction(definingStr, names, types,typeStr,index);
	}
	
	/***************************************************************************
	 * Define an auxiliary function for a quantifier body.
	 * If "index" is a valid index of paramTypes then
	 * generate for the bit vector representation of each value
	 * of paramTypes.get(index) a function application
	 * of the auxiliary function where the value is plugged in.
	 * If "index" is not a valid index of paramTypes then the resulting
	 * list contains only a single element representing the application of 
	 * the auxiliary function to variables given by paramNames.
	 * Condition:
	 * paramNames.size()==paramTypes.size()
	 * @param definingStr The SMT-LIB representation of the quantified expression.
	 * @param paramNames The names of the free variables part of the quantified expression.
	 * @param paramTypes The types of the free variables. Shall be canonical types.
	 * @param typeStr The SMT-LIB sort of the result of the auxiliary function.
	 * @param index If the bit vector representations of the values of
	 * a type from paramTypes shall be applied to the application of the
	 * auxiliary function, then "index" must be the index of this type.
	 * Otherwise "index" must be an invalid index of paramTypes.
	 * @return A list containing applications of the generated auxiliary function.
	 **************************************************************************/
	List<String> generateAuxillaryExpansionFunction(String definingStr,List<String>paramNames,
			List<Type>paramTypes,String typeStr,int index)
	{
		assert(paramNames.size()==paramTypes.size());
		String funName=getFunctionName(defaultAuxillaryExpansionFunctionName);
		String funStr=generateSMT_LIBFunctionDefinition(funName,paramNames,paramTypes,typeStr,definingStr);
		addDeclarationRemovable(funStr);
		if(paramNames.size()==0||paramTypes.size()==0)
		{
			List<String> lst=new ArrayList<String>(1);
			lst.add(funName);
			return lst;
		}
		else if(index<0||index>=paramNames.size())
		{
			String applicationParams=paramNames.stream().collect(Collectors.joining(" "));
			List<String> lst=new ArrayList<String>(1);
			lst.add("("+funName+" "+applicationParams+")");
			return lst;
		}
		else
		{
			List<String> values=computeValues(paramTypes.get(index));
			String applicationParams1=index==0?"":paramNames.subList(0, index).stream().collect(Collectors.joining(" "))+" ";
			String applicationParams2=index==(paramNames.size()-1)?"":" "+paramNames
					.subList(index+1, paramNames.size()).stream().collect(Collectors.joining(" "));			
			return values.stream().map(v ->{
				return "("+funName+" "+applicationParams1+v+applicationParams2+")";
			}).collect(Collectors.toList());
		}
	}
	
	/***************************************************************************
	 * Compute the ValueSymbol that correspond to unbound identifier in an expression.
	 * Remove the ValueSymbol that correspond to values introduced by the current
	 * declarations.
	 * @param exp The expression that shall be analysed.
	 * @return The ValueSymbol of the free variables.
	 **************************************************************************/
	Set<ValueSymbol> typedFreeSymbolsWithoutValues(Expression exp)
	{
		Set<ValueSymbol> freeVariables=TypedFreeVariables.compute(exp);
		freeVariables.removeAll(getVals());
		return freeVariables;
	}
	
	/***************************************************************************
	 * Determines the number of elements of a type.
	 * @exception TooLargeTypeSize If the given type has more elements than 
	 * SMT.maximalTypeSize.
	 * @param type The type whose size shall be computed. Shall be a canonical type.
	 * @return The size of the type.
	 **************************************************************************/
	public static int getTypeSize(Type type)
	{
		long s=type.getSize();
		if(s>SMT.maximalTypeSize)
		{
			SMT.tooLargeType("Too long type size", s, type);
		}
		return (int)s;
	}
	
	/***************************************************************************
	 * Determines the number of elements of a type.
	 * @exception TooLargeTypeSize If the given type has more elements 
	 * than SMT.maximalTypeSize minus buffer.
	 * @param type The type whose size shall be computed. Shall be a canonical type.
	 * @param buffer The given type has to be smaller than
	 * SMT.maximalTypeSize - buffer.
	 * @return The size of the type.
	 **************************************************************************/
	public static int getTypeSizeConsiderBuffer(Type type,int buffer)
	{
		long s=type.getSize();
		if(s>SMT.maximalTypeSize-buffer)
		{
			SMT.tooLargeType("The type is too large to ensure necessary buffer", s, type);
		}
		return (int)s;
	}
	
	/***************************************************************************
	 * Retrieve the necessary information for the enumeration function
	 * of the given type.
	 * If the enumeration function for the given type was not yet used
	 * then the TypeEnumerationInformation is computed.
	 * @param type The type whose enumeration function is of interest.
	 * Shall be a canonical type.
	 * @return The TypeEnumerationInformation associated to the given type.
	 **************************************************************************/
	public TypeEnumerationInformation getTypeEnumerationFunction(Type type)
	{
		//really necessary?
		if(type instanceof IdentifierType)
		{
			type=((TypeSymbol)((IdentifierType)type).ident.getSymbol()).getType();
		}
		Type type0=Utilities.getBaseType(type);
		if(SMT.expandBaseTypesOfSubTypes)
		{
			type=type0;
		}
		TypeEnumerationInformation info=typeEnumerations.get(type);
		if(info!=null)
		{
			return info;
		}		
		if(type0 instanceof IntType)
		{
			if(type instanceof SubType)
			{
				List<Value> lst=Values.getValueSeq(type).list();
				if(getUnquantify())
					info=intTrans.generateTypeEnumerationFunctionUseTable(lst, (IntType)type0);
				else
					info=intTrans.generateTypeEnumerationFunctionUseTableWithFunctionDefinition(lst, (IntType)type0);
			}
			else
			{
				info=intTrans.generateTypeEnumerationFunctionNonSubType((IntType)type0);
			}
		}
		else if(type0 instanceof BoolType)
		{
			if(type instanceof SubType)
			{
				List<Value> lst=Values.getValueSeq(type).list();
				if(getUnquantify())
					info=boolTrans.generateTypeEnumerationFunctionUseTable(lst);
				else
					info=boolTrans.generateTypeEnumerationFunctionUseTableWithFunctionDefinition(lst);
			}
			else
			{
				info=boolTrans.generateTypeEnumerationFunction();
			}
		}
		else if(type0 instanceof TupleType)
		{
			if(type instanceof SubType||SMT.defineEnumerationForTuplesAndMapsByTables)
			{
				List<Value> lst=Values.getValueSeq(type).list();
				if(getUnquantify())
					info=tupleTrans.generateTypeEnumerationFunctionUseTable(lst, (TupleType)type0);
				else
					info=tupleTrans.generateTypeEnumerationFunctionUseTableWithFunctionDefinition(lst, (TupleType)type0);
			}
			else
			{
				info=tupleTrans.generateTypeEnumerationFunction((TupleType)type0);
			}
		}
		else if(type0 instanceof MapType)
		{
			if(type instanceof SubType||SMT.defineEnumerationForTuplesAndMapsByTables)
			{
				List<Value> lst=Values.getValueSeq(type).list();
				if(getUnquantify())
					info=mapTrans.generateTypeEnumerationFunctionUseTable(lst, (MapType)type0);
				else
					info=mapTrans.generateTypeEnumerationFunctionUseTableWithFunctionDefinition(lst, (MapType)type0);
			}
			else
			{
				info=mapTrans.generateTypeEnumerationFunction((MapType)type0);
			}
		}
		else if(type0 instanceof SetType)
		{
			if(type instanceof SubType)
			{
				List<Value> lst=Values.getValueSeq(type).list();
				if(getUnquantify())
					info=setTrans.generateTypeEnumerationFunctionUseTable(lst, (SetType)type0);
				else
					info=setTrans.generateTypeEnumerationFunctionUseTableWithFunctionDefinition(lst, (SetType)type0);
			}
			else
			{
				info=setTrans.generateTypeEnumerationFunctionNonSubType((SetType)type0);
			}
		}
		else if(type0 instanceof UnitType)
		{
			String name=getFunctionName(defaultUnitTypeEnumerationName);
			String variable="x";
			String s="(define-fun "+name+" (("+variable+"(_ BitVec 1)))(_ BitVec 1) #b0)";
			addDeclaration(s);
			info=new TypeEnumerationInformation(name, 1, 1);
		}
		else
		{
			SMT.notImplemented(type, "The specified type cannot be enumerated!");
			return null;
		}
		typeEnumerations.put(type, info);
		return info;
	}
	
	/***************************************************************************
	 * Determines a common super type for the given types.
	 * Condition: 
	 * 		- The types have to be canonical
	 * 		- There has to be a common super type for type1 and type2
	 * 			i.e. there is a type that contains all the values of
	 * 			type1 and type2.
	 * @param type1 A type of interest. Shall be a canonical type.
	 * @param type2 A type of interest. Shall be a canonical type.
	 * @return A common super type of type1 and type2.
	 **************************************************************************/
	static Type determineCommonSuperType(Type type1,Type type2)
	{
		if(type1.equals(type2))
			return type1;
		Type t1=Utilities.getBaseType(type1);
		Type t2=Utilities.getBaseType(type2);
		assert(t1.getClass()==t2.getClass());
		Type resultType;
		if(t1 instanceof UnitType)
			resultType=new UnitType();
		else if(t1 instanceof BoolType)
			resultType=new BoolType();
		else if(t1 instanceof IntType)
		{
			IntType iT1=(IntType)t1;
			IntType iT2=(IntType)t2;
			long ival1=Long.min(iT1.ivalue1, iT2.ivalue1);
			long ival2=Long.max(iT1.ivalue2, iT2.ivalue2);
			resultType=new IntType(ival1, ival2);
		}
		else if(t1 instanceof TupleType)
		{
			TupleType tT1=(TupleType)t1;
			TupleType tT2=(TupleType)t2;
			int n=tT1.types.length;
			assert(n==tT2.types.length);
			Type[] tys=new Type[n];
			for(int i=0;i<n;i++)
			{
				tys[i]=determineCommonSuperType(tT1.types[i], tT2.types[i]);
			}
			resultType=new TupleType(tys);
		}
		else if(t1 instanceof MapType)
		{
			MapType mT1=(MapType)t1;
			MapType mT2=(MapType)t2;
			resultType=new MapType(mT1.type1, determineCommonSuperType(mT1.type2, mT2.type2));
		}
		else//t1 instanceof SetType
		{
			SetType sT1=(SetType)t1;
			SetType sT2=(SetType)t2;
			resultType=new SetType(determineCommonSuperType(sT1.type, sT2.type));
		}
		resultType.setCanonical();
		return resultType;
	}
	
	/***************************************************************************
	 * Determine the bit vector length that is necessary to
	 * represent the values of the given type.
	 * @param type The type of interest. Has to be a canonical type.
	 * @return The bit vector length that is necessary to
	 * represent values from type.
	 **************************************************************************/
	public long giveVectorLength(Type type)
	{
		assert(type.isCanonical());
		//elements of a subtype are always represented by bitvectors for the baseType.
		Type type0=Utilities.getBaseType(type);
		if(type0 instanceof BoolType)
		{
			return BoolTranslator.giveVectorLength();
		}
		else if(type0 instanceof IntType)
		{
			return IntTranslator.giveVectorLength((IntType)type0);
		}
		else if(type0 instanceof TupleType)
		{
			TupleType tT=(TupleType) type0;
			return tupleTrans.giveVectorLength(tT);
		}
		else if(type0 instanceof MapType)
		{
			MapType mT=(MapType)type0;
			return mapTrans.giveVectorLength(mT);
		}
		else if(type0 instanceof SetType)
		{
			SetType sT=(SetType)type0;
			return setTrans.giveVectorLength(sT);
		}
		else if(type instanceof UnitType)
		{
			return 1L;
		}
		else
		assert(false);
		return 0;
	}
	
	/***************************************************************************
	 * Determine if the bit vector sort that is used to represent
	 * the values of t contains values that do not correspond
	 * to values of t.
	 * If the method returns false there is a bijective mapping
	 * of the given type to the associated bit vector sort.
	 * @param t The type of interest. Has to be a canonical type.
	 * @return true if the bit vector sort associated to t
	 * contains values that do not correspond to values of t.
	 **************************************************************************/
	public boolean hasTypeCondition(Type t)
	{
		if(generateTypeConditions(t, "")==null)
			return false;
		else
			return true;
	}
	
	/***************************************************************************
	 * Generate a condition that checks whether the value of a SMT-LIB
	 * expression corresponds to a value of the given RISCAL type.
	 * @param ty The type of interest. Has to be a canonical type.
	 * @param str A SMT-LIB expression. Must be of the SMT-LIB sort that is
	 * used to represent the values of ty.
	 * @return If each value of the bit vector sort that is used to represent the
	 * values of ty corresponds to a value of ty than null. Otherwise a boolean
	 * valued SMT-LIB expression that is true iff the SMT-LIB
	 * expression s corresponds to a value of the type ty.
	 **************************************************************************/
	public String generateTypeConditions(Type ty,String str)
	{
		if(ty instanceof IntType)
		{
			IntType iT=(IntType)ty;
			return IntTranslator.generateCondition(iT.ivalue1, iT.ivalue2, str);
		}
		else if(ty instanceof SubType)
		{
			String cond=decTrans.getTypeCondition(ty);
			return "("+cond+" "+str+")";
		}
		else if(ty instanceof TupleType)
		{
			return tupleTrans.generateTypeConditions((TupleType)ty, str);
		}
		else if(ty instanceof MapType)
		{
			return mapTrans.generateTypeConditions((MapType)ty, str);
		}
		else if(ty instanceof UnitType)
		{
			return "(= "+str+" "+DeclarationTranslator.unit+")";
		}
		else if(ty instanceof SetType)
		{
			if(SMT.expandBaseTypesOfSubTypes)
			{
				return setTrans.generateTypeConditions((SetType)ty, str);
			}
			else
				return null;
		}
		else if(ty instanceof BoolType)
		{
			return null;
		}
		assert(false);
		return null;
	}
	
	/***************************************************************************
	 * Generate a fresh function name.
	 * @param s A suggestion for the function name.
	 * @return An unused function name.
	 **************************************************************************/
	public String getFunctionName(String s)
	{
		return decTrans.getNewFunctionName(s);
	}
  
	public String visit(Binder b)
	{
		String str=b.exp.accept(this);
		return "( "+b.ident.string+" "+str+")";
	}

  public String visit(AndExp exp)
  {
  	return boolTrans.compute(exp,false);
  }
  public String visit(ApplicationExp exp)
  {
  	//In SMT-LIB we cannot just handle constants as functions with no parameters.
  	if(exp.exps.length==0)
  	{
  		return exp.ident.string;
  	}
  	else
  	{
  		int len=exp.exps.length;
  		Type[] types=((FunctionSymbol)exp.ident.getSymbol()).types;
  		assert(len==types.length);
  		StringJoiner strJ=new StringJoiner(" ");
  		for(int i=0;i<len;i++)
  		{
  			strJ.add(compute(exp.exps[i], types[i]));
  		}
  		String params=strJ.toString();
  		return "("+exp.ident.string+" "+params+")";
  	}
  }

  public String visit(MapBuilderExp exp)
  {
  	return mapTrans.compute(exp);
  }

  public String visit(MapSelectionExp exp)
  {
  	return mapTrans.compute(exp);
  }

  public String visit(MapUpdateExp exp)
  {
  	return mapTrans.compute(exp);
  }
  
  public String visit(AssertExp exp)
  {
  	SMT.illegalUsage(exp, "Assert expressions should not occur in the translation to SMT-LIB!");
		return null;
  }
  
  public String visit(BigUnionExp exp)
  {
  	return setTrans.compute(exp);
  }

  public String visit(BigIntersectExp exp)
  {
  	return setTrans.compute(exp);
  }
  
  public String visit(ChooseExp exp)
  {
  	SMT.illegalUsage(exp, "Choose expressions should not occur in the translation to SMT-LIB!");
		return null;
  }

  public String visit(ChooseInElseExp exp)
  {
  	SMT.illegalUsage(exp, "Choose expressions should not occur in the translation to SMT-LIB!");
		return null;
  }

  public String visit(ChooseInExp exp)
  {
  	SMT.illegalUsage(exp, "Choose expressions should not occur in the translation to SMT-LIB!");
		return null;
  }    
  public String visit(DividesExp exp)
  {
  	return intTrans.compute(exp);
  }

  public String visit(EmptySetExp exp)
  {
  	return setTrans.compute(exp);
  }

  public String visit(EnumeratedSetExp exp)
  {
  	return setTrans.compute(exp);

  }
  
  public String visit(CartesianExp exp)
  {
  	return setTrans.compute(exp);
  }
  
	public String visit(EqualsExp exp)
	{
		Type ty=exp.exp1.getType();
		ty=Utilities.getBaseType(ty);
		if(ty instanceof IntType)
			return intTrans.compute(exp);
		else if(ty instanceof TupleType)
		{
			return tupleTrans.compute(exp);
		}
		else if(ty instanceof MapType)
		{
			return mapTrans.compute(exp);
		}
		else if(ty instanceof SetType)
		{
			return setTrans.compute(exp);
		}
		else
		{
			String str1=exp.exp1.accept(this);
			String str2=exp.exp2.accept(this);
			return "(= "+str1+" "+str2+")";
		}
	}
  
  public String visit(EquivExp exp)
  {
  	SMT.illegalUsage(exp, "EquivExp expressions should not occur in the translation to SMT-LIB!");
		return null;
  }
  
  public String visit(ExistsExp exp)
  {
  	return boolTrans.compute(exp,false);
  }
  public String visit(FalseExp exp)
  {
  	return boolTrans.compute(exp,false);
  }
  public String visit(ForallExp exp)
  {
  	return boolTrans.compute(exp,false);
  }
  public String visit(GreaterEqualExp exp)
  {
  	return intTrans.compute(exp);
  }
  public String visit(GreaterExp exp)
  {
  	return intTrans.compute(exp);
  }
  
  public String visit(HeightExp exp)
  {
  	SMT.illegalUsage(exp, "HeightExp expressions should not occur in the translation to SMT-LIB!");
		return null;
  } 
  
	public String visit(IdentifierExp exp)
	{
		return exp.ident.string;
	}
	
  public String visit(IfThenElseExp exp)
  {
  	if(exp.getType() instanceof IntType)
  		return intTrans.compute(exp);
  	String cond=exp.exp1.accept(this);
  	String thenExp=compute(exp.exp2, exp.getType());
  	String elseExp=compute(exp.exp3, exp.getType());
  	return "(ite "+cond+" "+thenExp+" "+elseExp+")";
  }
  public String visit(ImpliesExp exp)
  {
  	return boolTrans.compute(exp,false);
  }  

  public String visit(InSetExp exp)
  {
  	return setTrans.compute(exp);
  }

  public String visit(IntersectExp exp)
  {
  	return setTrans.compute(exp);
  }

  public String visit(IntervalExp exp)
  {
  	return setTrans.compute(exp);
  }
  public String visit(NumberLiteralExp exp)
  {
  	return intTrans.compute(exp);
  }
  public String visit(LessEqualExp exp)
  {
  	return intTrans.compute(exp);
  }
  public String visit(LessExp exp)
  {
  	return intTrans.compute(exp);
  }

  public String visit(LetExp exp)
  {
  	SMT.illegalUsage(exp, "Let expressions should not occur in the translation to SMT-LIB!");
		return null;
  }
	public String visit(LetParExp exp)
	{
		String str=Arrays.stream(exp.binder).map(b -> b.accept(this)).collect(Collectors.joining(" "));
		String expStr=exp.exp.accept(this);
		return "(let ("+str+") "+expStr+")";
	}

  public String visit(MatchExp exp)
  {
  	SMT.illegalUsage(exp, "MatchExp expressions should not occur in the translation to SMT-LIB!");
		return null;
  } 
  public String visit(MinusExp exp)
  {
  	return intTrans.compute(exp);
  }
  public String visit(NegationExp exp)
  {
  	return intTrans.compute(exp);
  }
  public String visit(FactorialExp exp)
  {
    return intTrans.compute(exp);
  }
	public String visit(NotEqualsExp exp)
	{
		Type ty=exp.exp1.getType();
		ty=Utilities.getBaseType(ty);
		if(ty instanceof IntType)
			return intTrans.compute(exp);
		else if(ty instanceof TupleType)
		{
			return tupleTrans.compute(exp);
		}
		else if(ty instanceof MapType)
		{
			return mapTrans.compute(exp);
		}
		else if(ty instanceof SetType)
		{
			return setTrans.compute(exp);
		}
		else
		{
			String str1=exp.exp1.accept(this);
			String str2=exp.exp2.accept(this);
			return "(distinct "+str1+" "+str2+")";
		}
	}
  public String visit(NotExp exp)
  {
  	return boolTrans.compute(exp,false);
  }

  public String visit(NumberExp exp)
  {
  	SMT.illegalUsage(exp, "NumberExp expressions should not occur in the translation to SMT-LIB!");
		return null;
  }
  public String visit(OrExp exp)
  {
  	return boolTrans.compute(exp,false);
  }
  public String visit(PlusExp exp)
  {
  	return intTrans.compute(exp);
  }
  
  public String visit(PlusExpMult exp)
  {
  	SMT.illegalUsage(exp, "PlusExpMult expressions should not occur in the translation to SMT-LIB!");
		return null;
  }
  
  public String visit(PowerExp exp)
  {
  	SMT.illegalUsage(exp, "PowerExp expressions should not occur in the translation to SMT-LIB!");
		return null;
  } 

  public String visit(PowersetExp exp)
  {
    return setTrans.compute(exp);
  } 

  public String visit(Powerset1Exp exp)
  {
  	return setTrans.compute(exp);
  } 

  public String visit(Powerset2Exp exp)
  {
  	return setTrans.compute(exp);
  } 

  public String visit(PrintExp exp)
  {
  	SMT.illegalUsage(exp, "PrintExp expressions should not occur in the translation to SMT-LIB!");
		return null;
  }

  public String visit(PrintInExp exp)
  {
  	SMT.illegalUsage(exp, "PrintInExp expressions should not occur in the translation to SMT-LIB!");
		return null;
  }

  public String visit(PrintTraceInExp exp)
  {
    SMT.illegalUsage(exp, "PrintTraceInExp expressions should not occur in the translation to SMT-LIB!");
    return null;
  }
  
  public String visit(CheckExp exp)
  {
  	SMT.illegalUsage(exp, "CheckExp expressions should not occur in the translation to SMT-LIB!");
		return null;
  }

  public String visit(RecApplicationExp exp)
  {
  	SMT.illegalUsage(exp, "Recursive applications should not occur in the translation to SMT-LIB!");
  	return null;
  }

  public String visit(RecIdentifierExp exp)
  {
  	SMT.illegalUsage(exp, "Recursive Ientifiers should not occur in the translation to SMT-LIB!");
  	return null;
  }

  public String visit(RecordExp exp)
  {
  	SMT.illegalUsage(exp, "RecordExp expressions should not occur in the translation to SMT-LIB!");
		return null;
  }

  public String visit(RecordSelectionExp exp)
  {
  	SMT.illegalUsage(exp, "RecordSelectionExp expressions should not occur in the translation to SMT-LIB!");
		return null;
  }

  public String visit(RecordUpdateExp exp)
  {
  	SMT.illegalUsage(exp, "RecordUpdateExp expressions should not occur in the translation to SMT-LIB!");
		return null;
  }

  public String visit(RemainderExp exp)
  {
  	return intTrans.compute(exp);
  }

  public String visit(SetBuilderExp exp)
  {
  	return setTrans.compute(exp);
  }
  public String visit(SumExp exp)
  {
  	return intTrans.compute(exp);
  }
  public String visit(ProductExp exp)
  {
  	return intTrans.compute(exp);
  }

  public String visit(MinExp exp)
  {
  	if(SMT.representMinMaxByUninterpretedFunctions)
  	{
  		SMT.illegalUsage(exp, "MinExp expressions should not occur in the translation to SMT-LIB!");
  		return null;
  	}
  	else
  		return intTrans.compute(exp);

  }

  public String visit(MaxExp exp)
  {
  	if(SMT.representMinMaxByUninterpretedFunctions)
  	{
  		SMT.illegalUsage(exp, "MaxExp expressions should not occur in the translation to SMT-LIB!");
  		return null;
  	}
  	else
  		return intTrans.compute(exp);
  }

  public String visit(SetSizeExp exp)
  {
  	return setTrans.compute(exp);
  }

  public String visit(SubsetExp exp)
  {
  	return setTrans.compute(exp);
  }      
  public String visit(TimesExp exp)
  {
  	return intTrans.compute(exp);
  }

  public String visit(TimesExpMult exp)
  {
  	SMT.illegalUsage(exp, "TimesExpMult expressions should not occur in the translation to SMT-LIB!");
		return null;
  }
  public String visit(TrueExp exp)
  {
    return boolTrans.compute(exp,false);
  }

  public String visit(TupleExp exp)
  {
  	return tupleTrans.compute(exp);
  }

  public String visit(TupleSelectionExp exp)
  {
  	return tupleTrans.compute(exp);
  }

  public String visit(TupleUpdateExp exp)
  {
  	return tupleTrans.compute(exp);
  }

  public String visit(UnionExp exp)
  {
  	return setTrans.compute(exp);
  }

  public String visit(UnitExp exp)
  {
    return DeclarationTranslator.unit;
  }

  public String visit(WithoutExp exp)
  {
  	return setTrans.compute(exp);
  }
 
	public String visit(Identifier ident)
	{
		return ident.string;
	}
	
	public String visit(QuantifiedVariable qVar)
	{
		String str=Arrays.stream(qVar.qvcore).map(qv -> qv.accept(this)).collect(Collectors.joining(" "));
		return "("+str+")";
	}
	
	public String visit(IdentifierTypeQuantifiedVar qvcore)
	{
		String typeString=qvcore.type.accept(this);
		return "("+qvcore.ident.string+" "+typeString+")";
	}
	
	public String visit(IdentifierSetQuantifiedVar qvcore)
	{
		SMT.illegalUsage(qvcore, "IdentifierSetQuantifiedVar should not occur in the translation to SMT-LIB!");
		return null;
	}
	  
  public String visit(BoolType type)
  {
  	return type.accept(boolTrans);
  }
  
  public String visit(IdentifierType type)
  {
  	assert(false);  	
  	return null;
  }
  public String visit(IntType type)
  {
  	return type.accept(intTrans);
  }
  public String visit(MapType type)
  {
  	return type.accept(mapTrans);
  }     
  public String visit(SetType type)
  {
  	return type.accept(setTrans);
  }    
  public String visit(SubType type)
  {
  	return type.ident.string;
  } 
  public String visit(TupleType type)
  {
  	return type.accept(tupleTrans);
  }   
  public String visit(UnitType type)
  {
  	return "(_ BitVec 1)";
  }
}
//---------------------------------------------------------------------------
// end of file
//---------------------------------------------------------------------------
