// ---------------------------------------------------------------------------
// IntTranslator.java
// A class for the translation of various integer expressions to SMT-LIB strings.
//
// Author: Franz Reichl <franz.x.reichl@gmail.com>
// $Id: IntTranslator.java,v 1.6 2020/06/24 09:54:53 schreine Exp $
// 
// Author: Franz Reichl <franz.x.reichl@gmail.com>
// Copyright (C) 2019-, Research Institute for Symbolic Computation (RISC)
// Johannes Kepler University, Linz, Austria, http://www.risc.jku.at
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
// ---------------------------------------------------------------------------
package riscal.smt;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Iterator;
import java.util.List;
import java.util.Set;
import java.util.function.Function;
import java.util.stream.Collectors;
import java.util.stream.LongStream;

import riscal.semantics.Types.Value;
import riscal.smt.ExpressionTranslator.TypeEnumerationInformation;
import riscal.syntax.AST.Binder;
import riscal.syntax.AST.Expression;
import riscal.syntax.AST.Expression.ApplicationExp;
import riscal.syntax.AST.Expression.DividesExp;
import riscal.syntax.AST.Expression.EqualsExp;
import riscal.syntax.AST.Expression.FactorialExp;
import riscal.syntax.AST.Expression.GreaterEqualExp;
import riscal.syntax.AST.Expression.GreaterExp;
import riscal.syntax.AST.Expression.HeightExp;
import riscal.syntax.AST.Expression.IdentifierExp;
import riscal.syntax.AST.Expression.IfThenElseExp;
import riscal.syntax.AST.Expression.LessEqualExp;
import riscal.syntax.AST.Expression.LessExp;
import riscal.syntax.AST.Expression.LetExp;
import riscal.syntax.AST.Expression.LetParExp;
import riscal.syntax.AST.Expression.MapSelectionExp;
import riscal.syntax.AST.Expression.MaxExp;
import riscal.syntax.AST.Expression.MinExp;
import riscal.syntax.AST.Expression.MinusExp;
import riscal.syntax.AST.Expression.NegationExp;
import riscal.syntax.AST.Expression.NotEqualsExp;
import riscal.syntax.AST.Expression.NumberLiteralExp;
import riscal.syntax.AST.Expression.PlusExp;
import riscal.syntax.AST.Expression.ProductExp;
import riscal.syntax.AST.Expression.RecApplicationExp;
import riscal.syntax.AST.Expression.RecIdentifierExp;
import riscal.syntax.AST.Expression.RemainderExp;
import riscal.syntax.AST.Expression.SetSizeExp;
import riscal.syntax.AST.Expression.SumExp;
import riscal.syntax.AST.Expression.TimesExp;
import riscal.syntax.AST.Expression.TupleSelectionExp;
import riscal.syntax.AST.QuantifiedVariable;
import riscal.syntax.AST.QuantifiedVariableCore;
import riscal.syntax.AST.QuantifiedVariableCore.IdentifierTypeQuantifiedVar;
import riscal.syntax.AST.Type;
import riscal.syntax.AST.Type.IntType;
import riscal.syntax.AST.Type.SubType;
import riscal.syntax.ASTVisitor;
import riscal.types.Environment.Symbol.ValueSymbol;
import riscal.util.Seq;
import riscal.util.Seq.Next;


public class IntTranslator extends ASTVisitor.Base<String>
{
	/**
	 * The underlying ExpressionTranslator
	 */
	private ExpressionTranslator expTrans;
	/**
	 * The bit vector length that is used for the current translation.
	 */
	private int bitVectorLength;
	/**
	 * A lower bound for the values that the currently translated
	 * expressions can attain.
	 */
	private long lowerBound;
	/**
	 * An upper bound for the values that the currently translated
	 * expressions can attain.
	 */
	private long upperBound;
	
	/**
	 * The basis for the generation of the name of the enumeration of a IntType.
	 */
	private static final String defaultTypeEnumerationName="intEnum";
	
	
	/**
	 * If true positive number literals are represented in the form #b111
	 * If false they are represented in the form (_ bv5 6)
	 */
	private static boolean positiveIntRepresentation=true;
	
	/***************************************************************************
	 * Create an IntTranslator with the given underlying ExpressionTranslator.
	 * @param et The underlying ExpressionTranslator.
	 **************************************************************************/
	public IntTranslator(ExpressionTranslator et)
	{
		this.expTrans=et;
	}
	
	/***************************************************************************
	 * Translate the given expression to SMT-LIB.
	 * @param exp The expression that shall be translated.
	 * Has to be an integer related expression.
	 * @return The result of the translation.
	 **************************************************************************/
	public String compute(Expression exp)
	{
		long[] a=getMaxLength(exp);
		int oldlen=bitVectorLength;
		long oldLowerBound=lowerBound;
		long oldUpperBound=upperBound;
		this.bitVectorLength=determineBitVectorLength(a[0], a[1]);
		this.lowerBound=a[0];
		this.upperBound=a[1];
		String str=exp.accept(this);
		if(exp.getType() instanceof IntType)
		{
			IntType it=(IntType)exp.getType();
			int len=determineBitVectorLength(it.ivalue1, it.ivalue2);
			//Since we used the maximal int interval we do not have problems with resizing
			str=resizeBitVector(str, bitVectorLength, len, lowerBound<0);
		}
		this.bitVectorLength=oldlen;
		this.lowerBound=oldLowerBound;
		this.upperBound=oldUpperBound;
		return str;
	}	
	
	/***************************************************************************
	 * Translate the given expression to SMT-LIB.
	 * If the given expression is an integer expression, than it
	 * shall be treated like an expression of type Int[lower,upper]
	 * @param exp The expression that shall be translated.
	 * Has to be an integer related expression.
	 * @param lower If exp is an integer expression, than lower has
	 * to be smaller than the actual lower bound of exp.
	 * @param upper If exp is an integer expression, than upper has
	 * to be larger than the actual upper bound of exp.
	 * @return The result of the translation.
	 **************************************************************************/
	public String compute(Expression exp,long lower,long upper)
	{
		int oldlen=bitVectorLength;
		long oldLowerBound=lowerBound;
		long oldUpperBound=upperBound;
		long[] a=getMaxLength(exp);
		long lB=a[0]<lower?a[0]:lower;
		long uB=a[1]>upper?a[1]:upper;
		this.bitVectorLength=determineBitVectorLength(lB, uB);
		this.lowerBound=lB;
		this.upperBound=uB;
		String str=exp.accept(this);
		if(exp.getType() instanceof IntType)
		{
			int len=determineBitVectorLength(lower, upper);
			//Since we used the maximal int interval we do not have problems with resizing
			str=resizeBitVector(str, bitVectorLength, len, lowerBound<0);
		}
		this.bitVectorLength=oldlen;
		this.lowerBound=oldLowerBound;
		this.upperBound=oldUpperBound;
		return str;
	}
	
	/***************************************************************************
	 * Takes a SMT-LIB expression that has the sort that corresponds
	 * to oldType. The method transforms this SMT-LIB expression to
	 * a SMT-LIB expression of the sort that corresponds to newType.
	 * Condition:
	 * oldType must be contained in newType
	 * @param s The SMT-LIB expression that shall be transformed.
	 * @param oldType The type that corresponds to the current sort.
	 * @param newType The type that corresponds to the sort the result
	 * shall have.
	 * @return A SMT-LIB expression of the sort that corresponds to
	 * newType
	 **************************************************************************/
	static String adaptTranslation(String s,IntType oldType,IntType newType)
	{
		long low=oldType.ivalue1;
		int len1=determineBitVectorLength(low, oldType.ivalue2);
		int len2=determineBitVectorLength(newType.ivalue1, newType.ivalue2);
		assert(len1<=len2);
		return resizeBitVector(s, len1, len2, low<0);		
	}
	

	
	/**
	 * Necessary for determining the position of the first set bit.
	 * Source: chess programming wiki. https://www.chessprogramming.org/BitScan#Bitscan_reverse
	 */
	private static final int index64[] = new int[]{
	    0, 47,  1, 56, 48, 27,  2, 60,
	   57, 49, 41, 37, 28, 16,  3, 61,
	   54, 58, 35, 52, 50, 42, 21, 44,
	   38, 32, 29, 23, 17, 11,  4, 62,
	   46, 55, 26, 59, 40, 36, 15, 53,
	   34, 51, 20, 43, 31, 22, 10, 45,
	   25, 39, 14, 33, 19, 30,  9, 24,
	   13, 18,  8, 12,  7,  6,  5, 63
	};
	
	/***************************************************************************
	 * Determine the position of the first set bit in a positive 64 bit integer.
	 * If the input is 0 the function returns 0.
	 * Source: chess programming wiki https://www.chessprogramming.org/BitScan#Bitscan_reverse
	 * @param x A non-negative long value of which we want to determine the position of the
	 * most significant bit
	 * @return The position of the most significant bit of x. (can be in the range 0..63)
	 **************************************************************************/
	private static int bitScanReverse(long x) 
	{
		if(x==0)
			return 0;
		long debruijn64 = 0x03f79d71b4cb0a89L;
		x |= x >>> 1; 
		x |= x >>> 2;
		x |= x >>> 4;
		x |= x >>> 8;
		x |= x >>> 16;
		x |= x >>> 32;
		return index64[(int) ((x * debruijn64) >>> 58)]+1;
	}
	
	/***************************************************************************
	 * Determine the minimal number such
	 * that bit vectors of this length can
	 * represent all integers in [a,b]
	 * The function requires that a<=b
	 * @param a The lower bound of an integer type
	 * @param b The upper bound of an integer type
	 * @return A bit vector length which is sufficient to represent the 
	 * numbers a..b
	 **************************************************************************/
	public static int determineBitVectorLength(long a,long b)
	{
		assert a<=b;
		/*
		 * If a>=0 we are considering a subset of the natural numbers.
		 * This means we can represent these number is an unsigned way.
		 * So the necessary length of bit vectors that can represent such numbers
		 * is given by the most significant bit of b.
		 */
		if(a>=0)
		{
			if(b==0)
				return 1;
			else
				return bitScanReverse(b);
		}			
		/*
		 * In this case we also have to consider negative numbers. 
		 * We use -(a+1) instead of -a-1 in order to prevent problems with the minimal long value.
		 * Because of the condition the necessary length of the bit vectors is given by the
		 * position of the most significant bit in b plus 1. (plus one because of the "sign bit")
		 */
		else if(-(a+1)<=b)
			return bitScanReverse(b)+1;
		/*
		 * In this case the length of the bit vectors cannot be concluded from the upper bound.
		 * So we consider the lower bound.
		 */
		else
			return bitScanReverse(-(a+1))+1;	
	}
	
	/***************************************************************************
	 * Compute a bit vector of length len that represents toBit
	 * The caller is responsible to ensure that the number toBit
	 * can be represented by a bit vector of size len.
	 * @param toBit The number which we want to represent by a sequence of bits. 
	 * If toBit is non negative: toBit has to be in in 0..2^(len+1)-1. 
	 * Note if the sequence shall represent a signed number toBit in 0..2^n -1
	 * If toBit is negative: toBit has to be in -2^n...2^n -1
	 * @param len The length, the sequence shall have. Must be at least 1.
	 * @return A string containing a bit representation of toBit
	 **************************************************************************/
	static String buildBitSequence(long toBit, long len)
	{
		if(toBit>=0)
			return toBit(toBit,len);
		else
		{
			if(len==1)
				return "1";			
			long x=1L<<(len-1);//2power len-1
			String str=toBit(toBit+x, len-1);
			return "1"+str;
		}
	}
	
	/***************************************************************************
	 * Compute a SMT-LIB bit vector constant of length len that
	 * represents the number toBit.
	 * The caller is responsible to ensure that the number toBit
	 * can be represented by a bit vector of size len.
	 * @param toBit The number that shall be represented
	 * @param len The length the bit vector shall have.
	 * @return A string containing a bit representation of toBit
	 **************************************************************************/
	public static String generateSMT_LIB_Constant(long toBit, long len)
	{
		if(positiveIntRepresentation||toBit<0)
		{
			return "#b"+buildBitSequence(toBit, len);
		}
		else
		{
			return "(_ bv"+Long.toString(toBit)+" "+Long.toString(len)+")";
		}	
	}
	
	/***************************************************************************
	 * Compute the SMT-LIB representations of the values of the given type.
	 * Add the #b prefix to these representations depending on bare.
	 * @param ty The type whose values shall be computed. Has to be canonical.
	 * @param bare If true do not add the #b prefix to the resulting values.
	 * @return A list containing all values of the given type 
	 * represented by their corresponding SMT-LIB strings.
	 **************************************************************************/
	public static List<String> computeValues(IntType ty,boolean bare)
	{
		int vectorLength=determineBitVectorLength(ty.ivalue1, ty.ivalue2);
		if(bare)
		{
			return LongStream.rangeClosed(ty.ivalue1, ty.ivalue2).boxed().map(
					ival -> buildBitSequence(ival, vectorLength)).collect(Collectors.toList());
		}
		else
		{
			return LongStream.rangeClosed(ty.ivalue1, ty.ivalue2).boxed().map(
					ival -> generateSMT_LIB_Constant(ival, vectorLength)).collect(Collectors.toList());
		}		
	}
		
	/***************************************************************************
	 * Translate a sequence of values with base type type to a list of SMT-LIB values.
	 * @param type The base type of the values. Must be canonical.
	 * @param seq A sequence of values whose basetype is type.
	 * @param bare If true do not add the #b prefix to the resulting values.
	 * @return A list containing the SMT-LIB representations of the
	 * values in seq.
	 **************************************************************************/
	static List<String> processValueSequence(Type type,Seq<Value> seq,boolean bare)
	{
		IntType base=(IntType) Utilities.getBaseType(type);
		int len=ExpressionTranslator.getTypeSize(type);
		List<String> result=new ArrayList<String>(len);
		int vectorLength;		
		vectorLength=determineBitVectorLength(base.ivalue1, base.ivalue2);
		Function <Value, String> proc;
		if(bare)
			proc=v -> processValueBare(vectorLength,v);
		else
			proc=v -> processValue(vectorLength,v);
		while (true)
    {
      Next<Value> next = seq.get();
      if (next == null) break;
      result.add(proc.apply(next.head));
      seq = next.tail;
    }
		return result;	
	}
	
	/***************************************************************************
	 * Translate a value to a bit vector of length vectorLength.
	 * The caller is responsible to ensure that the value can be
	 * represented by bit vectors of the denoted length.
	 * @param vectorLength The length of the resulting bit vector.
	 * @param val An int value that can be represented by bit vectors of length
	 * vectorLength.
	 * @return A SMT-LIB expression representing val.
	 **************************************************************************/
	private static String processValue(int vectorLength,Value val)
	{
		int intValue=((riscal.semantics.Types.Value.Int) val).getValue();
		return generateSMT_LIB_Constant(intValue, vectorLength);
	}
	
	/***************************************************************************
	 * Translate a value to a bit vector of length vectorLength.
	 * Do not add the #b prefix to the resulting bit sequence.
	 * The caller is responsible to ensure that the value can be
	 * represented by bit vectors of the denoted length.
	 * @param vectorLength The length of the resulting bit vector.
	 * @param val An int value that can be represented by bit vectors of length
	 * vectorLength.
	 * @return A bit sequence representing val.
	 **************************************************************************/
	private static String processValueBare(int vectorLength,Value val)
	{
		int intValue=((riscal.semantics.Types.Value.Int) val).getValue();
		return buildBitSequence(intValue, vectorLength);
	}
	
	static class IntValueContainer extends ExpressionTranslator.ValueContainer
	{
		/**
		 * The length the translation of the values shall have.
		 */
		private int vectorLength;
		/**
		 * Create an IntValueContainer for values of corresponding to IntType iT
		 * @param iT The type corresponding to the values that shall be
		 * processed by this IntValueContainer. Must be canonical.
		 */
		IntValueContainer(IntType iT)
		{
			vectorLength=determineBitVectorLength(iT.ivalue1, iT.ivalue2);
		}		
		/***************************************************************************
		 * Generate a bit vector constant representing the given value.
		 * @param val The value that shall be represented by a bit vector.
		 * @return The SMT-LIB representation of val
		 **************************************************************************/
		String process(Value val) 
		{
			return processValue(vectorLength, val);
		}
		/***************************************************************************
		 * Generate a bit vector constant representing the given value without
		 * the #b prefix.
		 * @param val The value that shall be represented by a bit vector.
		 * @return The SMT-LIB representation of val
		 **************************************************************************/
		String processToBareVector(Value val)
		{
			return processValueBare(vectorLength, val);
		}
	}
	
	/***************************************************************************
	 * Generate an enumeration function for the given type and return the
	 * associated information.
	 * @param iT The type for which an enumeration function shall be generated.
	 * @return The information associated to the enumeration function.
	 **************************************************************************/
	public TypeEnumerationInformation generateTypeEnumerationFunctionNonSubType(IntType iT)
	{
		String funName=expTrans.getFunctionName(defaultTypeEnumerationName);
		int len=ExpressionTranslator.getTypeSize(iT);
		int typeVectorLength=determineBitVectorLength(iT.ivalue1, iT.ivalue2);
		int argumentVectorLength=determineBitVectorLength(0, len-1);
		long shift=iT.ivalue1;
		String variable="x";
		if(shift==0)
		{
			expTrans.addDeclaration("(define-fun "+funName+" (("+ variable +"(_ BitVec "+Integer.toString(typeVectorLength)+"))) (_ BitVec "+
					Integer.toString(argumentVectorLength)+")"+variable+")");
		}
		else
		{
			String funDef="(bvsub "+variable+" "+IntTranslator.generateSMT_LIB_Constant(shift, typeVectorLength)+")";
			if(typeVectorLength!=argumentVectorLength)
			{
				funDef="((_ extract "+Integer.toString(argumentVectorLength-1)+" 0) "+funDef+")";
			}
			expTrans.addDeclaration("(define-fun "+funName+" (("+ variable +"(_ BitVec "+Integer.toString(typeVectorLength)+"))) (_ BitVec "+
					Integer.toString(argumentVectorLength)+")"+funDef+")");		
		}
		return new TypeEnumerationInformation(funName, argumentVectorLength, len);
	}
	
	/***************************************************************************
	 * Generates an enumeration function for the values given by lst.
	 * The result of the enumeration is given by the position 
	 * of the respective value in the list.
	 * The enumeration function is defined via a table.
	 * @param lst The values used to generate the enumeration function.
	 * @param iT The elements of lst must correspond to the type iT. Must be canonical.
	 * @return The information associated to the enumeration function.
	 **************************************************************************/
	public TypeEnumerationInformation generateTypeEnumerationFunctionUseTable(List<Value> lst,IntType iT)
	{
		String funName=expTrans.getFunctionName(defaultTypeEnumerationName);
		long len=lst.size();
		int typeVectorLength=determineBitVectorLength(iT.ivalue1, iT.ivalue2);
		int argumentVectorLength=determineBitVectorLength(0, len-1);
		expTrans.addDeclaration("(declare-fun "+funName+" ((_ BitVec "+Integer.toString(typeVectorLength)+")) "
				+ "(_ BitVec "+Long.toString(argumentVectorLength)+"))");
		long i=0;
		for(Value val:lst)
		{
			expTrans.addDeclaration("(assert (= ("+funName+" "+processValue(typeVectorLength,val)+") "+
					IntTranslator.generateSMT_LIB_Constant(i, argumentVectorLength)+"))");
			i++;
		}
		return new TypeEnumerationInformation(funName, argumentVectorLength, len);
	}
	
	public TypeEnumerationInformation generateTypeEnumerationFunctionUseTableWithFunctionDefinition(List<Value> lst,IntType iT)
	{
		String funName=expTrans.getFunctionName(defaultTypeEnumerationName);
		long len=lst.size();
		int typeVectorLength=determineBitVectorLength(iT.ivalue1, iT.ivalue2);
		int argumentVectorLength=determineBitVectorLength(0, len-1);
		StringBuilder strB=new StringBuilder();
		String variable="x";
		Iterator<Value> it=lst.iterator();
		for(int i=0;i<len-1;i++)
		{
			Value val=it.next();
			strB.append("(ite (= "+variable+" "+processValue(typeVectorLength,val)+")"+
					IntTranslator.generateSMT_LIB_Constant(i, argumentVectorLength)+" ");
		}
		strB.append(IntTranslator.generateSMT_LIB_Constant(len-1, argumentVectorLength));
		strB.append(")".repeat((int) (len-1)));
		String def="(define-fun "+funName+" (("+variable+" (_ BitVec "+Integer.toString(typeVectorLength)+"))) "
				+ "(_ BitVec "+Long.toString(argumentVectorLength)+") "+strB.toString()+")";
		expTrans.addDeclaration(def);
		return new TypeEnumerationInformation(funName, argumentVectorLength, len);
	}

	/***************************************************************************
	 * Compute the bit vector length that is necessary to represent the values
	 * of the given type.
	 * @param iT The type of interest. Must be canonical.
	 * @return The necessary length to represent the values of iT
	 **************************************************************************/
	public static long giveVectorLength(IntType iT)
	{
		long len=IntTranslator.determineBitVectorLength(iT.ivalue1, iT.ivalue2);
		return len;
	}
	
	/***************************************************************************
	 * Compute the bit vector length that is necessary to represent the values
	 * of the type of the given expression.
	 * @param exp An integer expression.
	 * @return The necessary length to represent the values of the type.
	 **************************************************************************/
	private static int getVectorLength(Expression exp)
	{
		long[] ar=getMaxLength(exp);
		return determineBitVectorLength(ar[0], ar[1]);
	}
	
	/***************************************************************************
	 * Generates a boolean valued SMT-LIB expression that is true
	 * iff the SMT-LIB expression str is greater equal than the bit vector
	 * representation of lower and less equal than the bit vector representation
	 * of upper.
	 * @param lower The lower bound for the condition.
	 * @param upper The upper bound for the condition.
	 * @param len The length that shall be used for the
	 * bit vector representations of lower and upper.
	 * @param str A SMT-LIB expression, whose sort are
	 * the bit vectors of length len.
	 * @return A boolean valued SMT-LIB expression.
	 **************************************************************************/
	public static String generateCondition(long lower,long upper,int len,String str)
	{
		long upperLimit=0<=lower?riscal.util.Arith.power(2, len)-1:riscal.util.Arith.power(2, len-1)-1;
		long lowerLimit=0<=lower?0:-riscal.util.Arith.power(2, len-1);
		if(lowerLimit==lower&&upperLimit==upper)
		{
			return null;
		}
		else if(lowerLimit==lower)
		{
			String upperBound=generateSMT_LIB_Constant(upper,len);
			if(lower<0)
			{
				return "(bvsle "+str+" "+upperBound+")";
			}
			else
			{
				return "(bvule "+str+" "+upperBound+")";
			}
		}
		else if(upperLimit==upper)
		{
			String lowerBound=generateSMT_LIB_Constant(lower,len);
			if(lower<0)
			{
				return "(bvsge "+str+" "+lowerBound+")";
			}
			else
			{
				return "(bvuge "+str+" "+lowerBound+")";
			}
		}
		else
		{
			String upperBound=generateSMT_LIB_Constant(upper,len);
			String lowerBound=generateSMT_LIB_Constant(lower,len);
			if(lower<0)
			{
				return "(and (bvsle "+lowerBound+" "+str+") (bvsge "+upperBound+" "+str+"))";
			}
			else
			{
				return "(and (bvule "+lowerBound+" "+str+") (bvuge "+upperBound+" "+str+"))";
			}
		}
	}
	
	/***************************************************************************
	 * Generates a boolean valued SMT-LIB expression that is true
	 * iff the SMT-LIB expression str is greater equal than the bit vector
	 * representation of lower and less equal than the bit vector representation
	 * of upper.
	 * @param lower The lower bound for the condition.
	 * @param upper The upper bound for the condition.
	 * @param str A SMT-LIB bit vector expression.
	 * @return A boolean valued SMT-LIB expression.
	 **************************************************************************/
	public static String generateCondition(long lower,long upper,String str)
	{
		return generateCondition(lower, upper, determineBitVectorLength(lower, upper), str);
	}
	
	/***************************************************************************
	 * Compute a bit vector of length len that represents a
	 * a must be representable by a bit vector of length len
	 * @param a The number that we want to represent by a bit sequence
	 * a>=0
	 * @param len The length, the sequence shall have. len >=1
	 * @return A string containing a bit sequence representing a
	 **************************************************************************/
	private static String toBit(long a, long len)
	{
		assert a>=0;
		StringBuilder strb=new StringBuilder();
		for(long i=len-1;i>=0;i--)
		{
			long x=1L<<i;
			if((a&x)==x)//checks if the i+1 th bit is set.
				strb.append("1");
			else
				strb.append("0");			
		}
		return strb.toString();
	}

	/***************************************************************************
	 * Resize a bit vector expression of length oldLength to a bit vector
	 * expression newLength representing the same value.
	 * @param str The SMT-LIB expression that shall be resized.
	 * Must be an expression of length oldLength.
	 * @param oldLength The current bit vector length of str.
	 * @param newLength The new bit vector length. 
	 * The caller is responsible to ensure that the represented number fits
	 * into the new vector. 
	 * @param oldSigned true if the given string could represent a negative
	 * number.
	 * @return A SMT-LIB bit vector expression of length newLength.
	 **************************************************************************/
	static String resizeBitVector(String str,int oldLength,int newLength,boolean oldSigned)
	{
		int diff=newLength-oldLength;
		if(diff==0)
			return str;
		if(diff>0)
		{
			if(oldSigned)
			{
				return "((_ sign_extend "+Integer.toString(diff)+") "+str+")";
			}
			else
			{
				return "((_ zero_extend "+Integer.toString(diff)+") "+str+")";
			}
		}
		else
		{
			return "((_ extract "+Integer.toString(newLength-1)+" 0) "+str+")";
		}
	}
	
	/***************************************************************************
	 * Determine the minimal and the maximal value that can be attained
	 * by exp or by one of its sub-expressions.
	 * @param exp The expression of interest.
	 * @return An array of length two containing the
	 * minimal and the maximal value that can occur in exp.
	 **************************************************************************/
	private static long[] getMaxLength(Expression exp)
	{
		long[] aux1;
		long[] aux2;
		long lower;
		long upper;
		long ival1;
		long ival2;
		
		if(exp instanceof EqualsExp)
		{
			aux1=getMaxLength(((EqualsExp) exp).exp1);
			aux2=getMaxLength(((EqualsExp) exp).exp2);
			lower=Math.min(aux1[0], aux2[0]);
			upper=Math.max(aux1[1], aux2[1]);
			return new long[] {lower, upper};
		}
		else if(exp instanceof NotEqualsExp)
		{
			aux1=getMaxLength(((NotEqualsExp) exp).exp1);
			aux2=getMaxLength(((NotEqualsExp) exp).exp2);
			lower=Math.min(aux1[0], aux2[0]);
			upper=Math.max(aux1[1], aux2[1]);
			return new long[] {lower, upper};
		}
		else if(exp instanceof LessExp)
		{
			aux1=getMaxLength(((LessExp) exp).exp1);
			aux2=getMaxLength(((LessExp) exp).exp2);
			lower=Math.min(aux1[0], aux2[0]);
			upper=Math.max(aux1[1], aux2[1]);
			return new long[] {lower, upper};
		}
		else if(exp instanceof LessEqualExp)
		{
			aux1=getMaxLength(((LessEqualExp) exp).exp1);
			aux2=getMaxLength(((LessEqualExp) exp).exp2);
			lower=Math.min(aux1[0], aux2[0]);
			upper=Math.max(aux1[1], aux2[1]);
			return new long[] {lower, upper};
		}
		else if(exp instanceof GreaterExp)
		{
			aux1=getMaxLength(((GreaterExp) exp).exp1);
			aux2=getMaxLength(((GreaterExp) exp).exp2);
			lower=Math.min(aux1[0], aux2[0]);
			upper=Math.max(aux1[1], aux2[1]);
			return new long[] {lower, upper};
		}
		else if(exp instanceof GreaterEqualExp)
		{
			aux1=getMaxLength(((GreaterEqualExp) exp).exp1);
			aux2=getMaxLength(((GreaterEqualExp) exp).exp2);
			lower=Math.min(aux1[0], aux2[0]);
			upper=Math.max(aux1[1], aux2[1]);
			return new long[] {lower, upper};
		}
		
		ival1=((IntType)exp.getType()).ivalue1;
		ival2=((IntType)exp.getType()).ivalue2;
		if(exp instanceof PlusExp)
		{
			aux1=getMaxLength(((PlusExp) exp).exp1);
			aux2=getMaxLength(((PlusExp) exp).exp2);
			lower=Math.min(ival1, Math.min(aux1[0], aux2[0]));
			upper=Math.max(ival2, Math.max(aux1[1], aux2[1]));
			return new long[] {lower, upper};
		}
		else if(exp instanceof MinusExp)
		{
			aux1=getMaxLength(((MinusExp) exp).exp1);
			aux2=getMaxLength(((MinusExp) exp).exp2);
			lower=Math.min(ival1, Math.min(aux1[0], aux2[0]));
			upper=Math.max(ival2, Math.max(aux1[1], aux2[1]));
			return new long[] {lower, upper};
		}
		else if(exp instanceof NegationExp)
		{
			aux1=getMaxLength(((NegationExp) exp).exp);
			lower=Math.min(ival1,aux1[0]);
			upper=Math.max(ival2,aux1[1]);
			return new long[] {lower, upper};
		}
		else if(exp instanceof TimesExp)
		{
			aux1=getMaxLength(((TimesExp) exp).exp1);
			aux2=getMaxLength(((TimesExp) exp).exp2);
			lower=Math.min(ival1, Math.min(aux1[0], aux2[0]));
			upper=Math.max(ival2, Math.max(aux1[1], aux2[1]));
			return new long[] {lower, upper};
		}
		else if(exp instanceof DividesExp)
		{
			aux1=getMaxLength(((DividesExp) exp).exp1);
			aux2=getMaxLength(((DividesExp) exp).exp2);
			lower=Math.min(ival1, Math.min(aux1[0], aux2[0]));
			upper=Math.max(ival2, Math.max(aux1[1], aux2[1]));
			return new long[] {lower, upper};
		}
		else if(exp instanceof RemainderExp)
		{
			aux1=getMaxLength(((RemainderExp) exp).exp1);
			aux2=getMaxLength(((RemainderExp) exp).exp2);
			lower=Math.min(ival1, Math.min(aux1[0], aux2[0]));
			upper=Math.max(ival2, Math.max(aux1[1], aux2[1]));
			return new long[] {lower, upper};
		}
		else if(exp instanceof SumExp)
		{
			aux1=getMaxLength(((SumExp)exp).exp);
			lower=Math.min(ival1,aux1[0]);
			upper=Math.max(ival2,aux1[1]);
			return new long[] {lower, upper};
		}
		else if(exp instanceof ProductExp)
		{
			aux1=getMaxLength(((ProductExp)exp).exp);
			lower=Math.min(ival1,aux1[0]);
			upper=Math.max(ival2,aux1[1]);
			return new long[] {lower, upper};
		}
		else if(exp instanceof LetParExp)
		{
			aux1=getMaxLength(((LetParExp)exp).exp);
			lower=Math.min(ival1,aux1[0]);
			upper=Math.max(ival2,aux1[1]);
			return new long[] {lower, upper};
		}
		else if(exp instanceof IdentifierExp)
		{
			return new long[] {ival1,ival2};
		}
		else if(exp instanceof ApplicationExp)
		{
			return new long[] {ival1,ival2};
		}
		else if(exp instanceof NumberLiteralExp)
		{
			return new long[] {ival1,ival2};
		}
		else if(exp instanceof IfThenElseExp)
		{
			aux1=getMaxLength(((IfThenElseExp) exp).exp2);
			aux2=getMaxLength(((IfThenElseExp) exp).exp3);
			lower=Math.min(ival1, Math.min(aux1[0], aux2[0]));
			upper=Math.max(ival2, Math.max(aux1[1], aux2[1]));
			return new long[] {lower, upper};
		}
		else if(exp instanceof SetSizeExp)
		{
			return new long[] {ival1,ival2};
		}
//		else if(exp instanceof NumberExp)
//		{
//			return new long[] {ival1,ival2};
//		}
		else if(exp instanceof TupleSelectionExp)
		{
			return new long[] {ival1,ival2};
		}
//		else if(exp instanceof RecordSelectionExp)
//		{
//			return new long[] {ival1,ival2};
//		}
		else if(exp instanceof MapSelectionExp)
		{
			return new long[] {ival1,ival2};
		}
//		else if(exp instanceof ChooseExp)
//		{
//			return new long[] {ival1,ival2};
//		}
		else if(exp instanceof MaxExp)
		{
			return new long[] {ival1,ival2};
		}
		else if(exp instanceof MinExp)
		{
			return new long[] {ival1,ival2};
		}
		else
		{
			SMT.illegalUsage(exp, "Currently it is not possible to determine the maximal bounds of the given Expression.");
			return null;
		}
	}
		
	public String visit(NumberLiteralExp exp)
	{
		return generateSMT_LIB_Constant(exp.literal.getInteger(), bitVectorLength);
	}
	
	public String visit(PlusExp exp)
	{
		String str1=exp.exp1.accept(this);
		String str2=exp.exp2.accept(this);
		return "(bvadd "+str1+" "+str2+")";
	}
	
	public String visit(MinusExp exp)
	{
		String str1=exp.exp1.accept(this);
		String str2=exp.exp2.accept(this);
		return "(bvsub "+str1+" "+str2+")";
	}
	
	public String visit(NegationExp exp)
	{
		String str1=exp.exp.accept(this);
		return "(bvneg "+str1+")";
	}
	
	public String visit(TimesExp exp)
	{
		String str1=exp.exp1.accept(this);
		String str2=exp.exp2.accept(this);
		return "(bvmul "+str1+" "+str2+")";
	}
	
	public String visit(DividesExp exp)
	{
		String str1=exp.exp1.accept(this);
		String str2=exp.exp2.accept(this);
		if(lowerBound<0)
		{
			return "(bvsdiv "+str1+" "+str2+")";
		}
		else
		{
			return "(bvudiv "+str1+" "+str2+")";
		}
	}
	
	/***************************************************************************
	 * Applies the SMT-LIB addition to the given arguments.
	 * @param summands Must contain at least one element. The elements
	 * of the list must be bit vector expressions of the same length.
	 * @return A SMT-LIB addition applied to the elements of summands.
	 **************************************************************************/
	private static String applyAddition(List<String> summands)
	{
		String args=String.join(" ", summands);
		if(summands.size()==1)
			return args;
		return "(bvadd "+args+")";
	}
	
	public String visit(SumExp exp)
	{
		List<String> vals=expTrans.computeValues(((ValueSymbol)((IdentifierTypeQuantifiedVar)exp.qvar.qvcore[0]).ident.getSymbol()).type);
		if(vals.size()==0)
			return generateSMT_LIB_Constant(0, bitVectorLength);
		String s=exp.exp.accept(this);
		Type ty=((ValueSymbol)exp.qvar.qvcore[0].ident.getSymbol()).type;
		String variable=exp.qvar.qvcore[0].ident.string;
		if(ty instanceof SubType&&SMT.expandBaseTypesOfSubTypes)
		{
			String cond=expTrans.generateTypeConditions(ty, variable);
			if(cond!=null)
			{
				s="(ite "+cond+" "+s+" "+generateSMT_LIB_Constant(0, bitVectorLength)+")";
			}
		}
		if(ExpressionTranslator.useAuxillaryFunctionInExpansions(vals.size(), s.length()))
		{
			Set<ValueSymbol> freeSyms=expTrans.typedFreeSymbolsWithoutValues(exp.exp);
			String typeString="(_ BitVec "+Integer.toString(bitVectorLength)+")";
			if(SMT.plugValuesDirectlyIntoAuxiliaryFunctions)
			{
				List<String> lst=expTrans.generateAuxillaryExpansionFunction(s, freeSyms, (ValueSymbol)exp.qvar.qvcore[0].ident.getSymbol(), typeString);
				return applyAddition(lst);
			}
			else
			{
				s=expTrans.generateAuxillaryExpansionFunction(s, freeSyms,typeString);
			}
		}		
		String expStr=s;
		String summands=vals.stream().map(str -> "(let (("+variable+" "+str+")) "+expStr+")").
				collect(Collectors.joining(" "));
		if(vals.size()==1)
			return summands;
		return "(bvadd "+summands+")";
	}
	
	/***************************************************************************
	 * Applies the SMT-LIB multiplication to the given arguments.
	 * @param factors Must contain at least one element. The elements
	 * of the list must be bit vector expressions of the same length.
	 * @return A SMT-LIB multiplication applied to the elements of summands.
	 **************************************************************************/
	private static String applyMuliplication(List<String> factors)
	{
		String args=String.join(" ", factors);
		if(factors.size()==1)
			return args;
		return "(bvmul "+args+")";
	}
	
	public String visit(ProductExp exp)
	{
		List<String> vals=expTrans.computeValues(((ValueSymbol)((IdentifierTypeQuantifiedVar)exp.qvar.qvcore[0]).ident.getSymbol()).type);
		if(vals.size()==0)
			return generateSMT_LIB_Constant(1, bitVectorLength);
		String s=exp.exp.accept(this);
		Type ty=((ValueSymbol)exp.qvar.qvcore[0].ident.getSymbol()).type;
		String variable=exp.qvar.qvcore[0].ident.string;
		if(ty instanceof SubType&&SMT.expandBaseTypesOfSubTypes)
		{
			String cond=expTrans.generateTypeConditions(ty, variable);
			if(cond!=null)
			{
				s="(ite "+cond+" "+s+" "+generateSMT_LIB_Constant(1, bitVectorLength)+")";
			}
		}
		if(ExpressionTranslator.useAuxillaryFunctionInExpansions(vals.size(), s.length()))
		{
			Set<ValueSymbol> freeSyms=expTrans.typedFreeSymbolsWithoutValues(exp.exp);
			String typeString="(_ BitVec "+Integer.toString(bitVectorLength)+")";
			if(SMT.plugValuesDirectlyIntoAuxiliaryFunctions)
			{
				List<String> lst=expTrans.generateAuxillaryExpansionFunction(s, freeSyms, (ValueSymbol)exp.qvar.qvcore[0].ident.getSymbol(), typeString);
				return applyMuliplication(lst);
			}
			else
			{
				s=expTrans.generateAuxillaryExpansionFunction(s, freeSyms,typeString);
			}
		}	
		String expStr=s;
		String factors=vals.stream().map(str -> "(let (("+exp.qvar.qvcore[0].ident.string+" "+str+")) "+expStr+")").
				collect(Collectors.joining(" "));
		if(vals.size()==1)
			return factors;
		return "(bvmul "+factors+")";
	}
	
	private String translateMinMax(Expression exp,String comparison,String defaultVal, String auxFun,QuantifiedVariable qv)
	{
		String expStr=compute(exp, lowerBound, upperBound);
		expTrans.addDeclaration("(define-fun "+auxFun+" ((x (_ BitVec "+Long.toString(bitVectorLength)+"))"
				+ "(y (_ BitVec "+Long.toString(bitVectorLength)+"))) (_ BitVec "+Long.toString(bitVectorLength)+")"
				+ " (ite ("+comparison+" x y) x y))");
		String typeString="(_ BitVec "+Integer.toString(bitVectorLength)+")";
		QuantifiedVariableCore[] qvc=qv.qvcore;
		String toBuild=defaultVal;
		String aux;
		Set<ValueSymbol> freeSyms=expTrans.typedFreeSymbolsWithoutValues(exp);
		for(int i=qvc.length-1;i>=0;i--)
		{
			String base;
			if(i==qvc.length-1)
				base=expStr;
			else
				base=toBuild;
			IdentifierTypeQuantifiedVar itqv=(IdentifierTypeQuantifiedVar)qvc[i];
			Type ty=((ValueSymbol)itqv.ident.getSymbol()).type;
			String variable=itqv.ident.string;
			String condition=null;
			if(itqv.wexp!=null)
			{
				freeSyms.addAll(expTrans.typedFreeSymbolsWithoutValues(itqv.wexp));
				condition=expTrans.compute(itqv.wexp);
			}
			if(ty instanceof SubType&&SMT.expandBaseTypesOfSubTypes)
			{
				String cond=expTrans.generateTypeConditions(ty, variable);
				condition=condition==null?cond:"(and "+condition+" "+cond+")";
			}
			aux=condition!=null?"(ite "+condition+" "+base+" "+defaultVal+")":base;
			List<String> vals=expTrans.computeValues(ty);
			if(ExpressionTranslator.useAuxillaryFunctionInExpansions(vals.size(), expStr.length()))
			{
				if(SMT.plugValuesDirectlyIntoAuxiliaryFunctions)
				{
					List<String> lst=expTrans.generateAuxillaryExpansionFunction(aux, freeSyms, (ValueSymbol)itqv.ident.getSymbol(), typeString);
					assert(lst.size()>0);
					Iterator<String> it=lst.iterator();
					toBuild=defaultVal;
					//new
					StringBuilder strB=new StringBuilder();
					strB.append(("("+auxFun).repeat(lst.size()));
					strB.append(" "+defaultVal);
					while(it.hasNext())
					{
						strB.append(" "+it.next()+")");
					}
					toBuild=strB.toString();
//					while(it.hasNext())
//					{
//						toBuild="("+auxFun+" "+toBuild+" "+it.next()+")";
//					}
					freeSyms.remove(itqv.ident.getSymbol());
					continue;
				}
				else
				{
					aux=expTrans.generateAuxillaryExpansionFunction(aux, freeSyms,typeString);	
				}
			}
			toBuild=defaultVal;
			//new
			StringBuilder strB=new StringBuilder();
			strB.append(("("+auxFun).repeat(vals.size()));
			strB.append(" "+defaultVal);
			for(String v:vals)
			{
				strB.append(" (let (("+variable+" "+v+")) "+aux+"))");
			}
			toBuild=strB.toString();
//			for(String v:vals)
//			{
//				toBuild="("+auxFun+" "+toBuild+" (let (("+variable+" "+v+")) "+aux+"))";
//			}
			freeSyms.remove(itqv.ident.getSymbol());
		}
		return toBuild;
	}
	
	public String visit(MinExp exp)
	{
		String auxFunName=expTrans.getFunctionName("MinHelper");
		String comparison=lowerBound<0?"bvslt":"bvult";
		String high=generateSMT_LIB_Constant(upperBound, bitVectorLength);
		return translateMinMax(exp.exp,comparison,high,auxFunName,exp.qvar);
	}
	
	public String visit(MaxExp exp)
	{
		String auxFunName=expTrans.getFunctionName("MaxHelper");
		String comparison=lowerBound<0?"bvsgt":"bvugt";
		String low=generateSMT_LIB_Constant(lowerBound, bitVectorLength);
		return translateMinMax(exp.exp,comparison,low,auxFunName,exp.qvar);
	}
	
	public String visit(FactorialExp exp)
	{
		SMT.illegalUsage(exp, "There should be no factorial expressions when translating to SMT-LIB!");
		return null;
	}
	
	//For negative numbers the RISCAL remainder corresponds
	//to SMT-Libs bvsrem
	public String visit(RemainderExp exp)
	{
		String str1=exp.exp1.accept(this);
		String str2=exp.exp2.accept(this);
		if(lowerBound<0)
		{
			return "(bvsrem "+str1+" "+str2+")";
		}
		else
		{
			return "(bvurem "+str1+" "+str2+")";
		}
	}
	
	public String visit(EqualsExp exp)
	{
		String str1=exp.exp1.accept(this);
		String str2=exp.exp2.accept(this);
		return "(= "+str1+" "+str2+")";
	}
	
	public String visit(NotEqualsExp exp)
	{
		String str1=exp.exp1.accept(this);
		String str2=exp.exp2.accept(this);
		return "(distinct "+str1+" "+str2+")";
	}
	
	public String visit(LessExp exp)
	{
		String str1=exp.exp1.accept(this);
		String str2=exp.exp2.accept(this);
		if(this.lowerBound<0)
		{
			return "(bvslt "+str1+" "+str2+")";
		}
		else
		{
			return "(bvult "+str1+" "+str2+")";
		}
	}
	
	public String visit(LessEqualExp exp)
	{
		String str1=exp.exp1.accept(this);
		String str2=exp.exp2.accept(this);
		if(this.lowerBound<0)
		{
			return "(bvsle "+str1+" "+str2+")";
		}
		else
		{
			return "(bvule "+str1+" "+str2+")";
		}
	}
	
	public String visit(GreaterExp exp)
	{
		String str1=exp.exp1.accept(this);
		String str2=exp.exp2.accept(this);
		if(this.lowerBound<0)
		{
			return "(bvsgt "+str1+" "+str2+")";
		}
		else
		{
			return "(bvugt "+str1+" "+str2+")";
		}
	}
	
	public String visit(GreaterEqualExp exp)
	{
		String str1=exp.exp1.accept(this);
		String str2=exp.exp2.accept(this);
		if(this.lowerBound<0)
		{
			return "(bvsge "+str1+" "+str2+")";
		}
		else
		{
			return "(bvuge "+str1+" "+str2+")";
		}
	}
	
	public String visit(SetSizeExp exp)
	{
		String card=expTrans.compute(exp);
		int len=getVectorLength(exp);
		return resizeBitVector(card, len, bitVectorLength, false);
	}
	
	public String visit(IfThenElseExp exp)
	{
		String cond=expTrans.compute(exp.exp1);
		String thenExp=exp.exp2.accept(this);
		String elseExp=exp.exp3.accept(this);
		return "(ite "+cond+" "+thenExp+" "+elseExp+")";
	}
	
	public String visit(IdentifierExp exp)
	{
		int len=getVectorLength(exp);
		boolean signed=((IntType)exp.getType()).ivalue1<0;
		String s=expTrans.compute(exp);
		return resizeBitVector(s, len, bitVectorLength, signed);
	}
	
	public String visit(ApplicationExp exp)
	{
		String s=expTrans.compute(exp);
		int len=getVectorLength(exp);
		boolean signed=((IntType)exp.getType()).ivalue1<0;
		return resizeBitVector(s, len, bitVectorLength, signed);
	}
	
	
	public String visit(Binder b)
	{
		String str=expTrans.compute(b.exp);
		return "( "+b.ident.string+" "+str+")";
	}
	
	public String visit(LetParExp exp)
	{
		String str=Arrays.stream(exp.binder).map(b -> b.accept(this)).collect(Collectors.joining(" "));
		String expStr=exp.exp.accept(this);
		return "(let ("+str+") "+expStr+")";
	}
		
	public String visit(TupleSelectionExp exp)
	{
		String s=expTrans.compute(exp);
		int len=getVectorLength(exp);
		boolean signed=((IntType)exp.getType()).ivalue1<0;
		return resizeBitVector(s, len, bitVectorLength, signed);
	}
	
	public String visit(MapSelectionExp exp)
	{
		String s=expTrans.compute(exp);
		int len=getVectorLength(exp);
		boolean signed=((IntType)exp.getType()).ivalue1<0;
		return resizeBitVector(s, len, bitVectorLength, signed);
	}	
	
	public String visit(IntType it)
	{
		int len=IntTranslator.determineBitVectorLength(it.ivalue1, it.ivalue2);
		return "(_ BitVec "+Integer.toString(len)+")";
	}
	

	public String visit(LetExp exp)
	{
		SMT.illegalUsage(exp, "Let expressions should not occur in the translation to SMT-LIB!");
		return null;
	}
	
	public String visit(RecApplicationExp exp)
	{
		SMT.notImplemented(exp, "Recursive Applications with integers are not yet implemented!");
		return null;
	}
	
	public String visit(RecIdentifierExp exp)
	{
		SMT.notImplemented(exp, "Recursive Identifers with integers are not yet implemented!");
		return null;
	}
	
	public String visit(HeightExp exp)
	{
		SMT.notImplemented(exp, "Heigth expressions are not yet implemented!");
		return null;
	}
}
//---------------------------------------------------------------------------
// end of file
//---------------------------------------------------------------------------
