// ---------------------------------------------------------------------------
// TupleTranslator.java
// A class for the translation of various tuple expressions to SMT-LIB strings.
//
// Author: Franz Reichl <franz.x.reichl@gmail.com>
// $Id: TupleTranslator.java,v 1.4 2020/07/02 10:15:59 schreine Exp $
// Copyright (C) 2019-, Research Institute for Symbolic Computation (RISC)
// Johannes Kepler University, Linz, Austria, http://www.risc.jku.at
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
// ---------------------------------------------------------------------------
package riscal.smt;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.function.Function;

import riscal.semantics.Types.Value;
import riscal.semantics.Types.Value.Array;
import riscal.smt.ExpressionTranslator.TypeEnumerationInformation;
import riscal.smt.ExpressionTranslator.ValueContainer;
import riscal.syntax.AST;
import riscal.syntax.AST.Expression;
import riscal.syntax.AST.Expression.ApplicationExp;
import riscal.syntax.AST.Expression.EqualsExp;
import riscal.syntax.AST.Expression.IdentifierExp;
import riscal.syntax.AST.Expression.IfThenElseExp;
import riscal.syntax.AST.Expression.LetParExp;
import riscal.syntax.AST.Expression.MapSelectionExp;
import riscal.syntax.AST.Expression.NotEqualsExp;
import riscal.syntax.AST.Expression.TupleExp;
import riscal.syntax.AST.Expression.TupleSelectionExp;
import riscal.syntax.AST.Expression.TupleUpdateExp;
import riscal.syntax.AST.Type;
import riscal.syntax.AST.Type.BoolType;
import riscal.syntax.AST.Type.TupleType;
import riscal.syntax.ASTVisitor;
import riscal.util.Seq;
import riscal.util.Seq.Next;

public class TupleTranslator extends ASTVisitor.Base<String>
{
	/**
	 * The underlying ExpressionTranslator
	 */
	private ExpressionTranslator expTrans;
	
	/**
	 * Assigns to the already encountered TupleTypes a TupleInformation
	 */
	private HashMap<Type, TupleInformation> visitedTupleTypes=new HashMap<AST.Type, TupleTranslator.TupleInformation>();
	
	/**
	 * The basis for the generation of the name of the tuple access function
	 */
	private static final String defaultAccessName="tupleAccess";
	/**
	 * The basis for the generation of the name of the map update function
	 */
	private static final String defaultUpdateName="tupleUpdate";
	/**
	 * The basis for the generation of the name of the enumeration of a TupleType.
	 */
	private static final String defaultTypeEnumerationName="tupleEnum";
	
	/***************************************************************************
	 * Create an ExpressionTranslator with the given underlying ExpressionTranslator.
	 * @param et The underlying ExpressionTranslator.
	 **************************************************************************/
	public TupleTranslator(ExpressionTranslator et) 
	{
		this.expTrans=et;
	}
	
	/***************************************************************************
	 * Translate the given expression to SMT-LIB.
	 * @param exp The expression that shall be translated.
	 * Has to be an tuple related expression.
	 * @return The result of the translation.
	 **************************************************************************/
	public String compute(Expression exp)
	{
		if(exp.getType() instanceof TupleType)
		{		
			storeType((TupleType) exp.getType());
		}
		return exp.accept(this);
	}
	
	/***************************************************************************
	 * Translate the given expression to SMT-LIB.
	 * The expression is treated like it had the type tT.
	 * @param exp The expression that shall be translated.
	 * Has to be a tuple expression.
	 * @param tT Must contain the type of exp. Shall be canonical.
	 * @return The result of the translation.
	 **************************************************************************/
	public String compute(Expression exp,TupleType tT)
	{
		if(exp instanceof TupleExp)
		{
			return processTupleExpression((TupleExp)exp, tT);
		}
		String s=exp.accept(this);		
		storeType(tT);
		TupleType old=(TupleType)Utilities.getBaseType(exp.getType());
		s=adaptTranslation(s, old, tT);
		return s;
	}
	
	/***************************************************************************
	 * Takes a SMT-LIB expression that has the sort that corresponds
	 * to oldType. The method transforms this SMT-LIB expression to
	 * a SMT-LIB expression of the sort that corresponds to newType.
	 * Condition:
	 * oldType must be contained in newType
	 * @param s The SMT-LIB expression that shall be transformed.
	 * @param oldType The type that corresponds to the current sort.
	 * @param newType The type that corresponds to the sort the result
	 * shall have.
	 * @return A SMT-LIB expression of the sort that corresponds to
	 * newType
	 **************************************************************************/
	String adaptTranslation(String s,TupleType oldType,TupleType newType)
	{
		assert(oldType.types.length==newType.types.length);
		
		//oldType must already have been stored.
		storeType(newType);
		TupleInformation info=visitedTupleTypes.get(oldType);
		long start=0;
		long end=0;
		long prefix=0;
		String res=null;
		for(int i=0;i<oldType.types.length;i++)
		{
			end+=info.componentLengths[i];
			String part="((_ extract "+Long.toString(end-1)+" "+Long.toString(start)+") "+s+")";
			String adaptedPart=expTrans.adaptTranslation(part, oldType.types[i], newType.types[i]);
			if(!part.equals(adaptedPart))
			{
				if(prefix!=start)
				{
					adaptedPart="(concat "+adaptedPart+"((_ extract "+Long.toString(start-1)+" "+Long.toString(prefix)+")"+s+"))";
				}					
				if(res==null)
				{
					res=adaptedPart;
				}
				else
				{
					res="(concat "+adaptedPart+" "+res+")";
				}		
				prefix=end;				
			}
			start=end;
		}
		if(start!=prefix)
		{
			if(res==null)
				return s;
			res="(concat ((_ extract "+Long.toString(end-1)+" "+Long.toString(prefix)+")"+s+") "+res+")";			
		}
		return res;
	}
	
	/***************************************************************************
	 * Represent the given expression by a SMT-LIB bit vector expression with respect
	 * to the given type. 
	 * @param exp The expression of interest
	 * @param type A type capable to represent the value of the given expression.
	 * @return A bit vector representation of exps.
	 **************************************************************************/
	private String treatTupleExps(Expression exp,Type type)
	{
		String s=expTrans.compute(exp,type);
		if(exp.getType() instanceof BoolType)
		{
			return "("+expTrans.getBoolean2BitVector()+" "+s+")";
		}
		return s;
	}
	
	/***************************************************************************
	 * Compute the SMT-LIB representation of a tuple expression with
	 * respect to the given type.
	 * @param exp The TupleExp that shall be translated.
	 * @param type exp shall be treated with respect to this type.
	 * Has to be capable to represent the value of exp.
	 * @return The SMT-LIB representation of the given expression.
	 **************************************************************************/
	private String processTupleExpression(TupleExp exp,TupleType type)
	{
		storeType(type);
		String s=treatTupleExps(exp.exps[0], type.types[0]);
		for(int i=1;i<exp.exps.length;i++)
		{
			s="(concat "+treatTupleExps(exp.exps[i], type.types[i])+" "+s+")";
		}
		return s;
	}
	
	/***************************************************************************
	 * Compute the TupleInformation for the given type.
	 * This included the generation of SMT-LIB function definitions
	 * for functions representing the tuple update and access.
	 * Both the update and the access are represented by
	 * a single function, respectively.
	 * @param tT The type of interest.
	 * @return The type information associated to the given type.
	 **************************************************************************/
	private TupleInformation computeTupleInformationSingleOperation(TupleType tT)
	{
		String accessName=expTrans.getFunctionName(defaultAccessName);
		String updateName=expTrans.getFunctionName(defaultUpdateName);
		long argumentLength=IntTranslator.determineBitVectorLength(1, tT.types.length);
		long typeLength;
		long accumulateLength = 0;
		long[] components=new long[tT.types.length];
		long maxComponent=0;
		for(int i=0;i<tT.types.length;i++)
		{
			typeLength=expTrans.giveVectorLength(tT.types[i]);
			accumulateLength+=typeLength;
			components[i]=typeLength;
			maxComponent=typeLength>maxComponent?typeLength:maxComponent;
		}
		TupleInformation info=new TupleInformation(accessName, updateName, accumulateLength, argumentLength,maxComponent, components);
		return info;		
	}
	
	/***************************************************************************
	 * Compute the TupleInformation for the given type.
	 * This included the generation of SMT-LIB function definitions
	 * for functions representing the tuple update and access
	 * For both the update and the access there is
	 * a single function for each component, respectively.
	 * @param tT The type of interest.
	 * @return The type information associated to the given type.
	 **************************************************************************/
	private TupleInformation computeTupleInformationMultipleOperations(TupleType tT)
	{
		int len=tT.types.length;
		String[] accessNames=new String[len];
		String[] updateNames=new String[len];
		long[] components=new long[len];
		long accumulateLength = 0;
		for(int i=0;i<len;i++)
		{
			long typeLength=expTrans.giveVectorLength(tT.types[i]);
			components[i]=typeLength;
			accumulateLength+=typeLength;
			accessNames[i]=expTrans.getFunctionName(defaultAccessName+"_component"+Integer.toString(i+1));
			updateNames[i]=expTrans.getFunctionName(defaultUpdateName+"_component"+Integer.toString(i+1));
		}
		return new TupleInformation(accessNames, updateNames, accumulateLength, components);
	}
	
	/***************************************************************************
	 * If not yet computed generate the TupleInformation for the
	 * given type. Represent update and access by single functions iff
	 * SMT.treatTupleOperationsAsSingleOperations is true.
	 * @param tT The type of interest.
	 **************************************************************************/
	private void storeType(TupleType tT)
	{
		TupleInformation info=visitedTupleTypes.get(tT);
		if(info!=null)
			return;
		if(SMT.treatTupleOperationsAsSingleOperations)
		{
			storeTypeSingleOperation(tT);
		}
		else
		{
			storeTypeMultipleOperations(tT);
		}
	}
	
	/***************************************************************************
	 * Compute the TupleInformation for the given type.
	 * Represent update and access by a single function, respectively.
	 * given type.
	 * @param tT The type of interest.
	 **************************************************************************/
	private void storeTypeSingleOperation(TupleType tT)
	{
		TupleInformation info=computeTupleInformationSingleOperation(tT);
		visitedTupleTypes.put(tT, info);	
		String accessName=info.accessName,updateName=info.updateName;
		long maxComponent=info.maximalComponentLength,argumentLength=info.accessVectorLength,accumulateLength=info.vectorLength;
		long[] components=info.componentLengths;	
		String variable="i";
		String vectorVariable="v";
		String newValue="val";
		String accDefStr,upDefStr;
		if(tT.types.length==1)
		{
			accDefStr=vectorVariable;
			upDefStr=newValue;
		}
		else
		{
			accDefStr="(ite (= "+variable+" "+IntTranslator.generateSMT_LIB_Constant(1, argumentLength)+")((_ extract "
			+Long.toString(maxComponent-1)+" 0)"+vectorVariable+")"+IntTranslator.generateSMT_LIB_Constant(0, maxComponent)+")";
			upDefStr="((_ extract "+Long.toString(components[0]-1)+" 0)"+newValue+")";
			upDefStr="(concat ((_ extract "+Long.toString(accumulateLength-1)+" "+Long.toString(components[0])+")"+vectorVariable+")"+upDefStr+")";
			upDefStr="(ite (= "+variable+" "+IntTranslator.generateSMT_LIB_Constant(1, argumentLength)+")"+upDefStr+
					" "+IntTranslator.generateSMT_LIB_Constant(0, accumulateLength)+")";
			long counter=components[0];
			for(int i=1;i<tT.types.length;i++)
			{
				long refillLen=maxComponent-components[i];
				String aux="((_ extract "+Long.toString(counter+components[i]-1)+" "+Long.toString(counter)+")"+vectorVariable+")";
				if(refillLen>0)
				{
					aux="((_ zero_extend "+Long.toString(refillLen)+")"+aux+")";
				}
				accDefStr="(ite (= "+variable+" "+IntTranslator.generateSMT_LIB_Constant(i+1, argumentLength)+")"+aux+" "+accDefStr+")";
				
				aux="((_ extract "+Long.toString(components[i]-1)+" 0)"+newValue+")";
				aux="(concat "+aux+"((_ extract "+Long.toString(counter-1)+" 0)"+vectorVariable+"))";		
				counter+=components[i];
				if(i<tT.types.length-1)
				{
					aux="(concat ((_ extract "+Long.toString(accumulateLength-1)+" "+Long.toString(counter)+")"+vectorVariable+")"+aux+")";
				}			
				upDefStr="(ite (= "+variable+" "+IntTranslator.generateSMT_LIB_Constant(i+1, argumentLength)+")"+aux+" "+upDefStr+")";			
			}
		}
		expTrans.addDeclaration("(define-fun "+accessName+" (("+vectorVariable+"(_ BitVec "+accumulateLength+
				"))("+variable+"(_ BitVec "+argumentLength+"))) (_ BitVec "+maxComponent+")"+accDefStr+")");
		expTrans.addDeclaration("(define-fun "+updateName+" (("+vectorVariable+"(_ BitVec "+accumulateLength+
				"))("+variable+"(_ BitVec "+argumentLength+"))("+newValue+"(_ BitVec "+maxComponent+"))) (_ BitVec "+accumulateLength+")"+upDefStr+")");
	}
	
	/***************************************************************************
	 * Compute the TupleInformation for the given type.
	 * Update and access are represented by single functions for
	 * the individual components, respectively.
	 * @param tT The type of interest.
	 **************************************************************************/
	private void storeTypeMultipleOperations(TupleType tT)
	{
		TupleInformation info=computeTupleInformationMultipleOperations(tT);
		visitedTupleTypes.put(tT, info);
		String[] acc=info.accessNames;
		String[] upd=info.updateNames;
		long vecLen=info.vectorLength;
		long[] components=info.componentLengths;
		int numComps=tT.types.length;
		long lowerBound=0;
		long upperBound=0;
		String vectorVariable="v";
		String newValue="val";
		String newValueBool="("+expTrans.getBoolean2BitVector()+" "+newValue+")";
		for(int i=0;i<numComps;i++)
		{
			upperBound=lowerBound+components[i]-1;
			String accDefStr="((_ extract "+Long.toString(upperBound)+" "+Long.toString(lowerBound)+")"+vectorVariable+")";
			if(Utilities.getBaseType(tT.types[i])instanceof BoolType)
			{
				accDefStr="("+expTrans.getBitVector2boolean()+" "+accDefStr+")";
				accDefStr="(define-fun "+acc[i]+" (("+vectorVariable+"(_ BitVec "+vecLen+"))) Bool "+accDefStr+")";
			}
			else
			{
				accDefStr="(define-fun "+acc[i]+" (("+vectorVariable+"(_ BitVec "+vecLen+")))(_ BitVec "+Long.toString(components[i])+")"+accDefStr+")";
			}
			expTrans.addDeclaration(accDefStr);
			lowerBound+=components[i];
		}
		lowerBound=components[0];
		upperBound=0;
		String aux1;
		String upDefStr;
		if(numComps==1)
		{
			if(Utilities.getBaseType(tT.types[0])instanceof BoolType)
			{
				upDefStr="(define-fun "+upd[0]+" (("+vectorVariable+"(_ BitVec "+vecLen+"))("+newValue+" Bool))(_ BitVec "+vecLen+")"+newValueBool+")";
			}
			else
			{
				upDefStr="(define-fun "+upd[0]+" (("+vectorVariable+"(_ BitVec "+vecLen+"))("+newValue+"(_ BitVec "+components[0]+")))(_ BitVec "+vecLen+")"+newValue+")";
			}			
			expTrans.addDeclaration(upDefStr);
		}
		else
		{
			aux1="((_ extract "+Long.toString(vecLen-1)+" "+Long.toString(lowerBound)+")"+vectorVariable+")";
			if(Utilities.getBaseType(tT.types[0])instanceof BoolType)
			{
				upDefStr="(concat "+aux1+" "+newValueBool+")";
				upDefStr="(define-fun "+upd[0]+" (("+vectorVariable+"(_ BitVec "+vecLen+"))("+newValue+" Bool))(_ BitVec "+vecLen+")"+upDefStr+")";
			}
			else
			{
				upDefStr="(concat "+aux1+" "+newValue+")";
				upDefStr="(define-fun "+upd[0]+" (("+vectorVariable+"(_ BitVec "+vecLen+"))("+newValue+"(_ BitVec "+components[0]+")))(_ BitVec "+vecLen+")"+upDefStr+")";
			}
			expTrans.addDeclaration(upDefStr);
		}
		for(int i=1;i<numComps-1;i++)
		{
			upperBound=lowerBound+components[i];
			aux1="((_ extract "+Long.toString(lowerBound-1)+" "+Long.toString(0)+")"+vectorVariable+")";
			String aux2="((_ extract "+Long.toString(vecLen-1)+" "+Long.toString(upperBound)+")"+vectorVariable+")";
			if(Utilities.getBaseType(tT.types[i])instanceof BoolType)
			{
				upDefStr="(concat "+aux2+"(concat "+newValueBool+" "+aux1+"))";
				upDefStr="(define-fun "+upd[i]+" (("+vectorVariable+"(_ BitVec "+vecLen+"))("+newValue+" Bool))(_ BitVec "+vecLen+")"+upDefStr+")";
			}
			else
			{
				upDefStr="(concat "+aux2+"(concat "+newValue+" "+aux1+"))";
				upDefStr="(define-fun "+upd[i]+" (("+vectorVariable+"(_ BitVec "+vecLen+"))("+newValue+"(_ BitVec "+components[i]+")))(_ BitVec "+vecLen+")"+upDefStr+")";
			}
			expTrans.addDeclaration(upDefStr);
			lowerBound+=components[i];
		}
		if(numComps>1)
		{
			aux1="((_ extract "+Long.toString(lowerBound-1)+" "+Long.toString(0)+")"+vectorVariable+")";
			if(Utilities.getBaseType(tT.types[numComps-1])instanceof BoolType)
			{
				upDefStr="(concat "+newValueBool+" "+aux1+")";
				upDefStr="(define-fun "+upd[numComps-1]+" (("+vectorVariable+"(_ BitVec "+vecLen+"))("+newValue+" Bool))(_ BitVec "+vecLen+")"+upDefStr+")";
			}
			else
			{
				upDefStr="(concat "+newValue+" "+aux1+")";
				upDefStr="(define-fun "+upd[numComps-1]+" (("+vectorVariable+"(_ BitVec "+vecLen+"))("+newValue+"(_ BitVec "+components[numComps-1]+")))(_ BitVec "+vecLen+")"+upDefStr+")";
			}
			expTrans.addDeclaration(upDefStr);
		}
	}
		
	/***************************************************************************
	 * Compute the SMT-LIB representations of the values of the given type.
	 * Add the #b prefix to these representations depending on bare.
	 * @param ty The type whose values shall be computed. Has to be canonical.
	 * @param bare If true do not add the #b prefix to the resulting values.
	 * @return A list containing all values of the given type 
	 * represented by their corresponding SMT-LIB strings.
	 **************************************************************************/
	public List<String> computeValues(TupleType ty,boolean bare)
	{
		storeType(ty);
		int numComps=ty.types.length;
		int typeSize=ExpressionTranslator.getTypeSize(ty);
		List<String> tupleVals=new ArrayList<String>(typeSize);
		List<List<String>> components = new ArrayList<List<String>>(numComps);
		for(int i=0;i<numComps;i++)
		{
			components.add(expTrans.computeValuesBare(ty.types[i]));
		}			
		//idea based on the generation of values in Values.java
		int[] counter=new int[numComps];
		for(int i=0;i<typeSize;i++)
		{
			int carry = 1;
			for(int j=0;j<numComps;j++)
			{
				counter[j] = counter[j] + carry;
				if(counter[j]==components.get(j).size())
				{
					counter[j]=0;
					carry = 1;
				}
				else
					carry = 0;
			}
			StringBuilder strB=new StringBuilder();
			if(!bare)
				strB.append("#b");
			for(int j=numComps-1;j>=0;j--)
			{
				strB.append(components.get(j).get(counter[j]));
			}
			tupleVals.add(strB.toString());
		}
		return tupleVals;
	}
	
	/***************************************************************************
	 * Compute the bit vector length that is necessary to represent the values
	 * of the given type.
	 * @param type The type of interest. Must be canonical.
	 * @return The necessary length to represent the values of type.
	 **************************************************************************/
	public long giveVectorLength(TupleType type)
	{
		long len=0;
		for(Type t:type.types)
		{
			len+=expTrans.giveVectorLength(t);
		}
		return len;
	}
	
	/***************************************************************************
	 * Translate a sequence of values with base type type to a list of SMT-LIB values.
	 * @param type The base type of the values. Must be canonical.
	 * @param seq A sequence of values whose basetype is type.
	 * @param bare If true do not add the #b prefix to the resulting values.
	 * @return A list containing the SMT-LIB representations of the
	 * values in seq.
	 **************************************************************************/
	List<String> processValueSequence(Type type,Seq<Value> seq,boolean bare)
	{
		type=Utilities.getBaseType(type);
		TupleType tT=(TupleType)type;
		storeType(tT);
		ValueContainer[] valCs=new ValueContainer[tT.types.length];
		int len=ExpressionTranslator.getTypeSize(type);
		List<String> result=new ArrayList<String>(len);
		for(int i=0;i<tT.types.length;i++)
		{
			valCs[i]=ExpressionTranslator.ValueContainerGenerator.getValueContainer(expTrans,tT.types[i]);
		}
		Function <Value, String> proc;
		if(bare)
			proc=v -> processValueBareVector(v,valCs);
		else
			proc=v -> processValue(v,valCs);
		while (true)
    {
      Next<Value> next = seq.get();
      if (next == null) break;
      result.add(proc.apply(next.head));
      seq = next.tail;
    }
		return result;
	}
	
	/***************************************************************************
	 * Translate the given value to a bit sequence
	 * @param val A map value that shall be translated.
	 * @param valCs ValueContainers for the types of the
	 * components of the tuple
	 * @return A bit sequence representing val.
	 **************************************************************************/
	private static String processValueBareVector(Value val,ValueContainer[] valCs)
	{
		Array v=(Array)val;
		Iterator<Value> it=v.iterator();
		int i=0;
		String str=valCs[i++].processToBareVector(it.next());
		while(it.hasNext())
		{
			str=valCs[i++].processToBareVector(it.next())+str;
		}
		return str;
	}
	
	/***************************************************************************
	 * Translate the given value to a SMT-LIB bit vector expression.
	 * @param val A map value that shall be translated.
	 * @param valCs ValueContainers for the types of the
	 * components of the tuple
	 * @return A SMT-LIB bit vector expression.
	 **************************************************************************/
	private static String processValue(Value val,ValueContainer[] valCs)
	{
		return "#b"+processValueBareVector(val,valCs);
	}
	
	static class TupleValueContainer extends ExpressionTranslator.ValueContainer
	{
		/**
		 * ValueContainers for processing the components of the respective type.
		 */
		private ValueContainer[] valCs;
		TupleValueContainer(ExpressionTranslator expTrans,TupleTranslator tTranslator,TupleType tT)
		{
			tTranslator.storeType(tT);
			valCs=new ValueContainer[tT.types.length];
			for(int i=0;i<tT.types.length;i++)
			{
				valCs[i]=ExpressionTranslator.ValueContainerGenerator.getValueContainer(expTrans,tT.types[i]);
			}				
		}
		/***************************************************************************
		 * Generate a bit vector constant representing the given value.
		 * @param val The value that shall be represented by a bit vector.
		 * @return The SMT-LIB representation of val
		 **************************************************************************/
		String process(Value val) 
		{
			return processValue(val,valCs);
		}
		/***************************************************************************
		 * Generate a bit sequence representing the given value.
		 * @param val The value that shall be represented by a bit vector.
		 * @return A bit sequence representing val.
		 **************************************************************************/
		String processToBareVector(Value val)
		{
			return processValueBareVector(val,valCs);
		}
	}
		
	/***************************************************************************
	 * Generate an enumeration function for the given type and return the
	 * associated information.
	 * @param tT The type for which an enumeration function shall be generated.
	 * @return The information associated to the enumeration function.
	 **************************************************************************/
	public TypeEnumerationInformation generateTypeEnumerationFunction(TupleType tT)
	{
		String funName=expTrans.getFunctionName(defaultTypeEnumerationName);
		storeType(tT);
		TupleInformation info=visitedTupleTypes.get(tT);
		long typeSize=ExpressionTranslator.getTypeSize(tT);
		int resultVectorLength=IntTranslator.determineBitVectorLength(0, typeSize-1);	
		String variable="x";
		//A Tuple with only unit components would make problems if not treated separately
		if(typeSize==1)
		{
			String def="(define-fun "+funName+" (("+variable+" (_ BitVec "+Long.toString(info.vectorLength)+")))"
					+ "(_ BitVec "+Long.toString(resultVectorLength)+") "+
					IntTranslator.generateSMT_LIB_Constant(0, resultVectorLength)+")";
			expTrans.addDeclaration(def);
			return new TypeEnumerationInformation(funName, resultVectorLength, typeSize);
		}
		long accumulatedSize=1;
		List<String> slst=new ArrayList<String>(tT.types.length);
		for(int i=0;i<tT.types.length;i++)
		{
			TypeEnumerationInformation enumInfo=expTrans.getTypeEnumerationFunction(tT.types[i]);
			if(enumInfo.numberOfElements==1)
				continue;
			int diff=resultVectorLength-enumInfo.enumerationFunctionResultVectorLength;
			if(diff<0)
				SMT.illegalUsage(null, "Enumeration of Tuples - erroneous vector length of components");
			String s="("+enumInfo.functionName+" ("+info.accessNames[i]+" "+variable+"))";
			s="((_ zero_extend "+Integer.toString(diff)+") "+s+")";
			if(accumulatedSize>1)
				s="(bvmul "+IntTranslator.generateSMT_LIB_Constant(accumulatedSize, resultVectorLength)+" "+s+")";
			slst.add(s);
			accumulatedSize*=enumInfo.numberOfElements;
		}
		String args=String.join(" ", slst);
		if(tT.types.length>1)
			args="(bvadd "+args+")";
		String def="(define-fun "+funName+" (("+variable+" (_ BitVec "+Long.toString(info.vectorLength)+")))"
				+ "(_ BitVec "+Long.toString(resultVectorLength)+") "+args+")";
		expTrans.addDeclaration(def);
		return new TypeEnumerationInformation(funName, resultVectorLength, typeSize);
	}
	
	/***************************************************************************
	 * Generates an enumeration function for the values given by lst.
	 * The result of the enumeration is given by the position 
	 * of the respective value in the list.
	 * The enumeration function is defined via a table.
	 * @param lst The values used to generate the enumeration function.
	 * @param tT The elements of lst must correspond to the type tT. Must be canonical.
	 * @return The information associated to the enumeration function.
	 **************************************************************************/
	public TypeEnumerationInformation generateTypeEnumerationFunctionUseTable(List<Value> lst,TupleType tT)
	{
		String funName=expTrans.getFunctionName(defaultTypeEnumerationName);
		storeType(tT);
		TupleInformation info=visitedTupleTypes.get(tT);
		long len=lst.size();
		long typeVectorLength=info.vectorLength;
		int argumentVectorLength=IntTranslator.determineBitVectorLength(0, len-1);
		expTrans.addDeclaration("(declare-fun "+funName+" ((_ BitVec "+Long.toString(typeVectorLength)+")) "
				+ "(_ BitVec "+Long.toString(argumentVectorLength)+"))");
		ValueContainer[] valCs=new ValueContainer[tT.types.length];
		for(int i=0;i<tT.types.length;i++)
		{
			valCs[i]=ExpressionTranslator.ValueContainerGenerator.getValueContainer(expTrans,tT.types[i]);
		}
		long i=0;		
		for(Value val:lst)
		{
			expTrans.addDeclaration("(assert (= ("+funName+" "+processValue(val,valCs)+") "+
					IntTranslator.generateSMT_LIB_Constant(i, argumentVectorLength)+"))");
			i++;
		}		
		return new TypeEnumerationInformation(funName, argumentVectorLength, len);
	}
	
	public TypeEnumerationInformation generateTypeEnumerationFunctionUseTableWithFunctionDefinition(List<Value> lst,TupleType tT)
	{
		String funName=expTrans.getFunctionName(defaultTypeEnumerationName);
		storeType(tT);
		TupleInformation info=visitedTupleTypes.get(tT);
		long len=lst.size();
		long typeVectorLength=info.vectorLength;
		int argumentVectorLength=IntTranslator.determineBitVectorLength(0, len-1);
		ValueContainer[] valCs=new ValueContainer[tT.types.length];
		for(int i=0;i<tT.types.length;i++)
		{
			valCs[i]=ExpressionTranslator.ValueContainerGenerator.getValueContainer(expTrans,tT.types[i]);
		}
		Iterator<Value> it=lst.iterator();
		StringBuilder strB=new StringBuilder();
		String variable="x";
		for(int i=0;i<len-1;i++)
		{
			Value val=it.next();
			strB.append("(ite (= "+variable+" "+processValue(val,valCs)+")"+
					IntTranslator.generateSMT_LIB_Constant(i, argumentVectorLength)+" ");
		}
		strB.append(IntTranslator.generateSMT_LIB_Constant(len-1, argumentVectorLength));
		strB.append(")".repeat((int) (len-1)));
		String def="(define-fun "+funName+" (("+variable+" (_ BitVec "+Long.toString(typeVectorLength)+"))) "
				+ "(_ BitVec "+Long.toString(argumentVectorLength)+") "+strB.toString()+")";
		expTrans.addDeclaration(def);
		return new TypeEnumerationInformation(funName, argumentVectorLength, len);
	}

	/***************************************************************************
	 * Generate the SMT-LIB representation of a tuple access expression.
	 * @param tuple The SMT-LIB representation of the tuple that shall
	 * be accessed.
	 * @param type The type of the original tuple expression.
	 * @param position The component that shall be accessed.
	 * @return A SMT-LIB representation for a tuple access.
	 **************************************************************************/
	private String generateTupleAccess(String tuple,TupleType type,int position)
	{		
		storeType(type);
		if(SMT.treatTupleOperationsAsSingleOperations)
		{
			return generateTupleAccessSingleOperation(tuple, type, position);
		}
		else
		{
			return generateTupleAccessMultipleOperations(tuple, type, position);
		}
	}
	
	/***************************************************************************
	 * Generate the SMT-LIB representation of a tuple access expression.
	 * Make use of the access representation by a single function.
	 * @param tuple The SMT-LIB representation of the tuple that shall
	 * be accessed.
	 * @param type The type of the original tuple expression.
	 * @param position The component that shall be accessed.
	 * @return A SMT-LIB representation for a tuple access.
	 **************************************************************************/
	private String generateTupleAccessSingleOperation(String tuple,TupleType type,int position)
	{	
		int index=position-1;
		TupleInformation info=visitedTupleTypes.get(type);
		String s="("+info.accessName+" "+tuple+" "+IntTranslator.generateSMT_LIB_Constant(position, info.accessVectorLength)+")";
		if(info.maximalComponentLength!=info.componentLengths[index])
		{
			s="((_ extract "+Long.toString(info.componentLengths[index]-1)+" 0)"+s+")";
		}
		if(Utilities.getBaseType(type.types[index]) instanceof BoolType)
		{
			s="("+expTrans.getBitVector2boolean()+" "+s+")";
		}
		return s;
	}
	
	/***************************************************************************
	 * Generate the SMT-LIB representation of a tuple access expression.
	 * Make use of the access representation by multiple functions.
	 * @param tuple The SMT-LIB representation of the tuple that shall
	 * be accessed.
	 * @param type The type of the original tuple expression.
	 * @param position The component that shall be accessed.
	 * @return A SMT-LIB representation for a tuple access.
	 **************************************************************************/
	private String generateTupleAccessMultipleOperations(String tuple,TupleType type,int position)
	{	
		int index=position-1;
		TupleInformation info=visitedTupleTypes.get(type);
		String s="("+info.accessNames[index]+" "+tuple+")";
		return s;
	}
	
	/***************************************************************************
	 * Generate the SMT-LIB representation of a tuple update expression.
	 * @param tuple The SMT-LIB representation of the tuple that shall
	 * be updated.
	 * @param newValue The expression that shall be used for the update.
	 * @param position The component that shall be updated.
	 * @return A SMT-LIB representation for a tuple update.
	 **************************************************************************/
	private String generateTupleUpdate(Expression tuple,Expression newValue,int position)
	{
		int index=position-1;
		TupleType tT=(TupleType)tuple.getType();
		storeType(tT);
		Type t=tT.types[index];
		String newVal=expTrans.compute(newValue, t);
		String tupleStr=compute(tuple);
		if(SMT.treatTupleOperationsAsSingleOperations)
		{
			if(Utilities.getBaseType(t) instanceof BoolType)
			{
				newVal="("+expTrans.getBoolean2BitVector()+" "+newVal+")";
			}
			return generateTupleUpdateSingleOperation(tupleStr,tT, newVal, index);
		}
		else
		{
			return generateTupleUpdateMultipleOperations(tupleStr,tT, newVal, index);
		}
	}
	
	/***************************************************************************
	 * Generate the SMT-LIB representation of a tuple update expression.
	 * Make use of the update representation by a single function.
	 * @param tupleStr The SMT-LIB representation of the tuple that shall
	 * be updated.
	 * @param tT The type of the original tuple expression.
	 * @param newValStr The SMT-LIB representation of the expression that
	 * shall be used for the update
	 * @param index The component that shall be updated.
	 * @return A SMT-LIB representation for a tuple update.
	 **************************************************************************/
	private String generateTupleUpdateSingleOperation(String tupleStr,TupleType tT,String newValStr,int index)
	{
		TupleInformation info=visitedTupleTypes.get(tT);
		long diff=info.maximalComponentLength-expTrans.giveVectorLength(tT.types[index]);
		assert(diff>=0);//the type of newValue cannot be larger then tT.types[index] and maximalComponentLength is the maximal length of the components
		String s;
		if(diff!=0)
		{
			s="((_ zero_extend "+Long.toString(diff)+")"+newValStr+")";
		}
		else
		{
			s=newValStr;
		}
		return "("+info.updateName+" "+tupleStr+" "+IntTranslator.generateSMT_LIB_Constant(index+1, info.accessVectorLength)+" "+s+")";
	}
	
	/***************************************************************************
	 * Generate the SMT-LIB representation of a tuple update expression.
	 * Make use of the update representation by by multiple functions.
	 * @param tupleStr The SMT-LIB representation of the tuple that shall
	 * be updated.
	 * @param tT The type of the original tuple expression.
	 * @param newValue The SMT-LIB representation of the expression that
	 * shall be used for the update
	 * @param index The component that shall be updated.
	 * @return A SMT-LIB representation for a tuple update.
	 **************************************************************************/
	private String generateTupleUpdateMultipleOperations(String tupleStr,TupleType tT,String newValue,int index)
	{
		TupleInformation info=visitedTupleTypes.get(tT);
		return "("+info.updateNames[index]+" "+tupleStr+" "+newValue+")";
	}
	
	/***************************************************************************
	 * Generates a boolean valued SMT-LIB expression that is true
	 * iff the SMT-LIB expression represents a bit vector value
	 * that represents a value of the given TupleType.
	 * @param ty The type of interest.
	 * @param str A bit vector expression.
	 * @return A boolean valued SMT-LIB expression.
	 **************************************************************************/
	public String generateTypeConditions(TupleType ty,String str)
	{
		List<String> lst=new ArrayList<String>();
		storeType(ty);
		for(int i=1;i<=ty.types.length;i++)
		{
			String s=generateTupleAccess(str, ty, i);
			s=expTrans.generateTypeConditions(ty.types[i-1], s);
			if(s!=null)
				lst.add(s);
		}
		if(lst.size()==0)
			return null;
		else if(lst.size()==1)//Boolector does not support and with a single argument
		{
			return String.join(" ", lst);
		}
		else
		{
			return "(and "+String.join(" ", lst)+")";
		}		
	}
	
	public String visit(TupleExp exp)
	{
		storeType((TupleType) exp.getType());
		Type[] types=((TupleType)exp.getType()).types;
		String s=treatTupleExps(exp.exps[0], types[0]);
		for(int i=1;i<exp.exps.length;i++)
		{
			s="(concat "+treatTupleExps(exp.exps[i], types[i])+" "+s+")";
		}
		return s;
	}
	
	public String visit(TupleUpdateExp exp)
	{
		int index=exp.index.getInteger().intValue();
		return generateTupleUpdate(exp.exp1, exp.exp2, index);
	}
	
	public String visit(TupleSelectionExp exp)
	{
		int index=exp.index.getInteger().intValue();
		String tupleStr=compute(exp.exp);
		return generateTupleAccess(tupleStr, (TupleType)exp.exp.getType(), index);
	}
	
	public String visit(EqualsExp exp)
	{
		Utilities.Pair<String, String> p=expTrans.computeUseSameType(exp.exp1, exp.exp2);
		return "(= "+p.first+" "+p.second+")";
	}
	
	public String visit(NotEqualsExp exp)
	{
		Utilities.Pair<String, String> p=expTrans.computeUseSameType(exp.exp1, exp.exp2);
		return "(distinct "+p.first+" "+p.second+")";
	}
	
//	public String visit(IfThenElseExp exp)
//	{
//		String cond=expTrans.compute(exp.exp1);
//		String thenExp=exp.exp2.accept(this);
//		String elseExp=exp.exp3.accept(this);
//		return "(ite "+cond+" "+thenExp+" "+elseExp+")";
//	}	
	
	public String visit(IfThenElseExp exp)
	{
		return expTrans.compute(exp);
	}
	
	public String visit(ApplicationExp exp)
	{
		return expTrans.compute(exp);	
	}
	
	public String visit(IdentifierExp exp)
	{
		return expTrans.compute(exp);
	}
	
	public String visit(LetParExp exp)
	{
		return expTrans.compute(exp);
	}
	
	public String visit(MapSelectionExp exp)
	{
		return expTrans.compute(exp);	
	}
	
	public String visit(TupleType tT)
	{
		assert(tT.isCanonical());
		storeType(tT);
		TupleInformation info=visitedTupleTypes.get(tT);		
		return "(_ BitVec "+Long.toString(info.vectorLength)+")";
	}

	private class TupleInformation
	{
		public String accessName;
		public String updateName;
		public String[] accessNames;
		public String[] updateNames;
		public long vectorLength;
		public long accessVectorLength;//the bitvectorlength needed to store numbers from 1 to type.type.length
		public long maximalComponentLength;
		public long[] componentLengths;
		public TupleInformation(String ac,String up,long vecLen,long accLen,long upLen,long[] compLeng)
		{
			accessName=ac;
			updateName=up;
			vectorLength=vecLen;
			accessVectorLength=accLen;
			maximalComponentLength=upLen;
			componentLengths=compLeng;
			accessNames=null;
			updateName=null;
		}
		public TupleInformation(String[] ac,String[] up,long vecLen,long[] compLen)
		{
			accessNames=ac;
			updateNames=up;
			vectorLength=vecLen;
			componentLengths=compLen;
		}
	}	
}
//---------------------------------------------------------------------------
// end of file
//---------------------------------------------------------------------------
