// ---------------------------------------------------------------------------
// Choices.java
// Replace choices from expressions (as far as possible)
// $Id: Choices.java,v 1.20 2021/05/06 16:01:33 schreine Exp schreine $
//
// Author: Wolfgang Schreiner <Wolfgang.Schreiner@risc.jku.at>
// Copyright (C) 2018-, Research Institute for Symbolic Computation (RISC)
// Johannes Kepler University, Linz, Austria, http://www.risc.jku.at
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
// ----------------------------------------------------------------------------
package riscal.tasks;

import java.util.*;

import riscal.syntax.*;
import riscal.syntax.AST.*;
import riscal.syntax.AST.Expression.*;
import riscal.types.Environment.Symbol.*;

public class Choices extends ASTCloner
{
  // this is the new version of choice elimination intended for SMT evaluation
  // (no consideration of evaluation efficiency); by eliminating choices
  // we avoid the generation of uninterpreted functions and their axiomatization
  // which affects the efficiency of most SMT solvers very negatively
  //
  // NOTE: this elimination is only sound under the assumption that the
  // specified choice can be indeed made: we generate for a choice
  // (choose x:T with F(x,y)) the constraint F(x,y) rather than
  // (exists x:T with F(x,y)) => F(x,y); this assumption is in particular true 
  // if the choice arises from the contract of a (previously verified)
  // function; it is the responsibility of the user to apply this option.
  //
  // the core idea is to pull out of a expressions chosen variables x1,x2,x3,...
  // as a sequence of quantified variables with constraints
  //
  //   x1:T1 with F(a,x1), x2:T2 with F(a,x1,x2), x3:T3 with F(a,x1,x2,x3), ...
  //
  // replacing the choice expressions by the variables x1,x2,x3,...; if the
  // name of a chosen variable xi is already visible in the current context,
  // it is renamed to a fresh variable xi, the corresponding constraint is 
  // instantiated as "let xi=yi in F(...,xi)";
  //
  // the pulled out quantified variables are added to the next enclosing 
  // quantified expression (respectively are closed by universal quantification 
  // on the outermost level of a theorem to be decided by the SMT solver)
  //
  // to avoid the blocking of this process by variable binders, we also
  // treat a binding expression "let x=T in E" as a quantified expression
  // that introduces quantified variable "x:D with x=y" which is pulled to
  // the next level of quantification level (only if a choice is made in E)
  //
  //   q(x, choose y with p1(x,y), choose y with p2(x,y))
  //   -> forall y with p1(x,y), y0 with (let y=y0 in p2(x,y)). q(x,x,x0)
  //
  //   exists x with p(x). q(x,choose y with p1(x,y), choose y with p2(x,y))
  //   -> exists x with p(x), y with p1(a,y), y0 with let y = y0 in p2(x,y).
  //        q(x,y,y0)
  //
  //   q(max x. choose y with p(x,y)) -> q(max x,y with p(x,y). y)
  // 

  private static class ChooseSearch extends ASTVisitor.Base<Void>
  {
    public static boolean execute(Expression exp)
    {
      ChooseSearch search = new ChooseSearch();
      exp.accept(search);
      return search.result;
    }
    private boolean result = false;
    public Void visit(ChooseExp exp) { result = true; return null; }
    public Void visit(ChooseInExp exp) { result = true; return null; }
  }
  
  /***************************************************************************
   * Remove choices in a formula (theorem) as far as possible.
   * @param exp the formula
   * @return the the formula with a lot of choices removed.
   **************************************************************************/
  public static Expression optimizeFormula(Expression exp)
  {
    Set<String> visible = FreeVariables.compute(exp);
    return optimizeFormula(exp, visible);
  }
  
  /***************************************************************************
   * Remove choices in a formula (theorem) as far as possible.
   * @param exp the formula
   * @param visibile the set of variables visible in the formula
   * @return the the formula with a lot of choices removed.
   **************************************************************************/
  private static Expression optimizeFormula(Expression exp, Set<String> visible)
  {
    Choices choices = new Choices(visible);
    Expression exp0 = choices.process(exp);
    if (!choices.vars.isEmpty())
      exp0 = new ForallExp(new QuantifiedVariable(choices.vars, null), exp0); 
    return exp0;
  }
  
  /***************************************************************************
   * Remove choices from expression.
   * @param exp the expression.
   * @return the processed result.
   **************************************************************************/
  private Expression process(Expression exp)
  {
    if (!ChooseSearch.execute(exp)) return exp;
    return (Expression)exp.accept(this);
  }
  
  // the set of currently visible variable names;
  private Set<String> visible;
  
  // the list of extracted variables (quantified variable cores with constraints)
  private List<QuantifiedVariableCore> vars;
  
  /***************************************************************************
   * Create object for removing choices.
   * @param visible the names of the currently visible variables.
   **************************************************************************/
  private Choices(Set<String> visible)
  {
    // also clone expression types
    super(true);
    this.visible = visible;
    this.vars = new ArrayList<QuantifiedVariableCore>();
  }
  
  // we have to consider all quantified expressions
  // - choices: ChooseExp, ChooseInExp, ChooseInElseExp
  // - formulas: ForallExp, ExistsExp
  // - terms: MinExp, MaxExp, NumberExp, SumExp, ProductExp, SetBuilderExp
  // - binders: LetExp, LetParExp
  // patterns (MatchExp) is currently not supported by SMT solver, thus ignored
 
  /***************************************************************************
   * Reset list of quantified variables.
   * @return the old variables.
   *************************************************************************/
  private List<QuantifiedVariableCore> resetVariables()
  {
    List<QuantifiedVariableCore> old = vars;
    vars = new ArrayList<QuantifiedVariableCore>();
    return old;
  }
  
  /***************************************************************************
   * Get (cloned) set of visible variables.
   * @return the old variables.
   *************************************************************************/
  private Set<String> getVisible()
  {
    Set<String> old = visible;
    visible = new HashSet<String>();
    return old;
  }
  
  /***************************************************************************
   * Get quantified variable from generated variable cores;
   * resets the generated variables to the old list and the
   * set of visible variables to the old set.
   * @param ovars the old list of quantified variables.
   * @param oset the old set of visible variables.
   * @return the new variable
   *************************************************************************/
  private QuantifiedVariable getVariable(List<QuantifiedVariableCore> ovars,
    Set<String> oset)
  {
    QuantifiedVariable result = new QuantifiedVariable(vars, null);
    vars = ovars;
    visible = oset;
    return result;
  }
  
  /**************************************************************************
   * Visit an expression.
   * @param exp the expression
   * @return a version of exp with choices minimized
   **************************************************************************/
  
  public Identifier visit(Identifier id)
  {
    // do not clone identifier, preserve its symbol
    return id;
  }
  
  public Expression visit(ChooseExp exp)
  {
    return quantifiedExpression(exp.qvar, null);
  }
  
  public Expression visit(ChooseInExp exp)
  {
    return quantifiedExpression(exp.qvar, exp.exp);
  }
  
  public Expression visit(ChooseInElseExp exp)
  {
    process(exp.exp2);
    return quantifiedExpression(exp.qvar, exp.exp1);
  }
  
  public Expression visit(ForallExp exp)
  {
    if (!ChooseSearch.execute(exp)) return exp;
    List<QuantifiedVariableCore> old = resetVariables();
    Set<String> oset = getVisible();
    Expression exp0 = quantifiedExpression(exp.qvar, exp.exp);
    return setType(exp, new ForallExp(getVariable(old, oset), exp0));
  }
  
  public Expression visit(ExistsExp exp)
  {
    if (!ChooseSearch.execute(exp)) return exp;
    List<QuantifiedVariableCore> old = resetVariables();
    Set<String> oset = getVisible();
    Expression exp0 = quantifiedExpression(exp.qvar, exp.exp);
    return setType(exp, new ExistsExp(getVariable(old, oset), exp0));
  }  
  
  public Expression visit(MinExp exp)
  {
    if (!ChooseSearch.execute(exp)) return exp;
    List<QuantifiedVariableCore> old = resetVariables();
    Set<String> oset = getVisible();
    Expression exp0 = quantifiedExpression(exp.qvar, exp.exp);
    return setType(exp, new MinExp(getVariable(old, oset), exp0));
  } 
  
  public Expression visit(MaxExp exp)
  {
    if (!ChooseSearch.execute(exp)) return exp;
    List<QuantifiedVariableCore> old = resetVariables();
    Set<String> oset = getVisible();
    Expression exp0 = quantifiedExpression(exp.qvar, exp.exp);
    return setType(exp, new MaxExp(getVariable(old, oset), exp0));
  } 
  
  public Expression visit(SumExp exp)
  {
    if (!ChooseSearch.execute(exp)) return exp;
    List<QuantifiedVariableCore> old = resetVariables();
    Set<String> oset = getVisible();
    Expression exp0 = quantifiedExpression(exp.qvar, exp.exp);
    return setType(exp, new SumExp(getVariable(old, oset), exp0));
  } 
  
  public Expression visit(ProductExp exp)
  {
    if (!ChooseSearch.execute(exp)) return exp;
    List<QuantifiedVariableCore> old = resetVariables();
    Set<String> oset = getVisible();
    Expression exp0 = quantifiedExpression(exp.qvar, exp.exp);
    return setType(exp, new ProductExp(getVariable(old, oset), exp0));
  } 
  
  public Expression visit(NumberExp exp)
  {
    if (!ChooseSearch.execute(exp)) return exp;
    List<QuantifiedVariableCore> old = resetVariables();
    Set<String> oset = getVisible();
    quantifiedExpression(exp.qvar, null);
    return setType(exp, new NumberExp(getVariable(old, oset)));
  } 
  
  public Expression visit(SetBuilderExp exp)
  {
    if (!ChooseSearch.execute(exp)) return exp;
    List<QuantifiedVariableCore> old = resetVariables();
    Set<String> oset = getVisible();
    Expression exp0 = quantifiedExpression(exp.qvar, exp.exp);
    return setType(exp, new SetBuilderExp(exp0, getVariable(old, oset)));
  } 
  
  public Expression visit(LetExp exp)
  {
    if (!ChooseSearch.execute(exp)) return exp;
    return setType(exp, letExpression(Arrays.asList(exp.binder), exp.exp, false));
  }
  
  public Expression visit(LetParExp exp)
  {
    if (!ChooseSearch.execute(exp)) return exp;
    return setType(exp, letExpression(Arrays.asList(exp.binder), exp.exp, true));
  }
    
  // --------------------------------------------------------------------------
  //
  // Auxiliaries
  //
  // --------------------------------------------------------------------------

  /***************************************************************************
   * Process let expression by translation to quantified variables with equalities
   * @param binder the binders of the let expression
   * @param exp the body of the let expression
   * @param parallel true if the binders bind "in parallel"
   * @return the result of the processing.
   **************************************************************************/
  private Expression letExpression(List<Binder> binder, Expression exp, 
    boolean parallel)
  {
    List<QuantifiedVariableCore> vars0 = new ArrayList<QuantifiedVariableCore>();
    for (Binder b : binder)
    {
      ValueSymbol symbol = (ValueSymbol)b.ident.getSymbol();
      Identifier ident = symbol.ident;
      Identifier ident0 = parallel && visible.contains(ident.string) ? 
          freshIdent(ident) : ident;
      Expression idexp0 = identifierExp(ident0);
      Expression bexp = process(b.exp); 
      Expression eq0 = new EqualsExp(idexp0, bexp);
      eq0.setType(new Type.BoolType());
      QuantifiedVariableCore var = 
          new QuantifiedVariableCore.IdentifierTypeQuantifiedVar(ident0, 
              symbol.type, eq0);
      vars.add(var);
      visible.add(ident0.string);
      if (ident != ident0) 
      {
        Expression idexp1 = identifierExp(ident);
        Expression idexp2 = identifierExp(ident0);
        Expression eq1 = new EqualsExp(idexp1, idexp2);
        eq1.setType(new Type.BoolType());
        QuantifiedVariableCore var0 = 
            new QuantifiedVariableCore.IdentifierTypeQuantifiedVar(ident, 
                symbol.type, eq1);
        vars0.add(var0);
      }
    }
    if (parallel)
    {
      for (QuantifiedVariableCore var : vars0)
      {
        vars.add(var);
        visible.add(var.ident.string);
      }
    }
    return process(exp);
  }
  
  /***************************************************************************
   * Construct typed identifier expression.
   * @param ident the identifier.
   * @return the expression.
   **************************************************************************/
  private static IdentifierExp identifierExp(Identifier ident)
  {
    ValueSymbol symbol = (ValueSymbol)ident.getSymbol();
    IdentifierExp result = new IdentifierExp(ident);
    result.setType(symbol.type);
    return result;
  }
  
  /****************************************************************************
   * Process quantified expression
   * @param qvar the quantified variable.
   * @param the body expression (may be null indicating a reference to the variables)
   * @return the replaced expression (may be the reference to the variables)
   ***************************************************************************/
  private Expression quantifiedExpression(QuantifiedVariable qvar, Expression exp)
  {
    List<Identifier> idents = new ArrayList<Identifier>();
    List<Identifier> idents0 = new ArrayList<Identifier>();
    List<QuantifiedVariableCore> core0 = new ArrayList<QuantifiedVariableCore>();
    int n = qvar.qvcore.length;
    for (int i = 0; i < n; i++)
    {
      QuantifiedVariableCore c = qvar.qvcore[i];
      Expression wexp = c.wexp;
      if (i == n-1 && qvar.exp != null)
      {
        wexp = AST.and(wexp, qvar.exp);
        wexp.setType(qvar.exp.getType());
      }
      Identifier ident = c.ident;
      Identifier ident0 = 
          visible.contains(ident.string) ? freshIdent(ident) : ident;
      idents.add(ident);
      idents0.add(ident0);
      QuantifiedVariableCore c0 = quantifiedVariable(c, wexp, ident0, idents, idents0);
      vars.add(c0);
      core0.add(c0);
    }
    if (exp != null) 
    {
      exp = CloneIds.bind(idents, idExps(idents0), exp);
      return process(exp);
    }
    if (core0.size() == 1) return identifierExp(core0.get(0).ident);
    List<Expression> exps = new ArrayList<Expression>();
    for (QuantifiedVariableCore c0 : core0) exps.add(identifierExp(c0.ident));
    return new TupleExp(exps);    
  }
  
  /****************************************************************************
   * Get a fresh version of identifier that is currently not visible.
   * @param ident the identifier.
   * @return a derived identifier that is not visible.
   ***************************************************************************/
  private Identifier freshIdent(Identifier ident)
  {
    String string = freshVariable(ident.string);
    Identifier ident0 = new Identifier(string);
    ValueSymbol symbol = (ValueSymbol)ident.getSymbol();
    ident0.setSymbol(new ValueSymbol(ident0, symbol.type, ValueSymbol.Kind.value));
    return ident0;
  }
  
  /****************************************************************************
   * Get the name of a variable that does not occur among the visible variables
   * @param the base name of the variable.
   * @return a derived name which is not visible.
   ***************************************************************************/
  private String freshVariable(String name)
  {
    int i = 0;
    while (true)
    {
      String name0 = "_" +  name + "_" + i;
      if (!visible.contains(name0))  return name0;
      i++;
    }
  }
  
  /****************************************************************************
   * Rename quantified variable core.
   * @param c the original variable core
   * @param wexp the condition attached to the core (may be null)
   * @param ident the new identifier for the core
   * @param core  the original identifiers so far.
   * @param core0 the replacements for the original identifiers so far.
   * @return the renamed version of the core.
   ***************************************************************************/
  private QuantifiedVariableCore quantifiedVariable(
    QuantifiedVariableCore c, Expression wexp, Identifier ident,
    List<Identifier> idents, List<Identifier> idents0)
  { 
    if (c instanceof QuantifiedVariableCore.IdentifierTypeQuantifiedVar)
    {
      QuantifiedVariableCore.IdentifierTypeQuantifiedVar c0 =
          (QuantifiedVariableCore.IdentifierTypeQuantifiedVar)c;
      visible.add(ident.string);
      if (wexp != null)
      { 
        wexp = optimizeFormula(wexp, visible);
        wexp = CloneIds.bind(idents, idExps(idents0), wexp);
      }
      return new QuantifiedVariableCore.IdentifierTypeQuantifiedVar(
          ident, c0.type, wexp);
    }
    else if (c instanceof QuantifiedVariableCore.IdentifierSetQuantifiedVar)
    {
      QuantifiedVariableCore.IdentifierSetQuantifiedVar c0 =
          (QuantifiedVariableCore.IdentifierSetQuantifiedVar)c;
      Expression exp0 = process(c0.exp);
      exp0.setType(c0.exp.getType());
      visible.add(ident.string);
      if (wexp != null)
      { 
        wexp = optimizeFormula(wexp, visible);
        wexp = CloneIds.bind(idents, idExps(idents0), wexp);
      }
      return new QuantifiedVariableCore.IdentifierSetQuantifiedVar(
          ident, exp0, wexp);
    }
    return null;
  }
  
  /***************************************************************************
   * Transform list of identifiers into corresponding expressions.
   * @param ids the identifiers.
   * @return the corresponding expressions.
   ***************************************************************************/
  private static List<Expression> idExps(List<Identifier> ids)
  {
    List<Expression> result = new ArrayList<Expression>();
    for (Identifier id : ids)
    {
      Expression exp = new IdentifierExp(id);
      ValueSymbol symbol = (ValueSymbol)id.getSymbol();
      exp.setType(symbol.type);
      result.add(exp);
    }
    return result;
  }
}
// ---------------------------------------------------------------------------
// end of file
// ---------------------------------------------------------------------------