// ---------------------------------------------------------------------------
// Environment.java
// The environment for type checking and translation.
// $Id: Environment.java,v 1.52 2020/05/19 17:02:27 schreine Exp $
//
// Author: Wolfgang Schreiner <Wolfgang.Schreiner@risc.jku.at>
// Copyright (C) 2016-, Research Institute for Symbolic Computation (RISC)
// Johannes Kepler University, Linz, Austria, http://www.risc.jku.at
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
// ----------------------------------------------------------------------------
package riscal.types;

import java.util.*;

import riscal.syntax.*;
import riscal.syntax.AST.Identifier;
import riscal.syntax.AST.Type;
import riscal.types.Environment.Symbol.*;
import riscal.semantics.Types.*;
import riscal.semantics.Systems.SysSem;
import riscal.Main;
import riscal.semantics.Systems;
import riscal.semantics.Systems.ActSem;
import riscal.semantics.Systems.ActSemComp;
import riscal.semantics.Systems.CompSem;
import riscal.util.*;
import riscal.tasks.*;

public class Environment
{
  // the generic table type we use
  public static class Map<S> extends LinkedHashMap<String,S>
  {
    private static final long serialVersionUID = 28091967L;
    public Map() { super(); }
    public Map(Map<S> t) { super(t); }
  }
  
  // the separate name spaces (functions may be overloaded)
  private Map<TypeSymbol> types;
  private Map<ValueSymbol> values;
  private Map<List<FunctionSymbol>> functions;
  private Map<SystemSymbol> systems;
  private Map<List<ActionSymbol>> actions; // for shared systems
  private Map<ComponentSymbol> components; // for distributed systems
  
  public static abstract class Symbol
  {
    public final AST.Identifier ident; // the identifier
    protected Symbol(AST.Identifier ident) 
    { 
      // link symbol to identifier and vice versa
      assert ident == null;
      this.ident = ident; 
      ident.setSymbol(this);
    }
    
    // the information on named types
    public final static class TypeSymbol extends Symbol
    {
      private AST.Type type; // may be temporarily null for recursive types
      public TypeSymbol(AST.Identifier ident, AST.Type type) 
      { 
        super(ident);
        this.type = type; 
      }
      public AST.Type getType() { return type; }
      public void setType(AST.Type type) { this.type = type; }
      public String toString() { return ident.string + ":" + type; }
    }

    // the information on values and variables
    public final static class ValueSymbol extends Symbol
    {
      // the kind of declaration
      // global value: stored in global frame
      // local value or variable: stored in local frame 
      public static enum Kind { global, value, variable; }
      public final AST.Type type;        // its type
      public final Kind kind;            // the declaration kind
      
      public ValueSymbol(AST.Identifier ident, AST.Type type, Kind kind)
      {
        super(ident);
        this.type = type;
        this.kind = kind;
      }
      public String toString() 
      { 
        return ident.string + ":" +
               (kind == Kind.global ? "global" : 
                kind == Kind.value ? "value" : "variable") 
                + ":" + (kind == Kind.global ? value : slot) + ":" + type; 
      }
      
      // the value of the object (null, if unknown)
      private Value value = null;              
      public Value getValue() { return value; }
      public void setValue(Value value) { this.value = value; }
      
      // the value sequence of the object (null, if unknown)
      private Seq<Value> values = null;              
      public Seq<Value> getValues() { return values; }
      public void setValues(Seq<Value> values) { this.values = values; }
      
      // if local, the frame slot of the value (-1, if unknown)
      private int slot = -1;
      public int getSlot() { return slot; }
      public void setSlot(int slot) { this.slot = slot; }
      
      // the number of slots in the "function" frame (-1, if unknown)
      private int slots = -1;
      public int getSlots() { return slots; }
      public void setSlots(int slots) { this.slots = slots; }
      
      // the syntactic (declaration) type (null, if unknown)
      private Type stype = null;
      public Type getSyntacticType() { return stype; }
      public void setSyntacticType(Type stype) { this.stype = stype; }
    }

    // the information of functions, predicates, and procedures
    public final static class FunctionSymbol extends Symbol
    {
      public final boolean multiple;     // is the function non-deterministic?
      public final ValueSymbol[] params; // its parameters
      public final AST.Type[] types;     // the argument types
      public final AST.Type type;        // the result type (Bool for predicates)
      
      public FunctionSymbol(boolean multiple, 
        AST.Identifier ident, ValueSymbol[] params, AST.Type type)
      {
        super(ident);
        this.multiple = multiple;
        this.params = params;
        this.types = new Type[params.length];
        for (int i=0; i<params.length; i++) types[i] = params[i].type;
        this.type = type;
      }
      
      // the definition of the symbol (null, if only declared)
      private AST.Declaration def = null;
      public boolean isDefined() { return def != null; }
      public AST.Declaration getDefinition() { return def; }
      public void makeDefined(AST.Declaration def) { this.def = def; }
      public String toString() 
      { 
        return ident.string + ":" + isDefined() + ":" + slots + ":" + 
          AST.toString(types, ",") + ":" + type; 
      }
      
      // the value of the object (null, if unknown)
      private FunSem value = null;              
      public FunSem getValue() { return value; }
      public void setValue(FunSem value) { this.value = value; }
      
      // the preconditions of the functions (null, if none)
      private List<ContextCondition> pre = null;
      public List<ContextCondition> getPreconditions() { return pre; }
      public void setPreconditions(List<ContextCondition> pre) { this.pre = pre; }
      
      // the number of slots in the function frame (-1, if unknown)
      private int slots = -1;
      public int getSlots() { return slots; }
      public void setSlots(int slots) { this.slots = slots; }
      
      // the result symbol (null, if unknown)
      private ValueSymbol result = null;
      public ValueSymbol getResult() { return result; }
      public void setResult(ValueSymbol result) { this.result = result; }
      
      // folder of associated tasks (may be null)
      private TaskFolder tasks = null;
      public TaskFolder getTaskFolder() { return tasks; }
      public void setTaskFolder(TaskFolder tasks) { this.tasks = tasks; }
    }
    
    // the information for systems
    public final static class SystemSymbol extends Symbol
    {
      // the kind of system
      public static enum Kind { shared, distributed }
      public final Kind kind;
      public SystemSymbol(AST.Identifier ident, Kind kind)
      {
        super(ident);
        this.kind = kind;
      }
      
      // the state variables of the system
      private ValueSymbol[] vars;
      public ValueSymbol[] getVars() { return vars; }
      public void setVars(ValueSymbol[] vars) { this.vars = vars; }
      
      // the poststate counterparts of the state variables
      private ValueSymbol[] newVars;
      public ValueSymbol[] getNewVars() { return newVars; }
      public void setNewVars(ValueSymbol[] newVars) { this.newVars = newVars; }
      
      // the number of slots in the action frame (-1, if unknown)
      private int slots = -1;
      public int getSlots() { return slots; }
      public void setSlots(int slots) { this.slots = slots; }
      
      // the value of the object (null, if unknown)
      private SysSem value = null;              
      public SysSem getValue() { return value; }
      public void setValue(SysSem value) { this.value = value; }
      
      // the names of the state variables
      private String[] names = null;
      public void setNames(String[] names) { this.names = names; }
      public String[] getNames() { return names; }
      
      // the symbols for the initial state variables
      private ValueSymbol[] oldVars = null;
      public ValueSymbol[] getOldVars() { return oldVars; }
      public void setOldVars(ValueSymbol[] oldVars) { this.oldVars = oldVars; }
      
      // the (processed) measures and invariants of the system
      List<ContextRelation> invariants; List<ContextFunction> variants;
      List<AST> cinvariants; List<AST> cvariants;
      public void setSpecification(
        List<ContextRelation> invariants, List<ContextFunction> variants,
        List<AST> cinvariants, List<AST> cvariants)
      {
        this.invariants = invariants; this.variants = variants;
        this.cinvariants = cinvariants; this.cvariants = cvariants;
      }
      public List<ContextRelation> getInvariants() { return invariants; }
      public List<ContextFunction> getVariants() { return variants; }
      public List<AST> getInvariantDeclarations() { return cinvariants; }
      public List<AST> getVariantDeclarations() { return cvariants; }
      
      // the components of the (distributed) system
      private ComponentSymbol[] components = null;
      public ComponentSymbol[] getComponents() { return components; }
      public void setComponents(ComponentSymbol[] components) { this.components = components; }
      
      // get executable function symbol from system
      private FunctionSymbol fun = null;
      public FunctionSymbol getFunctionSymbol()
      {
        if (fun != null) return fun;
        Identifier ident0 = new AST.Identifier("system " + ident.string);
        ident0.setPosition(ident.getPosition());
        fun = new FunctionSymbol(false, ident0,
          new ValueSymbol[] {}, new AST.Type.UnitType());
        fun.setValue((FunSem.Single)(Argument a)->
        {
          Systems.Visitor<Systems.StateClass> visitor = 
              new Systems.Visitor<Systems.StateClass>(getValue());
          Collection<Systems.StateClass> states = visitor.visit();
          int n = states.size();
          Main.println(n + " system " + (n == 1 ? "state" : "states")
              + " found with search depth " + visitor.getDepth() + ".");
          return new Value.Unit();
        });
        fun.setTaskFolder(new TaskFolder(null, "system " + ident.string, null));
        return fun;
      }
      
      // does s refer to the execution of a system
      public static boolean isSystem(FunctionSymbol s)
      {
        return s.ident.string.startsWith("system ");
      }
      
      // the number of components put into the system so far
      private int cnumber = 0;
      public int newComponent() { int c = cnumber; cnumber++; return c; }
    }
    
    // the information for actions
    public final static class ActionSymbol extends Symbol
    {
      public final SystemSymbol system;  // the associated system
      public final int bsize;            // the buffer size
      public final ValueSymbol[] params; // its parameters
      public final AST.Type[] types;     // their types
      public static enum Kind { command, function, predicate }
      public final Kind kind;            // the kind of action
      public ActionSymbol(SystemSymbol system, int bsize,
        AST.Identifier ident, ValueSymbol[] params, Kind kind)
      {
        super(ident);
        this.system = system;
        this.bsize = bsize;
        this.params = params;
        this.types = new Type[params.length];
        for (int i=0; i<params.length; i++) types[i] = params[i].type;
        this.kind = kind;
      }
      
      // the value of the object (null, if unknown)
      private ActSem value = null;              
      public ActSem getValue() { return value; }
      public void setValue(ActSem value) { this.value = value; }
      
      // the component of the (distributed) system to which the action belongs
      private ComponentSymbol component = null;
      public ComponentSymbol getComponent() { return component; }
      public void setComponent(ComponentSymbol component) { this.component = component; }
      
      // the component value of the object (null, if unknown)
      private ActSemComp valueComp = null;              
      public ActSemComp getValueComp() { return valueComp; }
      public void setValueComp(ActSemComp valueComp) { this.valueComp = valueComp; }
      
      // the number of the action (-1, if unknown)
      private int number = -1;
      public int getNumber() { return number; }
      public void setNumber(int number) { this.number = number; }
    }
    
    // the information for components
    public final static class ComponentSymbol extends Symbol
    {
      public final SystemSymbol system;        // the associated system
      public final int instances;              // the number of component instances
      private Map<List<ActionSymbol>> actions; // the component actions
      private List<ActionSymbol> alist;        // the actions as a list
      public final int number;                 // the number of the component
      
      public ComponentSymbol(SystemSymbol system, 
        AST.Identifier ident, int instances)
      {
        super(ident);
        this.system = system;
        this.instances = instances;
        this.actions = new Map<List<ActionSymbol>>();
        this.alist = new ArrayList<ActionSymbol>();
        this.number = system.newComponent();
      }
      
      // the number of actions put into the component so far
      private int anumber = 0;
      
      // put action into the component
      public ActionSymbol putAction(int bsize, AST.Identifier ident, 
        ValueSymbol[] params, ActionSymbol.Kind kind, boolean init)
      {
        ActionSymbol symbol = new ActionSymbol(system, bsize, ident, params, kind);
        symbol.setComponent(this);
        if (init) return symbol;
        symbol.setNumber(anumber);
        anumber++;
        putActionSymbol(symbol, actions);
        alist.add(symbol);
        return symbol;
      }
      
      // the state variables of the component
      private ValueSymbol[] vars;
      public ValueSymbol[] getVars() { return vars; }
      public void setVars(ValueSymbol[] vars) { this.vars = vars; }
      
      // the names of the state variables
      private String[] names = null;
      public void setNames(String[] names) { this.names = names; }
      public String[] getNames() { return names; }
      
      // the poststate counterparts of the state variables
      private ValueSymbol[] newVars;
      public ValueSymbol[] getNewVars() { return newVars; }
      public void setNewVars(ValueSymbol[] newVars) { this.newVars = newVars; }
      
      // the variable representing the component state within the system
      private ValueSymbol stateVar;
      public ValueSymbol getStateVar() { return stateVar; }
      public void setStateVar(ValueSymbol stateVar) { this.stateVar = stateVar; }

      // Returns the actions associated to the given identifier
      public List<ActionSymbol> getActions(AST.Identifier ident)
      {
        return actions.get(ident.string);
      }
      
      // returns the list of actions
      public List<ActionSymbol> getActions() { return alist; }
      
      // the value of the object (null, if unknown)
      private CompSem value = null;              
      public CompSem getValue() { return value; }
      public void setValue(CompSem value) { this.value = value; }
    }
  }
  
  /***************************************************************************
   * Create an empty environment
   **************************************************************************/
  public Environment()
  {
    this.types = new Map<TypeSymbol>();
    this.values = new Map<ValueSymbol>();
    this.functions = new Map<List<FunctionSymbol>>();
    this.systems = new Map<SystemSymbol>();
    this.actions = new Map<List<ActionSymbol>>();
    this.components = new Map<ComponentSymbol>();
  }
  
  /***************************************************************************
   * Create a duplicate of a given environment.
   * @param env the environment to be duplicated
   **************************************************************************/
  public Environment(Environment env)
  {
    this.types = new Map<TypeSymbol>(env.types);
    this.values = new Map<ValueSymbol>(env.values);
    this.functions = new Map<List<FunctionSymbol>>(env.functions);
    this.systems = new Map<SystemSymbol>(env.systems);
    this.actions = new Map<List<ActionSymbol>>(env.actions);
    this.components = new Map<ComponentSymbol>(env.components);
  }
  
  /**************************************************************************
   * Put a type into the environment (overriding any entry with same name)
   * @param ident the identifier of the type
   * @param type the type itself (must have canonical version)
   * @return the new type symbol 
   *************************************************************************/
  public TypeSymbol putType(AST.Identifier ident, AST.Type type)
  {
    TypeSymbol symbol = new TypeSymbol(ident, type);
    putTypeSymbol(symbol);
    return symbol;
  }
  
  /**************************************************************************
   * Put type symbol into environment (overriding any entry with same name)
   * @param symbol the symbol
   *************************************************************************/
  public void putTypeSymbol(TypeSymbol symbol)
  {
    types.put(symbol.ident.string, symbol);
  }
  
  /**************************************************************************
   * Put a value into the environment (overriding any entry with same name)
   * @param ident the identifier of the value
   * @param type the type of the value (must have canonical version)
   * @param level the declaration kind
   * @return the new value symbol 
   *************************************************************************/
  public ValueSymbol putValue(AST.Identifier ident, AST.Type type, 
    ValueSymbol.Kind kind)
  {
    ValueSymbol symbol = new ValueSymbol(ident, type, kind);
    putValueSymbol(symbol);
    return symbol;
  }
  
  /**************************************************************************
   * Put value symbol into environment (overriding any entry with same name)
   * @param symbol the symbol
   *************************************************************************/
  public void putValueSymbol(ValueSymbol symbol)
  {
    values.put(symbol.ident.string, symbol);
  }
  
  /**************************************************************************
   * Remove value symbols from environment
   * @param symbols the symbols to be removed
   * @return the original symbol map
   *************************************************************************/
  public Map<ValueSymbol> delValueSymbols(ValueSymbol[] symbols)
  {
    Map<ValueSymbol> values0 = values;
    values = new Map<ValueSymbol>(values);
    for (ValueSymbol symbol : symbols)
      values.remove(symbol.ident.string, symbol);
    return values0;
  }
  
  /**************************************************************************
   * Set value symbols to given map.
   * @param values0 the new symbol map.
   *************************************************************************/
  public void setValueSymbols(Map<ValueSymbol> values0)
  {
    values = values0;
  }
  
  /**************************************************************************
   * Put a function into the environment (overriding any entry with same 
   * name and argument types)
   * @param multiple is the function non-deterministic?
   * @param ident the identifier of the function
   * @param params the parameter symbols 
   * @param type the result type (Bool for predicates, must have canonical v.)
   * @return the function symbol
   *************************************************************************/
  public FunctionSymbol putFunction(boolean multiple,
    AST.Identifier ident, ValueSymbol[] params,
    AST.Type type)
  {
    FunctionSymbol symbol = new FunctionSymbol(multiple, ident, params, type);
    putFunctionSymbol(symbol);
    return symbol;
  }
  
  /**************************************************************************
   * Put function symbol into environment (overriding any entry with same name)
   * @param symbol the symbol
   *************************************************************************/
  public void putFunctionSymbol(FunctionSymbol symbol)
  {
    List<FunctionSymbol> symbols = functions.get(symbol.ident.string);
    if (symbols == null)
    {
      symbols = new LinkedList<FunctionSymbol>();
      symbols.add(symbol);
      functions.put(symbol.ident.string, symbols);
      return;
    }
    FunctionSymbol symbol0 = null;
    for (FunctionSymbol s : symbols)
    {
      if (Arrays.equals(s.types, symbol.types))
      {
        symbol0 = s;
        break;
      }
    }
    if (symbol0 != null) symbols.remove(symbol0);
    symbols.add(symbol);
  }
  
  /****************************************************************************
   * Returns the type associated to the given identifier
   * (sets the symbol of the identifier as a side effect).
   * @param ident the identifier
   * @return the associated type (null, if none)
   ***************************************************************************/
  public TypeSymbol getType(AST.Identifier ident)
  {
    TypeSymbol symbol = types.get(ident.string);
    ident.setSymbol(symbol);
    return symbol;
  }
  
  /****************************************************************************
   * Get all types in current environment.
   * @return the collection of all type symbols.
   ***************************************************************************/
  public Collection<TypeSymbol> getTypes()
  {
    return types.values();
  }
  
  /****************************************************************************
   * Returns the value associated to the given identifier
   * (sets the symbol of the identifier as a side effect).
   * @param ident the identifier
   * @return the associated value (null, if none)
   ***************************************************************************/
  public ValueSymbol getValue(AST.Identifier ident)
  {
    ValueSymbol symbol = values.get(ident.string);
    ident.setSymbol(symbol);
    return symbol;
  }
  
  /****************************************************************************
   * Get all values in current environment.
   * @return the collection of all value symbols.
   ***************************************************************************/
  public Collection<ValueSymbol> getValues()
  {
    return values.values();
  }
  
  /****************************************************************************
   * Returns the functions associated to the given identifier
   * @param ident the identifier
   * @return the associated functions (null, if none)
   ***************************************************************************/
  public List<FunctionSymbol> getFunctions(AST.Identifier ident)
  {
    return functions.get(ident.string);
  }
  
  /****************************************************************************
   * Get all function in current environment.
   * @return the collection of all value symbols.
   ***************************************************************************/
  public Collection<FunctionSymbol> getFunctions()
  {
    List<FunctionSymbol> result = new ArrayList<FunctionSymbol>();
    for (Collection<FunctionSymbol> fs : functions.values())
      result.addAll(fs);
    return result;
  }
  
  /**************************************************************************
   * Put system symbol into environment (overriding any entry with same name)
   * @param symbol the symbol
   *************************************************************************/
  public void putSystemSymbol(SystemSymbol symbol)
  {
    systems.put(symbol.ident.string, symbol);
  }
  
  /**************************************************************************
   * Put a system into the environment.
   * @param ident the identifier of the system.
   * @param ident the kind of system.
   * @return the new system symbol. 
   *************************************************************************/
  public SystemSymbol putSystem(AST.Identifier ident, SystemSymbol.Kind kind)
  {
    SystemSymbol symbol = new SystemSymbol(ident, kind);
    putSystemSymbol(symbol);
    return symbol;
  }
  
  /****************************************************************************
   * Returns the system associated to the given identifier
   * (sets the symbol of the identifier as a side effect).
   * @param ident the identifier
   * @return the associated system (null, if none)
   ***************************************************************************/
  public SystemSymbol getSystem(AST.Identifier ident)
  {
    SystemSymbol symbol = systems.get(ident.string);
    ident.setSymbol(symbol);
    return symbol;
  }
  
  /**************************************************************************
   * Put action symbol into environment (overriding any entry with same name)
   * @param symbol the symbol
   *************************************************************************/
  public void putActionSymbol(ActionSymbol symbol)
  {
    putActionSymbol(symbol, actions);
  }
  
  /**************************************************************************
   * Put action symbol into environment (overriding any entry with same name)
   * @param symbol the symbol
   * @param actions the environment
   *************************************************************************/
  private static void putActionSymbol(ActionSymbol symbol, 
    Map<List<ActionSymbol>> actions)
  {
    List<ActionSymbol> symbols = actions.get(symbol.ident.string);
    if (symbols == null)
    {
      symbols = new LinkedList<ActionSymbol>();
      symbols.add(symbol);
      actions.put(symbol.ident.string, symbols);
      return;
    }
    ActionSymbol symbol0 = null;
    for (ActionSymbol s : symbols)
    {
      if (Arrays.equals(s.types, symbol.types))
      {
        symbol0 = s;
        break;
      }
    }
    if (symbol0 != null) symbols.remove(symbol0);
    symbols.add(symbol);
  }
  
  /**************************************************************************
   * Put an action into the environment.
   * @param system the system associated to the action.
   * @param bsize the buffer size of the action
   * @param ident the identifier of the action.
   * @param params the parameters of the action.
   * @param ident the kind of action.
   * @return the new action symbol. 
   *************************************************************************/
  public ActionSymbol putAction(SystemSymbol system, int bsize,
    AST.Identifier ident, ValueSymbol[] params, ActionSymbol.Kind kind)
  {
    ActionSymbol symbol = new ActionSymbol(system, bsize, ident, params, kind);
    putActionSymbol(symbol);
    return symbol;
  }
  
  /****************************************************************************
   * Returns the actions associated to the given identifier
   * @param ident the identifier
   * @return the associated actions (null, if none)
   ***************************************************************************/
  public List<ActionSymbol> getActions(AST.Identifier ident)
  {
    return actions.get(ident.string);
  }
  
  /**************************************************************************
   * Put a component into the environment (overriding any entry with same name)
   * @param ident the identifier of the type
   * @param type the type itself (must have canonical version)
   * @return the new type symbol 
   *************************************************************************/
  public ComponentSymbol putComponent(SystemSymbol system, AST.Identifier ident, int number)
  {
    ComponentSymbol symbol = new ComponentSymbol(system, ident, number);
    putComponentSymbol(symbol);
    return symbol;
  }
  
  /**************************************************************************
   * Put component symbol into environment (overriding any entry with same name)
   * @param symbol the symbol
   *************************************************************************/
  public void putComponentSymbol(ComponentSymbol symbol)
  {
    components.put(symbol.ident.string, symbol);
  }
  
  /****************************************************************************
   * Returns the component associated to the given identifier
   * @param ident the identifier
   * @return the associated component (null, if none)
   ***************************************************************************/
  public ComponentSymbol getComponent(AST.Identifier ident)
  {
    ComponentSymbol symbol = components.get(ident.string);
    if (symbol != null) ident.setSymbol(symbol);
    return symbol;
  }
}
//---------------------------------------------------------------------------
// end of file
//---------------------------------------------------------------------------