// ---------------------------------------------------------------------------
// TraceWindow.java
// A window visualizing an execution trace
// $Id: TraceWindow.java,v 1.15 2018/07/11 09:38:30 schreine Exp $
//
// Author: Wolfgang Schreiner <Wolfgang.Schreiner@risc.jku.at>
// Copyright (C) 2018-, Research Institute for Symbolic Computation (RISC)
// Johannes Kepler University, Linz, Austria, http://www.risc.jku.at
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
// ----------------------------------------------------------------------------
package riscal.zest;

import org.eclipse.swt.*;
import org.eclipse.swt.layout.*;
import org.eclipse.swt.widgets.*;
import org.eclipse.swt.graphics.*;
import org.eclipse.core.runtime.NullProgressMonitor;
import org.eclipse.core.commands.ExecutionException;

import com.google.inject.Guice;
import com.google.inject.Injector;
import com.google.inject.Provider;
import com.google.inject.multibindings.MapBinder;

import javafx.application.*;
import javafx.scene.Scene;
import javafx.scene.input.*;

import java.util.*;
import java.util.List;

import org.eclipse.gef.common.adapt.*;
import org.eclipse.gef.common.adapt.inject.AdapterInjectionSupport;
import org.eclipse.gef.common.adapt.inject.AdapterInjectionSupport.LoggingMode;
import org.eclipse.gef.fx.swt.canvas.*;
import org.eclipse.gef.graph.*;
import org.eclipse.gef.layout.algorithms.*;
import org.eclipse.gef.mvc.fx.domain.*;
import org.eclipse.gef.mvc.fx.viewer.*;
import org.eclipse.gef.mvc.fx.handlers.*;
import org.eclipse.gef.mvc.fx.policies.*;
import org.eclipse.gef.mvc.fx.parts.*;
import org.eclipse.gef.mvc.fx.providers.*;
import org.eclipse.gef.zest.fx.*;
import org.eclipse.gef.zest.fx.parts.*;
import org.eclipse.gef.zest.fx.operations.*;
import org.eclipse.gef.zest.fx.behaviors.*;
import org.eclipse.gef.zest.fx.policies.*;
import org.eclipse.gef.zest.fx.handlers.*;
import org.eclipse.gef.zest.fx.providers.*;

import riscal.semantics.*;

public class TraceWindow
{
  // there may be only one instance of this object
  private static TraceWindow traceWindow;
  
  // the shell of the window
  private Shell shell;

  // the canvas to display the graph on
  private FXCanvasEx canvas;

  // the trace to display
  private List<Trace.Entry> trace;

  // the map of (nested) graphs to header strings
  private Map<Graph,String> headers = new HashMap<Graph,String>();

  // level of currently displayed procedure invocation
  private int level = 0;
  
  /****************************************************************************
   * Construct window connected to denoted GUI.
   * @param display the display.
   * @param image the icon image.
   * @param header the header text of the window.
   * @param width initial width
   * @param height initial height
   * @return a window connected to the GUI.
   ***************************************************************************/
  public static TraceWindow construct(Display display, 
    Image image, String header,
    int width, int height)
  {
    header = "Level 0: " + header;
    Shell shell = new Shell(display, 
        // SWT.SYSTEM_MODAL |
        SWT.CLOSE | SWT.MIN | SWT.MAX | SWT.RESIZE | SWT.TITLE | SWT.BORDER);
    shell.setText(header);
    shell.setImage(image);
    traceWindow = new TraceWindow(shell, width, height, header);
    return traceWindow;
  }

  /****************************************************************************
   * Construct panel linked to shell.
   * @param shell the shell
   * @param width initial width
   * @param height initial height
   * @param header the header for the top-level procedure
   ***************************************************************************/
  private TraceWindow(Shell shell, int width, int height, String header)
  {
    // resize it
    this.shell = shell;
    shell.setSize(width, height);

    // set shell layout for shell and panel
    shell.setLayout(new FillLayout());

    // extract trace to be displayed
    trace = Trace.extract();
    
    // create the graph
    Graph graph = createGraph(header);
    if (graph.getNodes().isEmpty()) return;
    
    // visualize the graph
    Platform.setImplicitExit(false);
    canvas = new FXCanvasEx(shell, SWT.NONE);
    Scene scene = createScene(graph);
    canvas.setScene(scene);
  }

  /****************************************************************************
   * Get shell of this trace window.
   ***************************************************************************/
  public Shell getShell() { return shell; }

  /****************************************************************************
   * Create the scene containing the denoted graph.
   * @param graph the graph to be displayed.
   * @return the scene.
   ***************************************************************************/
  private static Scene createScene(Graph graph)
  {    
    Injector injector = Guice.createInjector(new TraceModule());
    IDomain domain = injector.getInstance(IDomain.class);
    IViewer viewer = domain.getAdapter(
        AdapterKey.get(IViewer.class, IDomain.CONTENT_VIEWER_ROLE));  
    Scene scene = new Scene(((IViewer) viewer).getCanvas());
    domain.activate();
    Platform.runLater(new Runnable() 
    {
      @Override
      public void run() 
      {
        viewer.getContents().setAll(Collections.singletonList(graph));
      }
    });
    return scene;
  }

  // -------------------------------------------------------------------------
  //
  // graph construction
  //
  // -------------------------------------------------------------------------

  /***************************************************************************
   * Create a node of the graph.
   * @param label the label of the graph (may be null).
   * @param tooltip the tooltip (may be null).
   * @return the node
   **************************************************************************/
  private static Node node(String label, String tooltip) 
  {
    Node.Builder builder = new Node.Builder();
    if (label != null) builder.attr(ZestProperties.LABEL__NE, label);
    if (tooltip != null) builder.attr(ZestProperties.TOOLTIP__N, tooltip);
    // builder.attr(ZestProperties.LABEL_CSS_STYLE__NE, "-fx-font-family: monospace;");
    return builder.buildNode();
  }

  /***************************************************************************
   * Create an edge of the graph.
   * @param n the source node.
   * @param m the destination node.
   * @param label the label (may be null).
   * @return the edge.
   **************************************************************************/
  private static Edge edge(Node n, Node m, String label) 
  {
    Edge.Builder builder = new Edge.Builder(n, m);
    if (label != null) builder.attr(ZestProperties.SOURCE_LABEL__E, label);
    builder.attr(ZestProperties.SOURCE_LABEL_CSS_STYLE__E, 
        "-fx-font-style: italic; -fx-font-weight: bold;");
    builder.attr(ZestProperties.TARGET_DECORATION__E, 
        new javafx.scene.shape.Polygon(0, 0, 10, 3, 10, -3));
    builder.attr(ZestProperties.TARGET_DECORATION_CSS_STYLE__E, 
        "-fx-fill: white;");
    return builder.buildEdge();
  }

  /***************************************************************************
   * Create an dashed edge of the graph.
   * @param n the source node.
   * @param m the destination node.
   * @param label the label (may be null).
   * @return the edge.
   **************************************************************************/
  private static Edge dashed(Node n, Node m, String label) 
  {
    Edge.Builder builder = new Edge.Builder(n, m);
    if (label != null) builder.attr(ZestProperties.SOURCE_LABEL__E, label);
    builder.attr(ZestProperties.SOURCE_LABEL_CSS_STYLE__E, 
        "-fx-font-style: italic; -fx-font-weight: bold;");
    builder.attr(ZestProperties.TARGET_DECORATION__E, 
        new javafx.scene.shape.Polygon(0, 0, 10, 3, 10, -3));
    builder.attr(ZestProperties.TARGET_DECORATION_CSS_STYLE__E, 
        "-fx-fill: white;");
    builder.attr(ZestProperties.CURVE_CSS_STYLE__E, 
        "-fx-stroke-dash-array: 3 5;");
    return builder.buildEdge();
  }

  // the index of the next trace entry to be processed
  private int index = 0;

  /***************************************************************************
   * Create the graph visualizing the trace
   * @param  header the window header to be displayed for the graph
   * @return the graph.
   **************************************************************************/
  private Graph createGraph(String header)
  { 
    // the nodes and edges generated from the trace
    List<Node> nodes = new ArrayList<Node>();
    List<Edge> edges = new ArrayList<Edge>();

    // populate the graph
    level++;
    populate(nodes, edges);
    level--;
    
    HashMap<String, Object> attrs = new HashMap<>();
    GridLayoutAlgorithm algorithm = new GridLayoutAlgorithm();
    algorithm.setAspectRatio(5); // ???
    algorithm.setRowPadding(25); // ???
    attrs.put(ZestProperties.LAYOUT_ALGORITHM__G, algorithm);    
    Graph graph = new Graph(attrs, nodes, edges);
    headers.put(graph, header);
    return graph;
  }

  /***************************************************************************
   * Construct label from base and optional string of trace entry
   * @param entry the trace entry
   * @return the label
   **************************************************************************/
  private String label(Trace.Entry entry)
  {
    String result = entry.toString();
    String optional = entry.optionalString();
    if (optional != null) result += "\n(" + optional + ")";
    return result;
  }

  /***************************************************************************
   * Populate the graph from the trace starting with the current index.
   * @param nodes the list to add new nodes to
   * @param edges the list to add new edges to
   **************************************************************************/
  private void populate(List<Node> nodes, List<Edge> edges)
  {
    int counter = 0;
    Node last = null;
    while (true)
    {
      Node node = null;
      Edge edge = null;
      if (index >= trace.size()) return;
      Trace.Entry entry = trace.get(index);
      index++;
      if (last == null && !(entry instanceof Trace.Entry.Arguments)) return;
      if (entry instanceof Trace.Entry.Assignment)
      {
        Trace.Entry.Assignment entry0 = (Trace.Entry.Assignment)entry;
        node = node(Integer.toString(counter), entry0.context.toString());
        edge = edge(last, node, label(entry0));
        counter++;
      }
      else if (entry instanceof Trace.Entry.Choice)
      {
        Trace.Entry.Choice entry0 = (Trace.Entry.Choice)entry;
        node = node(Integer.toString(counter), entry0.context.toString());
        edge = edge(last, node, label(entry0));
        counter++;
      }
      else if (entry instanceof Trace.Entry.BoolTest)
      {
        Trace.Entry.BoolTest entry0 = (Trace.Entry.BoolTest)entry;
        node = node("", entry0.context.toString());
        edge = dashed(last, node, label(entry0));
        // counter++;
      }
      else if (entry instanceof Trace.Entry.MatchTest)
      {
        Trace.Entry.MatchTest entry0 = (Trace.Entry.MatchTest)entry;
        node = node("", entry0.context.toString());
        edge = dashed(last, node, label(entry0));
      }
      else if (entry instanceof Trace.Entry.Application)
      {
        Trace.Entry.Application entry0 = (Trace.Entry.Application)entry;
        node = node(entry0.toString(), entry0.toString());
        String header = label(entry0);
        edge = dashed(last, node, header);
        Graph graph = createGraph("Level " + level + ": " + header);
        if (!graph.getNodes().isEmpty())
          node.setNestedGraph(graph);
      }
      else if (entry instanceof Trace.Entry.Arguments)
      {
        Trace.Entry.Arguments entry0 = (Trace.Entry.Arguments)entry;
        node = node(Integer.toString(counter), entry0.context.toString());
        edge = null;
        counter++;
      }
      else if (entry instanceof Trace.Entry.Result)
      {
        Trace.Entry.Result entry0 = (Trace.Entry.Result)entry;
        node = node("", entry0.context.toString());
        edge = dashed(last, node, label(entry0));
      }
      nodes.add(node);
      if (edge != null) edges.add(edge);
      if (entry instanceof Trace.Entry.Result) return;
      last = node;
    }
  }

  // -------------------------------------------------------------------------
  //
  // event handling
  //
  // -------------------------------------------------------------------------

  // the module defining the event handlers
  public static class TraceModule extends ZestFxModule 
  {
    @Override
    protected void enableAdapterMapInjection() 
    {
      install(new AdapterInjectionSupport(LoggingMode.PRODUCTION));
    }
    @Override
    protected void bindNodePartAdapters(MapBinder<AdapterKey<?>, Object> adapterMapBinder) {
      // layout
      adapterMapBinder.addBinding(AdapterKey.defaultRole()).to(NodeLayoutBehavior.class);
      // pruning
      adapterMapBinder.addBinding(AdapterKey.defaultRole()).to(HidePolicy.class);
      adapterMapBinder.addBinding(AdapterKey.defaultRole()).to(ShowHiddenNeighborsPolicy.class);
      adapterMapBinder.addBinding(AdapterKey.defaultRole()).to(NodeHidingBehavior.class);

      // translate on-drag
      adapterMapBinder.addBinding(AdapterKey.defaultRole()).to(TranslateSelectedAndRelocateLabelsOnDragHandler.class);

      // show hidden neighbors on-type
      adapterMapBinder.addBinding(AdapterKey.role("show-hidden-neighbors"))
      .to(ShowHiddenNeighborsOnTypeHandler.class);

      // hide on-type
      adapterMapBinder.addBinding(AdapterKey.role("hide")).to(HideOnTypeHandler.class);

      adapterMapBinder.addBinding(AdapterKey.role("open-nested-graph")).to(TraceOpenNestedGraphOnDoubleClickHandler.class);

      // transform policy for relocation
      adapterMapBinder.addBinding(AdapterKey.defaultRole()).to(TransformPolicy.class);

      // resize policy to resize nesting nodes
      adapterMapBinder.addBinding(AdapterKey.defaultRole()).to(ResizePolicy.class);

      // anchor provider
      adapterMapBinder.addBinding(AdapterKey.defaultRole()).to(NodePartAnchorProvider.class);

      // feedback and handles
      adapterMapBinder
      .addBinding(AdapterKey.role(DefaultSelectionHandlePartFactory.SELECTION_HANDLES_GEOMETRY_PROVIDER))
      .toProvider(new Provider<ShapeBoundsProvider>() {
        @Override
        public ShapeBoundsProvider get() {
          return new ShapeBoundsProvider(1);
        }
      });

      // selection feedback provider
      adapterMapBinder
      .addBinding(AdapterKey.role(DefaultSelectionFeedbackPartFactory.SELECTION_FEEDBACK_GEOMETRY_PROVIDER))
      .toProvider(new Provider<ShapeBoundsProvider>() {
        @Override
        public ShapeBoundsProvider get() {
          return new ShapeBoundsProvider(0.5);
        }
      });

      // selection link feedback provider
      adapterMapBinder
      .addBinding(
          AdapterKey.role(DefaultSelectionFeedbackPartFactory.SELECTION_LINK_FEEDBACK_GEOMETRY_PROVIDER))
      .toProvider(new Provider<ShapeBoundsProvider>() {
        @Override
        public ShapeBoundsProvider get() {
          return new ShapeBoundsProvider(0.5);
        }
      });

      // hover feedback provider
      adapterMapBinder.addBinding(AdapterKey.role(DefaultHoverFeedbackPartFactory.HOVER_FEEDBACK_GEOMETRY_PROVIDER))
      .to(ShapeBoundsProvider.class);

      // focus feedback provider
      adapterMapBinder.addBinding(AdapterKey.role(DefaultFocusFeedbackPartFactory.FOCUS_FEEDBACK_GEOMETRY_PROVIDER))
      .toProvider(new Provider<ShapeBoundsProvider>() {
        @Override
        public ShapeBoundsProvider get() {
          return new ShapeBoundsProvider(0.5);
        }
      });

      // hover on-hover
      adapterMapBinder.addBinding(AdapterKey.defaultRole()).to(HoverOnHoverHandler.class);

      // normalize on drag
      adapterMapBinder.addBinding(AdapterKey.defaultRole()).to(ConnectedSupport.class);
    }

    @Override
    protected void bindIRootPartAdaptersForContentViewer(MapBinder<AdapterKey<?>, Object> adapterMapBinder) {
      super.bindIRootPartAdaptersForContentViewer(adapterMapBinder);

      adapterMapBinder.addBinding(AdapterKey.role("open-parent-graph")).to(TraceOpenParentGraphOnDoubleClickHandler.class);

      // the following are commented out because they trigger error messages

      // keyboard focus traversal
      // adapterMapBinder.addBinding(AdapterKey.defaultRole()).to(TraverseFocusOnTypeHandler.class);

      // select focused on type
      // adapterMapBinder.addBinding(AdapterKey.defaultRole()).to(SelectFocusedOnTypeHandler.class);
    }
  }

  // derived from org.eclipse.gef.zest.fx.handlers.OpenNestedGraphOnDoubleClickHandler
  public static class TraceOpenNestedGraphOnDoubleClickHandler extends AbstractHandler implements IOnClickHandler {

    @Override
    public void click(MouseEvent event) {
      if (event.getClickCount() == 2) {
        // double click, so open nested graph, if it exists
        final Graph nestedGraph = getHost().getContent().getNestedGraph();
        if (nestedGraph != null) {
          IViewer viewer = getHost().getRoot().getViewer();
          try {
            // navigate to nested graph
            viewer.getDomain().execute(new NavigateOperation(viewer, nestedGraph, true),
                new NullProgressMonitor());
            String header = traceWindow.headers.get(nestedGraph);
            if (header == null) header = "(unknown)";
            traceWindow.shell.setText(header);
          } catch (ExecutionException e) {
            throw new RuntimeException(e);
          }
        }
      }
    }

    @Override
    public NodePart getHost() {
      return (NodePart) super.getHost();
    }
  }

  // derived from org.eclipse.gef.zest.fx.handlers.OpenParentGraphOnDoubleClickHandler
  public static class TraceOpenParentGraphOnDoubleClickHandler extends AbstractHandler implements IOnClickHandler {

    @Override
    public void click(MouseEvent event) {
      if (event.getClickCount() == 2) {
        // do nothing in case there is an explicit event target
        if (isRegistered(event.getTarget()) && !isRegisteredForHost(event.getTarget())) {
          return;
        }

        // double click, so open nesting graph, if it exists
        IViewer viewer = getHost().getRoot().getViewer();
        final Graph currentGraph = (Graph) viewer.getContents().get(0);
        // due to some strange bug the currentGraph has to be used instead ???
        // final Graph nestingGraph = currentGraph.getNestingNode() != null ? currentGraph.getNestingNode().getGraph()
        //     : null;
        if (currentGraph != null) 
        {
          try {
            // navigate to parent graph
            viewer.getDomain().execute(new NavigateOperation(viewer, currentGraph, false),
                new NullProgressMonitor());
            String header = traceWindow.headers.get(currentGraph);
            if (header == null) header = "(unknown)";
            traceWindow.shell.setText(header);
          } catch (ExecutionException e) {
            throw new RuntimeException(e);
          }
        }
      }
    }

    @Override
    public ZestFxRootPart getHost() {
      return (ZestFxRootPart) super.getHost();
    }
  }
}
